/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <immintrin.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include <sched.h>
#include <time.h>
#include <math.h>

#include "rand.h"


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

int log_2(size_t n)
{
	int p = 0;
	size_t m = 1;
	while (m < n) {
		p++;
		m += m;
	}
	assert(m == n);
	return p;
}

void *mamalloc(size_t size)
{
	void *ptr = NULL;
	return posix_memalign(&ptr, 64, size) ? NULL : ptr;
}

void *align(const void *p)
{
	while (63 & (size_t) p) p++;
	return (void*) p;
}

size_t scalar(const uint32_t *keys, const uint32_t *vals, size_t tuples,
              const uint32_t *bloom_filter, int log_bloom_filter_bits,
              const uint32_t *hash_factors, size_t hash_functions,
              uint32_t *keys_out, uint32_t *vals_out)
{
	size_t i, o;
	const uint8_t shift = 32 - log_bloom_filter_bits;
	for (i = o = 0 ; i != tuples ; ++i) {
		uint32_t key = keys[i];
		size_t f = 0;
		do {
			uint32_t h = key * hash_factors[f];
			h >>= shift;
			asm goto("btl	%1, (%0)\n\t"
			         "jnc	%l[failure]"
			:: "r"(bloom_filter), "r"(h)
			: "cc" : failure);
		} while (++f != hash_functions);
		uint32_t val = vals[i];
#ifndef __MIC
		_mm_stream_si32(&((int*) keys_out)[o], key);
		_mm_stream_si32(&((int*) vals_out)[o], val);
#else
		keys_out[o] = key;
		vals_out[o] = val;
#endif
		o++;
failure:;
	}
	return o;
}

#ifndef __MIC

__m256i load_to_vector(const uint32_t *a, size_t n)
{
	size_t i;
	uint32_t b[8];
	assert(n <= 8);
	for (i = 0 ; i != 8 ; ++i)
		b[i] = i < n ? a[i] : 0;
	return _mm256_loadu_si256((__m256i*) b);
}

const uint64_t perm[256] = {0x0706050403020100ull,
     0x0007060504030201ull, 0x0107060504030200ull, 0x0001070605040302ull,
     0x0207060504030100ull, 0x0002070605040301ull, 0x0102070605040300ull,
     0x0001020706050403ull, 0x0307060504020100ull, 0x0003070605040201ull,
     0x0103070605040200ull, 0x0001030706050402ull, 0x0203070605040100ull,
     0x0002030706050401ull, 0x0102030706050400ull, 0x0001020307060504ull,
     0x0407060503020100ull, 0x0004070605030201ull, 0x0104070605030200ull,
     0x0001040706050302ull, 0x0204070605030100ull, 0x0002040706050301ull,
     0x0102040706050300ull, 0x0001020407060503ull, 0x0304070605020100ull,
     0x0003040706050201ull, 0x0103040706050200ull, 0x0001030407060502ull,
     0x0203040706050100ull, 0x0002030407060501ull, 0x0102030407060500ull,
     0x0001020304070605ull, 0x0507060403020100ull, 0x0005070604030201ull,
     0x0105070604030200ull, 0x0001050706040302ull, 0x0205070604030100ull,
     0x0002050706040301ull, 0x0102050706040300ull, 0x0001020507060403ull,
     0x0305070604020100ull, 0x0003050706040201ull, 0x0103050706040200ull,
     0x0001030507060402ull, 0x0203050706040100ull, 0x0002030507060401ull,
     0x0102030507060400ull, 0x0001020305070604ull, 0x0405070603020100ull,
     0x0004050706030201ull, 0x0104050706030200ull, 0x0001040507060302ull,
     0x0204050706030100ull, 0x0002040507060301ull, 0x0102040507060300ull,
     0x0001020405070603ull, 0x0304050706020100ull, 0x0003040507060201ull,
     0x0103040507060200ull, 0x0001030405070602ull, 0x0203040507060100ull,
     0x0002030405070601ull, 0x0102030405070600ull, 0x0001020304050706ull,
     0x0607050403020100ull, 0x0006070504030201ull, 0x0106070504030200ull,
     0x0001060705040302ull, 0x0206070504030100ull, 0x0002060705040301ull,
     0x0102060705040300ull, 0x0001020607050403ull, 0x0306070504020100ull,
     0x0003060705040201ull, 0x0103060705040200ull, 0x0001030607050402ull,
     0x0203060705040100ull, 0x0002030607050401ull, 0x0102030607050400ull,
     0x0001020306070504ull, 0x0406070503020100ull, 0x0004060705030201ull,
     0x0104060705030200ull, 0x0001040607050302ull, 0x0204060705030100ull,
     0x0002040607050301ull, 0x0102040607050300ull, 0x0001020406070503ull,
     0x0304060705020100ull, 0x0003040607050201ull, 0x0103040607050200ull,
     0x0001030406070502ull, 0x0203040607050100ull, 0x0002030406070501ull,
     0x0102030406070500ull, 0x0001020304060705ull, 0x0506070403020100ull,
     0x0005060704030201ull, 0x0105060704030200ull, 0x0001050607040302ull,
     0x0205060704030100ull, 0x0002050607040301ull, 0x0102050607040300ull,
     0x0001020506070403ull, 0x0305060704020100ull, 0x0003050607040201ull,
     0x0103050607040200ull, 0x0001030506070402ull, 0x0203050607040100ull,
     0x0002030506070401ull, 0x0102030506070400ull, 0x0001020305060704ull,
     0x0405060703020100ull, 0x0004050607030201ull, 0x0104050607030200ull,
     0x0001040506070302ull, 0x0204050607030100ull, 0x0002040506070301ull,
     0x0102040506070300ull, 0x0001020405060703ull, 0x0304050607020100ull,
     0x0003040506070201ull, 0x0103040506070200ull, 0x0001030405060702ull,
     0x0203040506070100ull, 0x0002030405060701ull, 0x0102030405060700ull,
     0x0001020304050607ull, 0x0706050403020100ull, 0x0007060504030201ull,
     0x0107060504030200ull, 0x0001070605040302ull, 0x0207060504030100ull,
     0x0002070605040301ull, 0x0102070605040300ull, 0x0001020706050403ull,
     0x0307060504020100ull, 0x0003070605040201ull, 0x0103070605040200ull,
     0x0001030706050402ull, 0x0203070605040100ull, 0x0002030706050401ull,
     0x0102030706050400ull, 0x0001020307060504ull, 0x0407060503020100ull,
     0x0004070605030201ull, 0x0104070605030200ull, 0x0001040706050302ull,
     0x0204070605030100ull, 0x0002040706050301ull, 0x0102040706050300ull,
     0x0001020407060503ull, 0x0304070605020100ull, 0x0003040706050201ull,
     0x0103040706050200ull, 0x0001030407060502ull, 0x0203040706050100ull,
     0x0002030407060501ull, 0x0102030407060500ull, 0x0001020304070605ull,
     0x0507060403020100ull, 0x0005070604030201ull, 0x0105070604030200ull,
     0x0001050706040302ull, 0x0205070604030100ull, 0x0002050706040301ull,
     0x0102050706040300ull, 0x0001020507060403ull, 0x0305070604020100ull,
     0x0003050706040201ull, 0x0103050706040200ull, 0x0001030507060402ull,
     0x0203050706040100ull, 0x0002030507060401ull, 0x0102030507060400ull,
     0x0001020305070604ull, 0x0405070603020100ull, 0x0004050706030201ull,
     0x0104050706030200ull, 0x0001040507060302ull, 0x0204050706030100ull,
     0x0002040507060301ull, 0x0102040507060300ull, 0x0001020405070603ull,
     0x0304050706020100ull, 0x0003040507060201ull, 0x0103040507060200ull,
     0x0001030405070602ull, 0x0203040507060100ull, 0x0002030405070601ull,
     0x0102030405070600ull, 0x0001020304050706ull, 0x0607050403020100ull,
     0x0006070504030201ull, 0x0106070504030200ull, 0x0001060705040302ull,
     0x0206070504030100ull, 0x0002060705040301ull, 0x0102060705040300ull,
     0x0001020607050403ull, 0x0306070504020100ull, 0x0003060705040201ull,
     0x0103060705040200ull, 0x0001030607050402ull, 0x0203060705040100ull,
     0x0002030607050401ull, 0x0102030607050400ull, 0x0001020306070504ull,
     0x0406070503020100ull, 0x0004060705030201ull, 0x0104060705030200ull,
     0x0001040607050302ull, 0x0204060705030100ull, 0x0002040607050301ull,
     0x0102040607050300ull, 0x0001020406070503ull, 0x0304060705020100ull,
     0x0003040607050201ull, 0x0103040607050200ull, 0x0001030406070502ull,
     0x0203040607050100ull, 0x0002030406070501ull, 0x0102030406070500ull,
     0x0001020304060705ull, 0x0506070403020100ull, 0x0005060704030201ull,
     0x0105060704030200ull, 0x0001050607040302ull, 0x0205060704030100ull,
     0x0002050607040301ull, 0x0102050607040300ull, 0x0001020506070403ull,
     0x0305060704020100ull, 0x0003050607040201ull, 0x0103050607040200ull,
     0x0001030506070402ull, 0x0203050607040100ull, 0x0002030506070401ull,
     0x0102030506070400ull, 0x0001020305060704ull, 0x0405060703020100ull,
     0x0004050607030201ull, 0x0104050607030200ull, 0x0001040506070302ull,
     0x0204050607030100ull, 0x0002040506070301ull, 0x0102040506070300ull,
     0x0001020405060703ull, 0x0304050607020100ull, 0x0003040506070201ull,
     0x0103040506070200ull, 0x0001030405060702ull, 0x0203040506070100ull,
     0x0002030405060701ull, 0x0102030405060700ull, 0x0001020304050607ull};

inline size_t _mm256_movemask_epi32(__m256i x)
{
	size_t r;
	asm("vmovmskps	%1, %0" : "=r"(r) : "x"(x));
	return r;
}

size_t vector(const uint32_t *keys, const uint32_t *vals, size_t tuples,
              const uint32_t *bloom_filter, int log_bloom_filter_bits,
              const uint32_t *hash_factors, size_t hash_functions,
              uint32_t *keys_out, uint32_t *vals_out)
{
	size_t i = 0, o = 0, b = 0;
	// buffer space
	const size_t buf_size = 128;
	uint32_t buf[(buf_size + 8) * 2 + 15];
	uint32_t *keys_buf = align(buf);
	uint32_t *vals_buf = &keys_buf[buf_size + 8];
	// constants
	const __m256i factors = load_to_vector(hash_factors, hash_functions);
	const __m128i shift = _mm_cvtsi32_si128(32 - log_bloom_filter_bits);
	const __m256i functions = _mm256_set1_epi32(hash_functions);
	const __m256i mask_31 = _mm256_set1_epi32(31);
	const __m256i mask_1 = _mm256_set1_epi32(1);
	const __m256i mask_0 = _mm256_set1_epi32(0);
	// non-constants
	__m256i key, val, fun;
	__m256i inv = _mm256_set1_epi32(-1);
	while (i + 8 <= tuples) {
		// load new items (mask out reloads)
		__m256i new_key = _mm256_maskload_epi32((int*) &keys[i], inv);
		__m256i new_val = _mm256_maskload_epi32((int*) &vals[i], inv);
		// reset old items
		fun = _mm256_andnot_si256(inv, fun);
		key = _mm256_andnot_si256(inv, key);
		val = _mm256_andnot_si256(inv, val);
		// pick hash function
		__m256i factor = _mm256_permutevar8x32_epi32(factors, fun);
		fun = _mm256_add_epi32(fun, mask_1);
		// combine old with new items
		key = _mm256_or_si256(key, new_key);
		val = _mm256_or_si256(val, new_val);
		// hash the keys and check who is almost done
		__m256i hash = _mm256_mullo_epi32(key, factor);
		__m256i last = _mm256_cmpeq_epi32(fun, functions);
		hash = _mm256_srl_epi32(hash, shift);
		// check bitmap
		__m256i div_32 = _mm256_srli_epi32(hash, 5);
		div_32 = _mm256_i32gather_epi32((int*) bloom_filter, div_32, 4);
		__m256i mod_32 = _mm256_and_si256(hash, mask_31);
		mod_32 = _mm256_sllv_epi32(mask_1, mod_32);
		div_32 = _mm256_and_si256(div_32, mod_32);
		inv = _mm256_cmpeq_epi32(div_32, mask_0);
		// branch to print winners
		if (!_mm256_testz_si256(div_32, last)) {
			__m256i out = _mm256_andnot_si256(inv, last);
			size_t j = _mm256_movemask_epi32(out);
			__m128i perm_comp = _mm_loadl_epi64((__m128i*) &perm[j ^ 255]);
			__m256i perm = _mm256_cvtepi8_epi32(perm_comp);
			out = _mm256_permutevar8x32_epi32(out, perm);
			__m256i key_out = _mm256_permutevar8x32_epi32(key, perm);
			__m256i val_out = _mm256_permutevar8x32_epi32(val, perm);
			_mm256_maskstore_epi32((int*) &keys_buf[b], out, key_out);
			_mm256_maskstore_epi32((int*) &vals_buf[b], out, val_out);
			b += _mm_popcnt_u64(j);
			if (b > buf_size) {
				j = 0;
				do {
					__m256i k = _mm256_load_si256((__m256i*) &keys_buf[j]);
					__m256i v = _mm256_load_si256((__m256i*) &vals_buf[j]);
					_mm256_stream_si256((__m256i*) &keys_out[o], k);
					_mm256_stream_si256((__m256i*) &vals_out[o], v);
					j += 8;
					o += 8;
				} while (j != buf_size);
				__m256i k = _mm256_load_si256((__m256i*) &keys_buf[j]);
				__m256i v = _mm256_load_si256((__m256i*) &vals_buf[j]);
				_mm256_store_si256((__m256i*) &keys_buf[0], k);
				_mm256_store_si256((__m256i*) &vals_buf[0], v);
				b -= buf_size;
			}
		}
		// permute to get rid of losers (and winners)
		inv = _mm256_or_si256(inv, last);
		size_t j = _mm256_movemask_epi32(inv);
		__m128i perm_comp = _mm_loadl_epi64((__m128i*) &perm[j]);
		__m256i perm = _mm256_cvtepi8_epi32(perm_comp);
		inv = _mm256_permutevar8x32_epi32(inv, perm);
		fun = _mm256_permutevar8x32_epi32(fun, perm);
		key = _mm256_permutevar8x32_epi32(key, perm);
		val = _mm256_permutevar8x32_epi32(val, perm);
		i += _mm_popcnt_u64(j);
	}
	// flush buffer
	size_t f, j;
	for (j = 0 ; j != b ; ++j) {
		_mm_stream_si32(&((int*) keys_out)[o], keys_buf[j]);
		_mm_stream_si32(&((int*) vals_out)[o], vals_buf[j]);
		o++;
	}
	// copy last items
	uint32_t l_keys[8];
	uint32_t l_vals[8];
	_mm256_storeu_si256((__m256i*) l_keys, key);
	_mm256_storeu_si256((__m256i*) l_vals, val);
	j = 8 - _mm_popcnt_u64(_mm256_movemask_epi32(inv));
	i += j;
	while (i != tuples) {
		l_keys[j] = keys[i];
		l_vals[j] = vals[i];
		i++;
		j++;
	}
	// process remaining items with scalar code
	for (i = 0 ; i != j ; ++i) {
		uint32_t k = l_keys[i];
		uint32_t v = l_vals[i];
		for (f = 0 ; f != hash_functions ; ++f) {
			uint32_t h = k * hash_factors[f];
			h >>= 32 - log_bloom_filter_bits;
			asm goto("btl	%1, (%0)\n\t"
			         "jnc	%l[failure]"
			:: "r"(bloom_filter), "r"(h)
			: "cc" : failure);
		}
		_mm_stream_si32(&((int*) keys_out)[o], k);
		_mm_stream_si32(&((int*) vals_out)[o], v);
		o++;
failure:;
	}
	return o;
}

#else

__m512i load_to_vector(const uint32_t *a, size_t n)
{
	assert(n <= 16);
	__mmask16 k = _mm512_int2mask((1 << n) - 1);
	__m512i v = _mm512_xor_epi32(v, v);
	v = _mm512_mask_loadunpacklo_epi32(v, k, &a[0]);
	v = _mm512_mask_loadunpackhi_epi32(v, k, &a[16]);
	return v;
}

size_t vector(const uint32_t *keys, const uint32_t *vals, size_t tuples,
              const uint32_t *bloom_filter, int log_bloom_filter_bits,
              const uint32_t *hash_factors, size_t hash_functions,
              uint32_t *keys_out, uint32_t *vals_out)
{
	size_t i = 0, o = 0, b = 0;
	// buffer space
	const size_t buf_size = 128;
	uint32_t buf[(buf_size + 16) * 2 + 15];
	uint32_t *keys_buf = align(buf);
	uint32_t *vals_buf = &keys_buf[buf_size + 16];
	// constants
	const __m512i factors = load_to_vector(hash_factors, hash_functions);
	const __m512i shift = _mm512_set1_epi32(32 - log_bloom_filter_bits);
	const __m512i functions = _mm512_set1_epi32(hash_functions);
	const __m512i mask_31 = _mm512_set1_epi32(31);
	const __m512i mask_1 = _mm512_set1_epi32(1);
	const __mmask16 z = _mm512_int2mask(0);
	// non-constants
	__m512i key, val, fun;
	__mmask16 k = _mm512_kxnor(k, k);
	while (i + 16 <= tuples) {
		// load new items (mask out reloads)
		key = _mm512_mask_loadunpacklo_epi32(key, k, &keys[i]);
		key = _mm512_mask_loadunpackhi_epi32(key, k, &keys[i + 16]);
		val = _mm512_mask_loadunpacklo_epi32(val, k, &vals[i]);
		val = _mm512_mask_loadunpackhi_epi32(val, k, &vals[i + 16]);
		i += _mm_countbits_64(_mm512_kconcatlo_64(z, k));
		// pick hash function
		fun = _mm512_mask_xor_epi32(fun, k, fun, fun);
		__m512i factor = _mm512_permutevar_epi32(fun, factors);
		fun = _mm512_add_epi32(fun, mask_1);
		// hash the keys and check who is almost done
		__m512i hash = _mm512_mullo_epi32(key, factor);
		__mmask16 q = _mm512_cmpeq_epi32_mask(fun, functions);
		hash = _mm512_srlv_epi32(hash, shift);
		// check bitmap
		__m512i div_32 = _mm512_srli_epi32(hash, 5);
		div_32 = _mm512_i32gather_epi32(div_32, bloom_filter, 4);
		__m512i mod_32 = _mm512_and_epi32(hash, mask_31);
		mod_32 = _mm512_sllv_epi32(mask_1, mod_32);
		div_32 = _mm512_and_si512(div_32, mod_32);
		k = _mm512_cmpneq_epi32_mask(div_32, mod_32);
		q = _mm512_kandnr(q, k);
		// branch to print winners
		if (!_mm512_kortestz(q, q)) {
			_mm512_mask_packstorelo_epi32(&keys_buf[b +  0], q, key);
			_mm512_mask_packstorehi_epi32(&keys_buf[b + 16], q, key);
			_mm512_mask_packstorelo_epi32(&vals_buf[b +  0], q, val);
			_mm512_mask_packstorehi_epi32(&vals_buf[b + 16], q, val);
			b += _mm_countbits_64(_mm512_kconcatlo_64(z, q));
			k = _mm512_kor(k, q);
			if (b > buf_size) {
				size_t j = 0;
				do {
					__m512 x = _mm512_load_ps(&keys_buf[j]);
					__m512 y = _mm512_load_ps(&vals_buf[j]);
					_mm512_storenrngo_ps(&keys_out[o], x);
					_mm512_storenrngo_ps(&vals_out[o], y);
					o += 16;
					j += 16;
				} while (j != buf_size);
				__m512 x = _mm512_load_ps(&keys_buf[j]);
				__m512 y = _mm512_load_ps(&vals_buf[j]);
				_mm512_store_ps(&keys_buf[0], x);
				_mm512_store_ps(&vals_buf[0], y);
				b -= buf_size;
			}
		}
	}
	// flush buffer
	size_t j;
	for (j = 0 ; j != b ; ++j) {
		keys_out[o] = keys_buf[j];
		vals_out[o] = vals_buf[j];
		o++;
	}
	// save last items in hash function order (to remain stable)
	fun = _mm512_mask_xor_epi32(fun, k, fun, fun);
	fun = _mm512_sub_epi32(functions, fun);
	uint32_t l_keys[32];
	uint32_t l_vals[32];
	size_t f = j = 0;
	do {
		k = _mm512_cmplt_epu32_mask(fun, mask_1);
		_mm512_mask_packstorelo_epi32(&l_keys[j +  0], k, key);
		_mm512_mask_packstorehi_epi32(&l_keys[j + 16], k, key);
		_mm512_mask_packstorelo_epi32(&l_vals[j +  0], k, val);
		_mm512_mask_packstorehi_epi32(&l_vals[j + 16], k, val);
		fun = _mm512_sub_epi32(fun, mask_1);
		j += _mm_countbits_64(_mm512_kconcatlo_64(z, k));
	} while (++f != hash_functions);
	while (i != tuples) {
		l_keys[j] = keys[i];
		l_vals[j] = vals[i];
		i++;
		j++;
	}
	// process remaining items with scalar code
	const uint8_t s = 32 - log_bloom_filter_bits;
	for (i = 0 ; i != j ; ++i) {
		uint32_t k = l_keys[i];
		uint32_t v = l_vals[i];
		for (f = 0 ; f != hash_functions ; ++f) {
			uint32_t h = k * hash_factors[f];
			h >>= s;
			asm goto("btl	%1, (%0)\n\t"
			         "jnc	%l[failure]"
			:: "r"(bloom_filter), "r"(h)
			: "cc" : failure);
		}
		keys_out[o] = k;
		vals_out[o] = v;
		o++;
failure:;
	}
	return o;
}

#endif

typedef struct {
	pthread_t id;
	int threads;
	int thread;
	int seed;
	int log_bloom_filter_bits;
	const uint32_t *hash_factors;
	size_t hash_functions;
	double filter_rate;
	size_t outer_tuples;
	size_t inner_tuples;
	size_t errors;
	size_t passed;
	uint64_t *times;
	volatile uint32_t *bloom_filter;
	volatile uint32_t *unique_table;
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread, d->threads);
	size_t i, r, m, f, b = 0;
	// thread local cardinalities
	size_t local_inner_tuples = d->inner_tuples / d->threads;
	size_t local_outer_tuples = d->outer_tuples / d->threads;
	if (d->thread + 1 == d->threads) {
		local_inner_tuples += d->inner_tuples % d->threads;
		local_outer_tuples += d->outer_tuples % d->threads;
	}
	// tables in volatile pointers for safe read-write access
	volatile uint32_t *vol_unique_table = d->unique_table;
	volatile uint32_t *vol_bloom_filter = d->bloom_filter;
	// tables in non-volatile pointers for read-only access
	uint32_t *unique_table = (uint32_t*) d->unique_table;
	uint32_t *bloom_filter = (uint32_t*) d->bloom_filter;
	// initialize random generator
	rand32_t *gen = rand32_init(d->seed);
	uint32_t payload_factor = rand32_next(gen) | 1;
	uint32_t ordered_factor = rand32_next(gen) | 1;
	// configure the bloom filter based on given error rate
	const uint32_t *hash_factors = d->hash_factors;
	size_t hash_functions = d->hash_functions;
	int log_bloom_filter_bits = d->log_bloom_filter_bits;
	size_t bloom_filter_bits = ((size_t) 1) << log_bloom_filter_bits;
	// initialize inner table (built) => unique keys only
	size_t unique_buckets = d->inner_tuples / 0.7;
	const uint32_t knuth_factor = 0x9e3779b1;
	for (i = 0 ; i != local_inner_tuples ; ++i) {
		uint32_t key, tab;
		// generate new unique key
		do {
			key = rand32_next(gen);
			uint64_t h = (uint32_t) (key * knuth_factor);
			h = (h * unique_buckets) >> 32;
			while ((tab = vol_unique_table[h]) != 0 && tab != key)
				if (++h == unique_buckets) h = 0;
			// atomic update in table
			asm("lock\n\t"
			    "cmpxchgl	%2, (%1)"
			: "=a"(tab)
			: "r"(&vol_unique_table[h]), "r"(key), "a"(tab)
			: "cc", "memory");
		} while (tab == key);
		// add key in the Bloom filter
		for (f = 0 ; f != hash_functions ; ++f) {
			uint64_t h = (uint32_t) (key * hash_factors[f]);
			h = (h * bloom_filter_bits) >> 32;
			// atomic bit set in the Bloom filter
			asm("lock\n\t"
			    "btsl	%1, (%0)"
			:: "r"(vol_bloom_filter), "r"((uint32_t) h)
			: "cc", "memory");
		}
	}
	pthread_barrier_wait(&d->barrier[b++]);
	// initialize outer table (probed) => keys & payloads
	uint32_t *keys = malloc(local_outer_tuples * sizeof(uint32_t));
	uint32_t *vals = malloc(local_outer_tuples * sizeof(uint32_t));
	size_t passed_correctly = 0;
	uint32_t error_limit = ~0;
	error_limit *= d->filter_rate;
	for (i = r = 0 ; i != local_outer_tuples ; ++i) {
		uint32_t key, tab;
		uint64_t h;
		if (rand32_next(gen) > error_limit)
			do {
				key = rand32_next(gen);
				h = (uint32_t) (key * knuth_factor);
				h = (h * unique_buckets) >> 32;
				while ((tab = unique_table[h]) != 0 && tab != key)
					if (++h == unique_buckets) h = 0;
			} while (key == tab);
		else {
			do {
				h = rand32_next(gen);
				h = (h * unique_buckets) >> 32;
			} while ((key = unique_table[h]) == 0);
			passed_correctly++;
		}
		keys[i] = key;
		vals[i] = key * payload_factor;
	}
	free(gen);
	size_t local_output_tuples = local_outer_tuples * (d->filter_rate + 0.05);
	if (local_output_tuples > local_outer_tuples)
		local_output_tuples = local_outer_tuples;
	uint32_t *keys_out = mamalloc(local_output_tuples * sizeof(uint32_t));
	uint32_t *vals_out = mamalloc(local_output_tuples * sizeof(uint32_t));
	for (i = 0 ; i != local_output_tuples ; ++i) {
		keys_out[i] = 0xDEADBEEF;
		vals_out[i] = 0xCAFEBABE;
	}
	size_t (*method[2]) (const uint32_t *keys, const uint32_t *vals, size_t tuples,
	                     const uint32_t *bloom_filter, int log_bloom_filter_bits,
	                     const uint32_t *hash_factors, size_t hash_functions,
	                     uint32_t *keys_out, uint32_t *vals_out) =
		{scalar, vector};
	uint32_t sum = 0;
	uint32_t sum_ordered = 0;
	for (m = 0 ; m != 2 ; ++m) {
		pthread_barrier_wait(&d->barrier[b++]);
		uint64_t t = thread_time();
		size_t passed = method[m](keys, vals, local_outer_tuples,
		                          bloom_filter, log_bloom_filter_bits,
		                          hash_factors, hash_functions,
	                              keys_out, vals_out);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		// save and check results
		d->times[m] = t;
		assert(passed <= local_output_tuples);
		if (m) assert(passed == d->passed);
		else d->passed = passed;
		// check results
		size_t errors = 0;
		uint32_t checksum = 0;
		uint32_t checksum_ordered = 0;
		for (i = 0 ; i != passed ; ++i) {
			// check if key qualifies the filter
			uint32_t key = keys_out[i];
			for (f = 0 ; f != hash_functions ; ++f) {
				uint64_t h = (uint32_t) (key * hash_factors[f]);
				h = (h * bloom_filter_bits) >> 32;
				assert(((bloom_filter[h >> 5]) & (1 << (h & 31))) != 0);
			}
			// check if false positive or not
			uint64_t h = (uint32_t) (key * knuth_factor);
			h = (h * unique_buckets) >> 32;
			while (unique_table[h] != 0 && unique_table[h] != key)
				if (++h == unique_buckets) h = 0;
			errors += unique_table[h] == 0 ? 1 : 0;
			// check matching payload and update checksums
			assert(vals_out[i] == key * payload_factor);
			checksum += key;
			checksum_ordered *= ordered_factor;
			checksum_ordered += key;
		}
		if (m) {
			assert(errors == d->errors);
			assert(sum         == checksum);
			assert(sum_ordered == checksum_ordered);
		} else {
			d->errors = errors;
			sum         = checksum;
			sum_ordered = checksum_ordered;
			assert(passed - errors == passed_correctly);
		}
	}
	free(keys);
	free(vals);
	free(keys_out);
	free(vals_out);
	pthread_exit(NULL);
}

int parse_log_size(const char *s)
{
	char *p = NULL;
	int bits = 0;
	long long base = 1;
	long long num = strtol(s, &p, 10);
	while (base < num) {
		base += base;
		bits++;
	}
	assert(base == num);
	while (isspace(*p)) p++;
	if      (*p == 'K') bits += 10;
	else if (*p == 'M') bits += 20;
	else if (*p == 'G') bits += 30;
	else p--;
	p++;
	assert(*p == 'B');
	p++;
	while (isspace(*p)) p++;
	assert(*p == 0);
	return bits;
}

char *print_log_size(int bits)
{
	static char buf[16];
	if      (bits >= 30) sprintf(buf, "%d GB", 1 << (bits - 30));
	else if (bits >= 20) sprintf(buf, "%d MB", 1 << (bits - 20));
	else if (bits >= 10) sprintf(buf, "%d KB", 1 << (bits - 10));
	else                 sprintf(buf, "%d  B", 1 << bits);
	return buf;
}

#ifndef __INTEL_COMPILER
uint64_t _mm_tzcnt_64(uint64_t i)
{
	uint64_t o;
	asm("tzcntq	%1, %0" : "=r"(o) : "r"(i) : "cc");
	return o;
}
#endif

int main(int argc, char **argv)
{
	int i, j, t, threads = hardware_threads();
	int log_bloom_filter_bits = argc > 1 ? parse_log_size(argv[1]) + 3 : 17;
	double bits_per_item = argc > 2 ? atof(argv[2]) : 10;
	int hash_functions = argc > 3 ? atoi(argv[3]) : 5;
	double filter_rate = argc > 4 ? atof(argv[4]) : 0.05;
	size_t bloom_filter_bits = ((size_t) 1) << log_bloom_filter_bits;
	size_t outer_tuples = argc > 5 ? atoll(argv[5]) : 1000 * 1000 * 1000;
	size_t inner_tuples = bloom_filter_bits / bits_per_item;
	// check arguments
	assert(filter_rate >= 0.0 && filter_rate <= 1.0);
	assert(threads > 0 && threads <= hardware_threads());
#ifdef __MIC
	assert(hash_functions >= 1 && hash_functions <= 16);
	fprintf(stderr, "Running on Phi!\n");
#else
	assert(hash_functions >= 1 && hash_functions <= 8);
	fprintf(stderr, "Running on Haswell!\n");
#endif
	// print configuration
	double error = pow(1 - exp(hash_functions * -1.0 / bits_per_item), hash_functions);
	fprintf(stderr, "Threads: %d\n", threads);
	fprintf(stderr, "Outer table tuples: %ld\n", outer_tuples);
	fprintf(stderr, "Inner table tuples: %ld\n", inner_tuples);
	fprintf(stderr, "Bloom filter size: %s\n", print_log_size(log_bloom_filter_bits - 3));
	fprintf(stderr, "Bits / item: %.2f\n", bits_per_item);
	fprintf(stderr, "Hash functions: %d\n", hash_functions);
	fprintf(stderr, "Expected filter rate: %.5f%%\n", filter_rate * 100.0);
	fprintf(stderr, "Expected error  rate: %.5f%%\n", error * 100.0);
	// generate distinct primes as multiplicative hashing factors
	srand(time(NULL));
	uint32_t hash_factors[16];
	for (i = 0 ; i != hash_functions ; ++i) {
		uint32_t f = (rand() << 1) | 1;
		for (j = 0 ; j != i ; ++j) {
			assert(j < 16);
			if (_mm_tzcnt_64(hash_factors[j] - f) > 5) break;
		}
		if (i != j) i--;
		else hash_factors[i] = f;
	}
	// allocate bloom filter and table for distinct keys
	size_t unique_buckets = inner_tuples / 0.7;
	size_t bloom_filter_words = (bloom_filter_bits + 31) >> 5;
	uint32_t *unique_table = calloc(unique_buckets, sizeof(uint32_t));
	uint32_t *bloom_filter = calloc(bloom_filter_words, sizeof(uint32_t));
	// initialize synchronization barriers
	int b, barriers = 7;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	// names of methods
	char *names[] = {"Scalar", "Vector"};
	int m, methods = sizeof(names) / sizeof(char*);
	// initialize threads
	info_t info[threads];
	uint64_t times[methods * threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].inner_tuples = inner_tuples;
		info[t].outer_tuples = outer_tuples;
		info[t].filter_rate = filter_rate;
		info[t].log_bloom_filter_bits = log_bloom_filter_bits;
		info[t].hash_functions = hash_functions;
		info[t].hash_factors = hash_factors;
		info[t].bloom_filter = bloom_filter;
		info[t].unique_table = unique_table;
		info[t].seed = rand();
		info[t].threads = threads;
		info[t].thread = t;
		info[t].times = &times[methods * t];
		info[t].barrier = barrier;
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	// print rate
	uint64_t errors = 0;
	uint64_t passed = 0;
	for (t = 0 ; t != threads ; ++t) {
		errors += info[t].errors;
		passed += info[t].passed;
	}
	double fr = (passed - errors) * 100.0 / outer_tuples;
	double er = errors * 100.0 / (outer_tuples - (passed - errors));
	fprintf(stderr, "Measured filter rate: %.5f%%\n", fr);
	fprintf(stderr, "Measured error  rate: %.5f%%\n", er);
	fflush(stderr);
	// print times
	for (m = 0 ; m != methods ; ++m) {
		uint64_t ns = 0;
		for (t = 0 ; t != threads ; ++t)
			ns += info[t].times[m];
		ns /= threads;
		printf("%-6s, %6.3f\n", names[m], outer_tuples * 1.0 / ns);
	}
	// cleanup
	free(bloom_filter);
	free(unique_table);
	return EXIT_SUCCESS;
}
