/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <immintrin.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include <sched.h>
#include <time.h>

#include "inner_outer.h"


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *mamalloc(size_t size)
{
	void *ptr = NULL;
	return posix_memalign(&ptr, 64, size) ? NULL : ptr;
}

void *align(const void *p)
{
	while (63 & (size_t) p) p++;
	return (void*) p;
}

int log_2(size_t n)
{
	size_t b = 1;
	int p = 0;
	while (b < n) {
		b += b;
		p++;
	}
	assert(b == n);
	return p;
}

uint64_t sum(const uint32_t *data, size_t size)
{
	size_t i;
	uint64_t sum = 0;
	for (i = 0 ; i != size ; ++i)
		sum += data[i];
	return sum;
}

typedef struct {
	uint32_t key;
	uint32_t val;
} bucket_t;

size_t build(const uint32_t *keys, size_t size, bucket_t *table,
             int log_table_bytes, const uint32_t table_factors[2],
             uint32_t invalid_key, uint32_t payload_factor)
{
	size_t i;
	const size_t buckets = (((size_t) 1) << log_table_bytes) / sizeof(table[0]);
	const uint8_t shift = 32 - log_2(buckets);
	const uint32_t factor_1 = table_factors[0];
	const uint32_t factor_2 = table_factors[1];
	for (i = 0 ; i != buckets ; ++i)
		table[i].key = invalid_key;
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		uint32_t val = key * payload_factor;
		size_t h = ((uint32_t) (key * factor_1)) >> shift;
		uint32_t tab_key = table[h].key;
		if (tab_key != invalid_key) {
			h = ((uint32_t) (key * factor_2)) >> shift;
			tab_key = table[h].key;
		}
		size_t l = 0;
		while (tab_key != invalid_key) {
			uint32_t tab_val = table[h].val;
			table[h].key = key;
			table[h].val = val;
			key = tab_key;
			val = tab_val;
			size_t h1 = (uint32_t) (key * factor_1);
			size_t h2 = (uint32_t) (key * factor_2);
			h1 >>= shift;
			h2 >>= shift;
			h = h1 + h2 - h;
			if (++l == buckets) return i;
			tab_key = table[h].key;
		}
		table[h].key = key;
		table[h].val = val;
	}
	return i;
}

#ifdef __MIC
#define WIDTH 8
#else
#define WIDTH 2
#endif

typedef struct {
	uint32_t key[WIDTH];
	uint32_t val[WIDTH];
} bucket_wide_t;

size_t build_wide(const uint32_t *keys, size_t size, bucket_wide_t *table,
                  int log_table_bytes, const uint32_t table_factors[2],
                  uint32_t invalid_key, uint32_t payload_factor)
{
	size_t i, j;
	const size_t buckets = (((size_t) 1) << log_table_bytes) / sizeof(table[0]);
	const uint8_t shift = 32 - log_2(buckets);
	const uint32_t factor_1 = table_factors[0];
	const uint32_t factor_2 = table_factors[1];
	for (i = 0 ; i != buckets ; ++i)
		for (j = 0 ; j != WIDTH ; ++j)
			table[i].key[j] = invalid_key;
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		uint32_t val = key * payload_factor;
		size_t h = ((uint32_t) (key * factor_1)) >> shift;
		for (j = 0 ; j != WIDTH ; ++j)
			if (table[h].key[j] == invalid_key) break;
		if (j == WIDTH)
			h = ((uint32_t) (key * factor_2)) >> shift;
		size_t l = 0;
		for (;;) {
			for (j = 0 ; j != WIDTH ; ++j)
				if (table[h].key[j] == invalid_key) break;
			if (j != WIDTH) break;
			uint32_t tab_key = table[h].key[0];
			uint32_t tab_val = table[h].val[0];
			for (j = 0 ; j != WIDTH - 1 ; ++j) {
				table[h].key[j] = table[h].key[j + 1];
				table[h].val[j] = table[h].val[j + 1];
			}
			table[h].key[j] = key;
			table[h].val[j] = val;
			key = tab_key;
			val = tab_val;
			size_t h1 = ((uint32_t) (key * factor_1)) >> shift;
			size_t h2 = ((uint32_t) (key * factor_2)) >> shift;
			h = h1 + h2 - h;
			if (++l == buckets) return i;
		}
		table[h].key[j] = key;
		table[h].val[j] = val;
	}
	return i;
}

typedef struct {
	uint64_t sum;
	uint32_t count;
} avg_t;

size_t probe_sum(const uint32_t *keys, size_t size,
                 const bucket_t *table, int log_table_bytes,
                 const uint32_t table_factors[2],
                 uint64_t *payload_checksum)
{
	size_t i, j;
	uint64_t sum = 0;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const uint8_t shift = 32 - log_2(buckets);
	const uint32_t factor_1 = table_factors[0];
	const uint32_t factor_2 = table_factors[1];
	const uint64_t *table_64 = (const uint64_t*) table;
	for (i = j = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		size_t h = ((uint32_t) (key * factor_1)) >> shift;
		uint64_t tab = table_64[h];
		if (key == (uint32_t) tab) {
			sum += tab >> 32;
			j++;
		} else {
			h = ((uint32_t) (key * factor_2)) >> shift;
			tab = table_64[h];
			if (key == (uint32_t) tab) {
				sum += tab >> 32;
				j++;
			}
		}
	}
	*payload_checksum = sum;
	return j;
}

inline void store(uint32_t *p, uint32_t v)
{
#ifndef __MIC
	_mm_stream_si32((int*) p, v);
#else
	*p = v;
#endif
}

size_t probe(const uint32_t *keys, uint32_t *vals, size_t size,
             const bucket_t *table, int log_table_bytes,
             const uint32_t table_factors[2])
{
	size_t i;
	uint32_t *vals_orig = vals;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const uint8_t shift = 32 - log_2(buckets);
	const uint32_t factor_1 = table_factors[0];
	const uint32_t factor_2 = table_factors[1];
	const uint64_t *table_64 = (const uint64_t*) table;
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		size_t h = ((uint32_t) (key * factor_1)) >> shift;
		uint64_t tab = table_64[h];
		if (key == (uint32_t) tab)
			store(vals++, tab >> 32);
		else {
			h = ((uint32_t) (key * factor_2)) >> shift;
			tab = table_64[h];
			if (key == (uint32_t) tab)
				store(vals++, tab >> 32);
		}
	}
	return vals - vals_orig;;
}

size_t probe_nob(const uint32_t *keys, uint32_t *vals, size_t size,
                 const bucket_t *table, int log_table_bytes,
                 const uint32_t table_factors[2])
{
	size_t i, o;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const uint8_t shift = 32 - log_2(buckets);
	const uint32_t factor_1 = table_factors[0];
	const uint32_t factor_2 = table_factors[1];
	const uint64_t *table_64 = (const uint64_t*) table;
	for (i = o = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		size_t h1 = ((uint32_t) (key * factor_1)) >> shift;
		size_t h2 = ((uint32_t) (key * factor_2)) >> shift;
		uint64_t tab_1 = table_64[h1];
		uint64_t tab_2 = table_64[h2];
		uint64_t out_1 = -(key == (uint32_t) tab_1 ? 1 : 0);
		uint64_t out_2 = -(key == (uint32_t) tab_2 ? 1 : 0);
		uint64_t tab = (tab_1 & out_1) | (tab_2 & out_2);
		uint64_t out = out_1 | out_2;
		store(&vals[o], (tab >> 32) & out);
		o -= (size_t) out;
	}
	return o;
}

#ifndef __MIC

inline uint32_t _mm_reduce_or_epi32(__m128i x)
{
	x = _mm_or_si128(x, _mm_shuffle_epi32(x, _MM_SHUFFLE(1,0,3,2)));
	x = _mm_or_si128(x, _mm_shuffle_epi32(x, _MM_SHUFFLE(2,3,0,1)));
	return _mm_cvtsi128_si32(x);
}

size_t probe_hor(const uint32_t *keys, uint32_t *vals, size_t size,
                 const bucket_wide_t *table, int log_table_bytes,
                 const uint32_t table_factors[2])
{
	size_t i, j, o, p1 = 0, p2 = 0;
	assert(sizeof(table[0]) == 16);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 4);
	const __m128i shift = _mm_cvtsi32_si128(32 - log_2(buckets));
	const __m128i factor_1 = _mm_set1_epi32(table_factors[0]);
	const __m128i factor_2 = _mm_set1_epi32(table_factors[1]);
	const void *table_void = (const void*) table;
	for (i = o = 0 ; i != size ; i += 4) {
		__m128i key = _mm_load_si128((__m128i*) &keys[i]);
		__m128i h1 = _mm_mullo_epi32(key, factor_1);
		__m128i h2 = _mm_mullo_epi32(key, factor_2);
		h1 = _mm_srl_epi32(h1, shift);
		h2 = _mm_srl_epi32(h2, shift);
		h1 = _mm_slli_epi32(h1, 4);
		h2 = _mm_slli_epi32(h2, 4);
		for (j = 0 ; j != 4 ; ++j) {
			asm("movd	%1, %%eax" : "=a"(p1) : "x"(h1), "0"(p1));
			asm("movd	%1, %%edx" : "=d"(p2) : "x"(h2), "0"(p2));
			__m128i tab_1 = _mm_load_si128((__m128i*) (table_void + p1));
			__m128i tab_2 = _mm_load_si128((__m128i*) (table_void + p2));
			__m128i tab_key = _mm_unpacklo_epi64(tab_1, tab_2);
			__m128i tab_val = _mm_unpackhi_epi64(tab_1, tab_2);
			__m128i key_x4 = _mm_shuffle_epi32(key, 0);
			__m128i out = _mm_cmpeq_epi32(tab_key, key_x4);
			tab_val = _mm_and_si128(tab_val, out);
			uint32_t val = _mm_reduce_or_epi32(tab_val);
			_mm_stream_si32(&((int*) vals)[o], val);
			o += !_mm_testz_si128(out, out) ? 1 : 0;
			key = _mm_srli_si128(key, 4);
			h1 = _mm_srli_si128(h1, 4);
			h2 = _mm_srli_si128(h2, 4);
		}
	}
	return o;
}

#else

size_t probe_hor(const uint32_t *keys, uint32_t *vals, size_t size,
                 const bucket_wide_t *table, int log_table_bytes,
                 const uint32_t table_factors[2])
{
	size_t i, j, o;
	uint32_t buf[47];
	uint32_t *h1_buf = align(buf);
	uint32_t *h2_buf  = &h1_buf[16];
	assert(sizeof(table[0]) == 64);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 6);
	const __m512i shift = _mm512_set1_epi32(32 - log_2(buckets));
	const __m512i factor_1 = _mm512_set1_epi32(table_factors[0]);
	const __m512i factor_2 = _mm512_set1_epi32(table_factors[1]);
	const __mmask16 blend_00FF = _mm512_int2mask(0xFF00);
	const void *table_void = (const void*) table;
	for (i = o = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32(&keys[i]);
		__m512i h1 = _mm512_mullo_epi32(key, factor_1);
		__m512i h2 = _mm512_mullo_epi32(key, factor_2);
		h1 = _mm512_srlv_epi32(h1, shift);
		h2 = _mm512_srlv_epi32(h2, shift);
		h1 = _mm512_slli_epi32(h1, 6);
		h2 = _mm512_slli_epi32(h2, 6);
		_mm512_store_epi32(h1_buf, h1);
		_mm512_store_epi32(h2_buf, h2);
		for (j = 0 ; j != 16 ; ++j) {
			size_t p1 = h1_buf[j];
			size_t p2 = h2_buf[j];
			__m512i tab_1 = _mm512_load_epi32(table_void + p1);
			__m512i tab_2 = _mm512_load_epi32(table_void + p2);
			__m512i key_x16 = _mm512_extload_epi32(&keys[i + j], _MM_UPCONV_EPI32_NONE,
			                                       _MM_BROADCAST_1X16, _MM_HINT_NONE);
			__m512i tab_1_R = _mm512_permute4f128_epi32(tab_1, _MM_PERM_BADC);
			__m512i tab_2_R = _mm512_permute4f128_epi32(tab_2, _MM_PERM_BADC);
			__m512i tab_key = _mm512_mask_blend_epi32(blend_00FF, tab_1,   tab_2_R);
			__m512i tab_val = _mm512_mask_blend_epi32(blend_00FF, tab_1_R, tab_2);
			__mmask16 k = _mm512_cmpeq_epi32_mask(tab_key, key_x16);
			vals[o] = _mm512_mask_reduce_or_epi32(k, tab_val);
			o += !_mm512_kortestz(k, k) ? 1 : 0;
		}
	}
	return o;
}

#endif

#ifndef __MIC

inline __m256i _mm256_packlo_epi32(__m256i x, __m256i y)
{
	__m256 a = _mm256_castsi256_ps(x);
	__m256 b = _mm256_castsi256_ps(y);
	__m256 c = _mm256_shuffle_ps(a, b, _MM_SHUFFLE(2,0,2,0));
	__m256i z = _mm256_castps_si256(c);
	return _mm256_permute4x64_epi64(z, _MM_SHUFFLE(3,1,2,0));
}

inline __m256i _mm256_packhi_epi32(__m256i x, __m256i y)
{
	__m256 a = _mm256_castsi256_ps(x);
	__m256 b = _mm256_castsi256_ps(y);
	__m256 c = _mm256_shuffle_ps(a, b, _MM_SHUFFLE(3,1,3,1));
	__m256i z = _mm256_castps_si256(c);
	return _mm256_permute4x64_epi64(z, _MM_SHUFFLE(3,1,2,0));
}

const uint64_t perm[256] = {0x0001020304050607ull,
     0x0102030405060700ull, 0x0002030405060701ull, 0x0203040506070100ull,
     0x0001030405060702ull, 0x0103040506070200ull, 0x0003040506070201ull,
     0x0304050607020100ull, 0x0001020405060703ull, 0x0102040506070300ull,
     0x0002040506070301ull, 0x0204050607030100ull, 0x0001040506070302ull,
     0x0104050607030200ull, 0x0004050607030201ull, 0x0405060703020100ull,
     0x0001020305060704ull, 0x0102030506070400ull, 0x0002030506070401ull,
     0x0203050607040100ull, 0x0001030506070402ull, 0x0103050607040200ull,
     0x0003050607040201ull, 0x0305060704020100ull, 0x0001020506070403ull,
     0x0102050607040300ull, 0x0002050607040301ull, 0x0205060704030100ull,
     0x0001050607040302ull, 0x0105060704030200ull, 0x0005060704030201ull,
     0x0506070403020100ull, 0x0001020304060705ull, 0x0102030406070500ull,
     0x0002030406070501ull, 0x0203040607050100ull, 0x0001030406070502ull,
     0x0103040607050200ull, 0x0003040607050201ull, 0x0304060705020100ull,
     0x0001020406070503ull, 0x0102040607050300ull, 0x0002040607050301ull,
     0x0204060705030100ull, 0x0001040607050302ull, 0x0104060705030200ull,
     0x0004060705030201ull, 0x0406070503020100ull, 0x0001020306070504ull,
     0x0102030607050400ull, 0x0002030607050401ull, 0x0203060705040100ull,
     0x0001030607050402ull, 0x0103060705040200ull, 0x0003060705040201ull,
     0x0306070504020100ull, 0x0001020607050403ull, 0x0102060705040300ull,
     0x0002060705040301ull, 0x0206070504030100ull, 0x0001060705040302ull,
     0x0106070504030200ull, 0x0006070504030201ull, 0x0607050403020100ull,
     0x0001020304050706ull, 0x0102030405070600ull, 0x0002030405070601ull,
     0x0203040507060100ull, 0x0001030405070602ull, 0x0103040507060200ull,
     0x0003040507060201ull, 0x0304050706020100ull, 0x0001020405070603ull,
     0x0102040507060300ull, 0x0002040507060301ull, 0x0204050706030100ull,
     0x0001040507060302ull, 0x0104050706030200ull, 0x0004050706030201ull,
     0x0405070603020100ull, 0x0001020305070604ull, 0x0102030507060400ull,
     0x0002030507060401ull, 0x0203050706040100ull, 0x0001030507060402ull,
     0x0103050706040200ull, 0x0003050706040201ull, 0x0305070604020100ull,
     0x0001020507060403ull, 0x0102050706040300ull, 0x0002050706040301ull,
     0x0205070604030100ull, 0x0001050706040302ull, 0x0105070604030200ull,
     0x0005070604030201ull, 0x0507060403020100ull, 0x0001020304070605ull,
     0x0102030407060500ull, 0x0002030407060501ull, 0x0203040706050100ull,
     0x0001030407060502ull, 0x0103040706050200ull, 0x0003040706050201ull,
     0x0304070605020100ull, 0x0001020407060503ull, 0x0102040706050300ull,
     0x0002040706050301ull, 0x0204070605030100ull, 0x0001040706050302ull,
     0x0104070605030200ull, 0x0004070605030201ull, 0x0407060503020100ull,
     0x0001020307060504ull, 0x0102030706050400ull, 0x0002030706050401ull,
     0x0203070605040100ull, 0x0001030706050402ull, 0x0103070605040200ull,
     0x0003070605040201ull, 0x0307060504020100ull, 0x0001020706050403ull,
     0x0102070605040300ull, 0x0002070605040301ull, 0x0207060504030100ull,
     0x0001070605040302ull, 0x0107060504030200ull, 0x0007060504030201ull,
     0x0706050403020100ull, 0x0001020304050607ull, 0x0102030405060700ull,
     0x0002030405060701ull, 0x0203040506070100ull, 0x0001030405060702ull,
     0x0103040506070200ull, 0x0003040506070201ull, 0x0304050607020100ull,
     0x0001020405060703ull, 0x0102040506070300ull, 0x0002040506070301ull,
     0x0204050607030100ull, 0x0001040506070302ull, 0x0104050607030200ull,
     0x0004050607030201ull, 0x0405060703020100ull, 0x0001020305060704ull,
     0x0102030506070400ull, 0x0002030506070401ull, 0x0203050607040100ull,
     0x0001030506070402ull, 0x0103050607040200ull, 0x0003050607040201ull,
     0x0305060704020100ull, 0x0001020506070403ull, 0x0102050607040300ull,
     0x0002050607040301ull, 0x0205060704030100ull, 0x0001050607040302ull,
     0x0105060704030200ull, 0x0005060704030201ull, 0x0506070403020100ull,
     0x0001020304060705ull, 0x0102030406070500ull, 0x0002030406070501ull,
     0x0203040607050100ull, 0x0001030406070502ull, 0x0103040607050200ull,
     0x0003040607050201ull, 0x0304060705020100ull, 0x0001020406070503ull,
     0x0102040607050300ull, 0x0002040607050301ull, 0x0204060705030100ull,
     0x0001040607050302ull, 0x0104060705030200ull, 0x0004060705030201ull,
     0x0406070503020100ull, 0x0001020306070504ull, 0x0102030607050400ull,
     0x0002030607050401ull, 0x0203060705040100ull, 0x0001030607050402ull,
     0x0103060705040200ull, 0x0003060705040201ull, 0x0306070504020100ull,
     0x0001020607050403ull, 0x0102060705040300ull, 0x0002060705040301ull,
     0x0206070504030100ull, 0x0001060705040302ull, 0x0106070504030200ull,
     0x0006070504030201ull, 0x0607050403020100ull, 0x0001020304050706ull,
     0x0102030405070600ull, 0x0002030405070601ull, 0x0203040507060100ull,
     0x0001030405070602ull, 0x0103040507060200ull, 0x0003040507060201ull,
     0x0304050706020100ull, 0x0001020405070603ull, 0x0102040507060300ull,
     0x0002040507060301ull, 0x0204050706030100ull, 0x0001040507060302ull,
     0x0104050706030200ull, 0x0004050706030201ull, 0x0405070603020100ull,
     0x0001020305070604ull, 0x0102030507060400ull, 0x0002030507060401ull,
     0x0203050706040100ull, 0x0001030507060402ull, 0x0103050706040200ull,
     0x0003050706040201ull, 0x0305070604020100ull, 0x0001020507060403ull,
     0x0102050706040300ull, 0x0002050706040301ull, 0x0205070604030100ull,
     0x0001050706040302ull, 0x0105070604030200ull, 0x0005070604030201ull,
     0x0507060403020100ull, 0x0001020304070605ull, 0x0102030407060500ull,
     0x0002030407060501ull, 0x0203040706050100ull, 0x0001030407060502ull,
     0x0103040706050200ull, 0x0003040706050201ull, 0x0304070605020100ull,
     0x0001020407060503ull, 0x0102040706050300ull, 0x0002040706050301ull,
     0x0204070605030100ull, 0x0001040706050302ull, 0x0104070605030200ull,
     0x0004070605030201ull, 0x0407060503020100ull, 0x0001020307060504ull,
     0x0102030706050400ull, 0x0002030706050401ull, 0x0203070605040100ull,
     0x0001030706050402ull, 0x0103070605040200ull, 0x0003070605040201ull,
     0x0307060504020100ull, 0x0001020706050403ull, 0x0102070605040300ull,
     0x0002070605040301ull, 0x0207060504030100ull, 0x0001070605040302ull,
     0x0107060504030200ull, 0x0007060504030201ull, 0x0706050403020100ull};

size_t probe_ver(const uint32_t *keys, uint32_t *vals, size_t size,
                 const bucket_t *table, int log_table_bytes,
                 const uint32_t table_factors[2])
{
	size_t i, o, b;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const __m128i shift = _mm_cvtsi32_si128(32 - log_2(buckets));
	const __m256i factor_1 = _mm256_set1_epi32(table_factors[0]);
	const __m256i factor_2 = _mm256_set1_epi32(table_factors[1]);
#if defined __INTEL_COMPILER && __INTEL_COMPILER < 1600
	const long *table_64 = (const long*) table;
#else
	const long long *table_64 = (const long long*) table;
#endif
	const size_t buf_size = 128;
	int buf_space[buf_size + 8 + 15];
	int *buf = align((void*) buf_space);
	for (i = o = b = 0 ; i != size ; i += 8) {
		__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
		// hash
		__m256i h1 = _mm256_mullo_epi32(key, factor_1);
		__m256i h2 = _mm256_mullo_epi32(key, factor_2);
		h1 = _mm256_srl_epi32(h1, shift);
		h2 = _mm256_srl_epi32(h2, shift);
		// gather all
		__m256i tab_1_L = _mm256_i32gather_epi64(table_64, _mm256_castsi256_si128(h1), 8);
		__m256i tab_2_L = _mm256_i32gather_epi64(table_64, _mm256_castsi256_si128(h2), 8);
		h1 = _mm256_permute4x64_epi64(h1, _MM_SHUFFLE(1,0,3,2));
		h2 = _mm256_permute4x64_epi64(h2, _MM_SHUFFLE(1,0,3,2));
		__m256i tab_1_H = _mm256_i32gather_epi64(table_64, _mm256_castsi256_si128(h1), 8);
		__m256i tab_2_H = _mm256_i32gather_epi64(table_64, _mm256_castsi256_si128(h2), 8);
		// mix keys and payloads
		__m256i key_1 = _mm256_packlo_epi32(tab_1_L, tab_1_H);
		__m256i key_2 = _mm256_packlo_epi32(tab_2_L, tab_2_H);
		__m256i val_1 = _mm256_packhi_epi32(tab_1_L, tab_1_H);
		__m256i val_2 = _mm256_packhi_epi32(tab_2_L, tab_2_H);
		// matches
		__m256i out_1 = _mm256_cmpeq_epi32(key_1, key);
		__m256i out_2 = _mm256_cmpeq_epi32(key_2, key);
		val_1 = _mm256_and_si256(val_1, out_1);
		val_2 = _mm256_and_si256(val_2, out_2);
		__m256i out = _mm256_or_si256(out_1, out_2);
		__m256i tab = _mm256_or_si256(val_1, val_2);
		// store output
		size_t j = _mm256_movemask_ps(_mm256_castsi256_ps(out));
		__m128i perm_comp = _mm_loadl_epi64((__m128i*) &perm[j]);
		__m256i perm = _mm256_cvtepi8_epi32(perm_comp);
		tab = _mm256_permutevar8x32_epi32(tab, perm);
		out = _mm256_permutevar8x32_epi32(out, perm);
		_mm256_maskstore_epi32(&buf[b], out, tab);
		b += _mm_popcnt_u64(j);
		// flush buffer
		if (b > buf_size) {
			size_t b_i = 0;
			do {
				__m256i x = _mm256_load_si256((__m256i*) &buf[b_i]);
				_mm256_stream_si256((__m256i*) &vals[o], x);
				b_i += 8;
				o += 8;
			} while (b_i != buf_size);
			__m256i x = _mm256_load_si256((__m256i*) &buf[b_i]);
			_mm256_store_si256((__m256i*) &buf[0], x);
			b -= buf_size;
		}
	}
	// clean buffer
	size_t b_i = 0;
	while (b_i != b) _mm_stream_si32(&((int*) vals)[o++], buf[b_i++]);
	return o;
}

size_t probe_ver_sel(const uint32_t *keys, uint32_t *vals, size_t size,
                     const bucket_t *table, int log_table_bytes,
                     const uint32_t table_factors[2])
{
	size_t i, o, b;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const __m128i shift = _mm_cvtsi32_si128(32 - log_2(buckets));
	const __m256i factor_1 = _mm256_set1_epi32(table_factors[0]);
	const __m256i factor_2 = _mm256_set1_epi32(table_factors[1]);
	const __m256i perm_extend_lo = _mm256_setr_epi32(0,0,1,1,2,2,3,3);
	const __m256i perm_extend_hi = _mm256_setr_epi32(4,4,5,5,6,6,7,7);
	const __m256i all_set = _mm256_set1_epi32(-1);
#if defined __INTEL_COMPILER && __INTEL_COMPILER < 1600
	const long *table_64 = (const long*) table;
#else
	const long long *table_64 = (const long long*) table;
#endif
	const size_t buf_size = 128;
	int buf_space[buf_size + 8 + 15];
	int *buf = align((void*) buf_space);
	for (i = o = b = 0 ; i != size ; i += 8) {
		__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
		// hash
		__m256i h1 = _mm256_mullo_epi32(key, factor_1);
		__m256i h2 = _mm256_mullo_epi32(key, factor_2);
		h1 = _mm256_srl_epi32(h1, shift);
		h2 = _mm256_srl_epi32(h2, shift);
		// gather 1st
		__m256i tab_L = _mm256_i32gather_epi64(table_64, _mm256_castsi256_si128(h1), 8);
		h1 = _mm256_permute4x64_epi64(h1, _MM_SHUFFLE(1,0,3,2));
		__m256i tab_H = _mm256_i32gather_epi64(table_64, _mm256_castsi256_si128(h1), 8);
		// match keys
		__m256i tab = _mm256_packlo_epi32(tab_L, tab_H);
		__m256i out = _mm256_cmpeq_epi32(tab, key);
		__m256i alt = _mm256_xor_si256(out, all_set);
		__m256i alt_L = _mm256_permutevar8x32_epi32(alt, perm_extend_lo);
		__m256i alt_H = _mm256_permutevar8x32_epi32(alt, perm_extend_hi);
		// gather 2nd
		tab_L = _mm256_mask_i32gather_epi64(tab_L, table_64, _mm256_castsi256_si128(h2), alt_L, 8);
		h2 = _mm256_permute4x64_epi64(h2, _MM_SHUFFLE(1,0,3,2));
		tab_H = _mm256_mask_i32gather_epi64(tab_H, table_64, _mm256_castsi256_si128(h2), alt_H, 8);
		// match keys
		tab = _mm256_packlo_epi32(tab_L, tab_H);
		out = _mm256_cmpeq_epi32(tab, key);
		tab = _mm256_packhi_epi32(tab_L, tab_H);
		// store output
		size_t j = _mm256_movemask_ps(_mm256_castsi256_ps(out));
		__m128i perm_comp = _mm_loadl_epi64((__m128i*) &perm[j]);
		__m256i perm = _mm256_cvtepi8_epi32(perm_comp);
		tab = _mm256_permutevar8x32_epi32(tab, perm);
		out = _mm256_permutevar8x32_epi32(out, perm);
		_mm256_maskstore_epi32(&buf[b], out, tab);
		b += _mm_popcnt_u64(j);
		// flush buffer
		if (b > buf_size) {
			size_t b_i = 0;
			do {
				__m256i x = _mm256_load_si256((__m256i*) &buf[b_i]);
				_mm256_stream_si256((__m256i*) &vals[o], x);
				b_i += 8;
				o += 8;
			} while (b_i != buf_size);
			__m256i x = _mm256_load_si256((__m256i*) &buf[b_i]);
			_mm256_store_si256((__m256i*) &buf[0], x);
			b -= buf_size;
		}
	}
	// clean buffer
	size_t b_i = 0;
	while (b_i != b) _mm_stream_si32(&((int*) vals)[o++], buf[b_i++]);
	return o;
}

size_t probe_ver_sep(const uint32_t *keys, uint32_t *vals, size_t size,
                     const bucket_t *table, int log_table_bytes,
                     const uint32_t table_factors[2])
{
	size_t i, o, b;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const __m128i shift = _mm_cvtsi32_si128(32 - log_2(buckets));
	const __m256i factor_1 = _mm256_set1_epi32(table_factors[0]);
	const __m256i factor_2 = _mm256_set1_epi32(table_factors[1]);
	const __m256i all_set = _mm256_set1_epi32(-1);
	const int *key_table = (const int*) table;
	const int *val_table = &key_table[1];
	const size_t buf_size = 128;
	int buf_space[buf_size + 8 + 15];
	int *buf = align((void*) buf_space);
	for (i = o = b = 0 ; i != size ; i += 8) {
		__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
		// hash
		__m256i h1 = _mm256_mullo_epi32(key, factor_1);
		__m256i h2 = _mm256_mullo_epi32(key, factor_2);
		h1 = _mm256_srl_epi32(h1, shift);
		h2 = _mm256_srl_epi32(h2, shift);
		// gather 1st key and find non-matches
		__m256i tab = _mm256_i32gather_epi32(key_table, h1, 8);
		__m256i fst = _mm256_cmpeq_epi32(tab, key);
		__m256i snd = _mm256_xor_si256(fst, all_set);
		// gather 2nd key and find matches
		tab = _mm256_mask_i32gather_epi32(tab, key_table, h2, snd, 8);
		__m256i h = _mm256_blendv_epi8(h1, h2, snd);
		__m256i out = _mm256_cmpeq_epi32(tab, key);
		// gather payloads for matches
		tab = _mm256_mask_i32gather_epi32(tab, val_table, h, out, 8);
		// store output
		size_t j = _mm256_movemask_ps(_mm256_castsi256_ps(out));
		__m128i perm_comp = _mm_loadl_epi64((__m128i*) &perm[j]);
		__m256i perm = _mm256_cvtepi8_epi32(perm_comp);
		tab = _mm256_permutevar8x32_epi32(tab, perm);
		out = _mm256_permutevar8x32_epi32(out, perm);
		_mm256_maskstore_epi32(&buf[b], out, tab);
		b += _mm_popcnt_u64(j);
		// flush buffer
		if (b > buf_size) {
			size_t b_i = 0;
			do {
				__m256i x = _mm256_load_si256((__m256i*) &buf[b_i]);
				_mm256_stream_si256((__m256i*) &vals[o], x);
				b_i += 8;
				o += 8;
			} while (b_i != buf_size);
			__m256i x = _mm256_load_si256((__m256i*) &buf[b_i]);
			_mm256_store_si256((__m256i*) &buf[0], x);
			b -= buf_size;
		}
	}
	// clean buffer
	size_t b_i = 0;
	while (b_i != b) _mm_stream_si32(&((int*) vals)[o++], buf[b_i++]);
	return o;
}

inline __m256i _mm256_i32gather_epi64_emu(const uint64_t *table, __m128i index)
{
	__m128i index_R = _mm_shuffle_epi32(index, _MM_SHUFFLE(1,0,3,2));
	__m128i i12 = _mm_cvtepi32_epi64(index);
	__m128i i34 = _mm_cvtepi32_epi64(index_R);
	size_t i1 = _mm_cvtsi128_si64(i12);
	size_t i3 = _mm_cvtsi128_si64(i34);
	__m128i d1 = _mm_loadl_epi64((__m128i*) &table[i1]);
	__m128i d3 = _mm_loadl_epi64((__m128i*) &table[i3]);
	i12 = _mm_srli_si128(i12, 8);
	i34 = _mm_srli_si128(i34, 8);
	size_t i2 = _mm_cvtsi128_si64(i12);
	size_t i4 = _mm_cvtsi128_si64(i34);
	__m128i d2 = _mm_loadl_epi64((__m128i*) &table[i2]);
	__m128i d4 = _mm_loadl_epi64((__m128i*) &table[i4]);
	__m256i d12 = _mm256_castsi128_si256(_mm_unpacklo_epi64(d1, d2));
	__m256i d34 = _mm256_castsi128_si256(_mm_unpacklo_epi64(d3, d4));
	return _mm256_permute2x128_si256(d12, d34, _MM_SHUFFLE(0,2,0,0));
}

size_t probe_ver_emu(const uint32_t *keys, uint32_t *vals, size_t size,
                     const bucket_t *table, int log_table_bytes,
                     const uint32_t table_factors[2])
{
	size_t i, o, b;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const __m128i shift = _mm_cvtsi32_si128(32 - log_2(buckets));
	const __m256i factor_1 = _mm256_set1_epi32(table_factors[0]);
	const __m256i factor_2 = _mm256_set1_epi32(table_factors[1]);
	const uint64_t *table_64 = (const uint64_t*) table;
	const size_t buf_size = 128;
	int buf_space[buf_size + 8 + 15];
	int *buf = align((void*) buf_space);
	for (i = o = b = 0 ; i != size ; i += 8) {
		__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
		// hash
		__m256i h1 = _mm256_mullo_epi32(key, factor_1);
		__m256i h2 = _mm256_mullo_epi32(key, factor_2);
		h1 = _mm256_srl_epi32(h1, shift);
		h2 = _mm256_srl_epi32(h2, shift);
		// gather all
		__m256i tab_1_L = _mm256_i32gather_epi64_emu(table_64, _mm256_castsi256_si128(h1));
		__m256i tab_2_L = _mm256_i32gather_epi64_emu(table_64, _mm256_castsi256_si128(h2));
		h1 = _mm256_permute4x64_epi64(h1, _MM_SHUFFLE(1,0,3,2));
		h2 = _mm256_permute4x64_epi64(h2, _MM_SHUFFLE(1,0,3,2));
		__m256i tab_1_H = _mm256_i32gather_epi64_emu(table_64, _mm256_castsi256_si128(h1));
		__m256i tab_2_H = _mm256_i32gather_epi64_emu(table_64, _mm256_castsi256_si128(h2));
		// mix keys and payloads
		__m256i key_1 = _mm256_packlo_epi32(tab_1_L, tab_1_H);
		__m256i key_2 = _mm256_packlo_epi32(tab_2_L, tab_2_H);
		__m256i val_1 = _mm256_packhi_epi32(tab_1_L, tab_1_H);
		__m256i val_2 = _mm256_packhi_epi32(tab_2_L, tab_2_H);
		// matches
		__m256i out_1 = _mm256_cmpeq_epi32(key_1, key);
		__m256i out_2 = _mm256_cmpeq_epi32(key_2, key);
		val_1 = _mm256_and_si256(val_1, out_1);
		val_2 = _mm256_and_si256(val_2, out_2);
		__m256i out = _mm256_or_si256(out_1, out_2);
		__m256i tab = _mm256_or_si256(val_1, val_2);
		// store output
		size_t j = _mm256_movemask_ps(_mm256_castsi256_ps(out));
		__m128i perm_comp = _mm_loadl_epi64((__m128i*) &perm[j]);
		__m256i perm = _mm256_cvtepi8_epi32(perm_comp);
		tab = _mm256_permutevar8x32_epi32(tab, perm);
		out = _mm256_permutevar8x32_epi32(out, perm);
		_mm256_maskstore_epi32(&buf[b], out, tab);
		b += _mm_popcnt_u64(j);
		// flush buffer
		if (b > buf_size) {
			size_t b_i = 0;
			do {
				__m256i x = _mm256_load_si256((__m256i*) &buf[b_i]);
				_mm256_stream_si256((__m256i*) &vals[o], x);
				b_i += 8;
				o += 8;
			} while (b_i != buf_size);
			__m256i x = _mm256_load_si256((__m256i*) &buf[b_i]);
			_mm256_store_si256((__m256i*) &buf[0], x);
			b -= buf_size;
		}
	}
	// clean buffer
	size_t b_i = 0;
	while (b_i != b) _mm_stream_si32(&((int*) vals)[o++], buf[b_i++]);
	return o;
}

#else

size_t probe_ver(const uint32_t *keys, uint32_t *vals, size_t size,
                 const bucket_t *table, int log_table_bytes,
                 const uint32_t table_factors[2])
{
	size_t i, o, b;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const __m512i shift = _mm512_set1_epi32(32 - log_2(buckets));
	const __m512i factor_1 = _mm512_set1_epi32(table_factors[0]);
	const __m512i factor_2 = _mm512_set1_epi32(table_factors[1]);
	const __m512i unpack = _mm512_set_epi32(15,13,11,9,7,5,3,1,14,12,10,8,6,4,2,0);
	const __mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	const __mmask16 blend_5555 = _mm512_int2mask(0x5555);
	const __mmask16 z = _mm512_int2mask(0);
	const size_t buf_size = 128;
	int buf_space[buf_size + 16 + 15];
	int *buf = align((void*) buf_space);
	for (i = o = b = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32((__m256i*) &keys[i]);
		// hash
		__m512i h1 = _mm512_mullo_epi32(key, factor_1);
		__m512i h2 = _mm512_mullo_epi32(key, factor_2);
		h1 = _mm512_srlv_epi32(h1, shift);
		h2 = _mm512_srlv_epi32(h2, shift);
		// gather all
		__m512i tab_1_L = _mm512_i32logather_epi64(h1, table, 8);
		__m512i tab_2_L = _mm512_i32logather_epi64(h2, table, 8);
		h1 = _mm512_permute4f128_epi32(h1, _MM_PERM_BADC);
		h2 = _mm512_permute4f128_epi32(h2, _MM_PERM_BADC);
		__m512i tab_1_H = _mm512_i32logather_epi64(h1, table, 8);
		__m512i tab_2_H = _mm512_i32logather_epi64(h2, table, 8);
		// mix keys and payloads
		__m512i key_1 = _mm512_mask_blend_epi32(blend_AAAA, tab_1_L, _mm512_swizzle_epi32(tab_1_H, _MM_SWIZ_REG_CDAB));
		__m512i key_2 = _mm512_mask_blend_epi32(blend_AAAA, tab_2_L, _mm512_swizzle_epi32(tab_2_H, _MM_SWIZ_REG_CDAB));
		__m512i val_1 = _mm512_mask_blend_epi32(blend_5555, tab_1_H, _mm512_swizzle_epi32(tab_1_L, _MM_SWIZ_REG_CDAB));
		__m512i val_2 = _mm512_mask_blend_epi32(blend_5555, tab_2_H, _mm512_swizzle_epi32(tab_2_L, _MM_SWIZ_REG_CDAB));
		key_1 = _mm512_permutevar_epi32(unpack, key_1);
		key_2 = _mm512_permutevar_epi32(unpack, key_2);
		val_1 = _mm512_permutevar_epi32(unpack, val_1);
		val_2 = _mm512_permutevar_epi32(unpack, val_2);
		// matches
		__mmask16 k1 = _mm512_cmpeq_epi32_mask(key_1, key);
		__mmask16 k2 = _mm512_cmpeq_epi32_mask(key_2, key);
		__m512i val = _mm512_mask_blend_epi32(k1, val_2, val_1);
		__mmask16 k = _mm512_kor(k1, k2);
		// store to output
		_mm512_mask_packstorelo_epi32(&buf[b +  0], k, val);
		_mm512_mask_packstorehi_epi32(&buf[b + 16], k, val);
		b += _mm_countbits_64(_mm512_kconcatlo_64(z, k));
		// flush buffer
		if (b > buf_size) {
			b -= buf_size;
			size_t b_i = 0;
			do {
				__m512 x = _mm512_load_ps(&buf[b_i]);
				_mm512_storenrngo_ps(&vals[o], x);
				b_i += 16;
				o += 16;
			} while (b_i != buf_size);
			__m512 x = _mm512_load_ps(&buf[b_i]);
			_mm512_store_ps(&buf[0], x);
		}
	}
	// clean buffer
	size_t b_i = 0;
	while (b_i != b) vals[o++] = buf[b_i++];
	return o;
}

size_t probe_ver_sel(const uint32_t *keys, uint32_t *vals, size_t size,
                     const bucket_t *table, int log_table_bytes,
                     const uint32_t table_factors[2])
{
	size_t i, o, b;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const __m512i shift = _mm512_set1_epi32(32 - log_2(buckets));
	const __m512i factor_1 = _mm512_set1_epi32(table_factors[0]);
	const __m512i factor_2 = _mm512_set1_epi32(table_factors[1]);
	const __m512i mask_unpack = _mm512_set_epi32(15,13,11,9,7,5,3,1,14,12,10,8,6,4,2,0);
	const __mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	const __mmask16 blend_5555 = _mm512_int2mask(0x5555);
	const __mmask16 z = _mm512_int2mask(0);
	const size_t buf_size = 128;
	int buf_space[buf_size + 16 + 15];
	int *buf = align((void*) buf_space);
	for (i = o = b = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32((__m512i*) &keys[i]);
		// hash
		__m512i h1 = _mm512_mullo_epi32(key, factor_1);
		__m512i h2 = _mm512_mullo_epi32(key, factor_2);
		h1 = _mm512_srlv_epi32(h1, shift);
		h2 = _mm512_srlv_epi32(h2, shift);
		// gather 1st
		__m512i tab_L = _mm512_i32logather_epi64(h1, table, 8);
		h1 = _mm512_permute4f128_epi32(h1, _MM_PERM_BADC);
		__m512i tab_H = _mm512_i32logather_epi64(h1, table, 8);
		// match keys
		__m512i tab_key_1 = _mm512_mask_blend_epi32(blend_AAAA, tab_L, _mm512_swizzle_epi32(tab_H, _MM_SWIZ_REG_CDAB));
		tab_key_1 = _mm512_permutevar_epi32(mask_unpack, tab_key_1);
		__mmask16 k = _mm512_cmpneq_epi32_mask(tab_key_1, key);
		// gather 2nd
		tab_L = _mm512_mask_i32logather_epi64(tab_L, k, h2, table, 8);
		h2 = _mm512_permute4f128_epi32(h2, _MM_PERM_BADC);
		k = _mm512_kmerge2l1h(k, k);
		tab_H = _mm512_mask_i32logather_epi64(tab_H, k, h2, table, 8);
		// match keys
		__m512i tab_key = _mm512_mask_blend_epi32(blend_AAAA, tab_L, _mm512_swizzle_epi32(tab_H, _MM_SWIZ_REG_CDAB));
		__m512i tab_val = _mm512_mask_blend_epi32(blend_5555, tab_H, _mm512_swizzle_epi32(tab_L, _MM_SWIZ_REG_CDAB));
		tab_key = _mm512_permutevar_epi32(mask_unpack, tab_key);
		tab_val = _mm512_permutevar_epi32(mask_unpack, tab_val);
		k = _mm512_cmpeq_epi32_mask(tab_key, key);
		// store to output
		_mm512_mask_packstorelo_epi32(&buf[b +  0], k, tab_val);
		_mm512_mask_packstorehi_epi32(&buf[b + 16], k, tab_val);
		b += _mm_countbits_64(_mm512_kconcatlo_64(z, k));
		// flush buffer
		if (b > buf_size) {
			b -= buf_size;
			size_t b_i = 0;
			do {
				__m512 x = _mm512_load_ps(&buf[b_i]);
				_mm512_storenrngo_ps(&vals[o], x);
				b_i += 16;
				o += 16;
			} while (b_i != buf_size);
			__m512 x = _mm512_load_ps(&buf[b_i]);
			_mm512_store_ps(&buf[0], x);
		}
	}
	// clean buffer
	size_t b_i = 0;
	while (b_i != b) vals[o++] = buf[b_i++];
	return o;
}

size_t probe_ver_sep(const uint32_t *keys, uint32_t *vals, size_t size,
                     const bucket_t *table, int log_table_bytes,
                     const uint32_t table_factors[2])
{
	size_t i, o, b;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const __m512i shift = _mm512_set1_epi32(32 - log_2(buckets));
	const __m512i factor_1 = _mm512_set1_epi32(table_factors[0]);
	const __m512i factor_2 = _mm512_set1_epi32(table_factors[1]);
	const __m512i mask_unpack = _mm512_set_epi32(15,13,11,9,7,5,3,1,14,12,10,8,6,4,2,0);
	const __mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	const __mmask16 blend_5555 = _mm512_int2mask(0x5555);
	const __mmask16 z = _mm512_int2mask(0);
	const int *key_table = (const int*) table;
	const int *val_table = &key_table[1];
	const size_t buf_size = 128;
	int buf_space[buf_size + 16 + 15];
	int *buf = align((void*) buf_space);
	for (i = o = b = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32((__m512i*) &keys[i]);
		// hash
		__m512i h1 = _mm512_mullo_epi32(key, factor_1);
		__m512i h2 = _mm512_mullo_epi32(key, factor_2);
		h1 = _mm512_srlv_epi32(h1, shift);
		h2 = _mm512_srlv_epi32(h2, shift);
		// gather 1st keys and check non-equal
		__m512i tab = _mm512_i32gather_epi32(h1, key_table, 8);
		__mmask16 k = _mm512_cmpneq_epi32_mask(tab, key);
		// gather 2nd keys and check equal
		tab = _mm512_mask_i32gather_epi32(tab, k, h2, key_table, 8);
		__m512i h = _mm512_mask_blend_epi32(k, h1, h2);
		k = _mm512_cmpeq_epi32_mask(tab, key);
		// gather payloads for matching keys only
		tab = _mm512_mask_i32gather_epi32(tab, k, h, val_table, 8);
		// store to output
		_mm512_mask_packstorelo_epi32(&buf[b +  0], k, tab);
		_mm512_mask_packstorehi_epi32(&buf[b + 16], k, tab);
		b += _mm_countbits_64(_mm512_kconcatlo_64(z, k));
		// flush buffer
		if (b > buf_size) {
			b -= buf_size;
			size_t b_i = 0;
			do {
				__m512 x = _mm512_load_ps(&buf[b_i]);
				_mm512_storenrngo_ps(&vals[o], x);
				b_i += 16;
				o += 16;
			} while (b_i != buf_size);
			__m512 x = _mm512_load_ps(&buf[b_i]);
			_mm512_store_ps(&buf[0], x);
		}
	}
	// clean buffer
	size_t b_i = 0;
	while (b_i != b) vals[o++] = buf[b_i++];
	return o;
}

size_t probe_ver_emu(const uint32_t *keys, uint32_t *vals, size_t size,
                     const bucket_t *table, int log_table_bytes,
                     const uint32_t table_factors[2])
{
	size_t e, i, o, b;
	assert(sizeof(table[0]) == 8);
	const size_t buckets = ((size_t) 1) << (log_table_bytes - 3);
	const __m512i shift = _mm512_set1_epi32(32 - log_2(buckets));
	const __m512i factor_1 = _mm512_set1_epi32(table_factors[0]);
	const __m512i factor_2 = _mm512_set1_epi32(table_factors[1]);
	const __m512i unpack = _mm512_set_epi32(15,13,11,9,7,5,3,1,14,12,10,8,6,4,2,0);
	const __mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	const __mmask16 blend_5555 = _mm512_int2mask(0x5555);
	const __mmask16 z = _mm512_int2mask(0);
	const size_t buf_size = 128;
	int buf_space[buf_size + 16 + 15];
	int *buf = align((void*) buf_space);
	const uint64_t *table_64 = (const uint64_t*) table;
	uint32_t index_space[16 + 15];
	uint32_t *index = align((void*) index_space);
	uint64_t data_space[16 + 7];
	uint64_t *data = align((void*) data_space);
	for (i = o = b = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32((__m256i*) &keys[i]);
		// hash
		__m512i h1 = _mm512_mullo_epi32(key, factor_1);
		__m512i h2 = _mm512_mullo_epi32(key, factor_2);
		h1 = _mm512_srlv_epi32(h1, shift);
		h2 = _mm512_srlv_epi32(h2, shift);
		// gather all
		_mm512_store_epi32(index, h1);
		for (e = 0 ; e != 16 ; ++e)
			data[e] = table_64[index[e]];
		__m512i tab_1_L = _mm512_load_epi32(&data[0]);
		__m512i tab_1_H = _mm512_load_epi32(&data[8]);
		_mm512_store_epi32(index, h2);
		for (e = 0 ; e != 16 ; ++e)
			data[e] = table_64[index[e]];
		__m512i tab_2_L = _mm512_load_epi32(&data[0]);
		__m512i tab_2_H = _mm512_load_epi32(&data[8]);
		// mix keys and payloads
		__m512i key_1 = _mm512_mask_blend_epi32(blend_AAAA, tab_1_L, _mm512_swizzle_epi32(tab_1_H, _MM_SWIZ_REG_CDAB));
		__m512i key_2 = _mm512_mask_blend_epi32(blend_AAAA, tab_2_L, _mm512_swizzle_epi32(tab_2_H, _MM_SWIZ_REG_CDAB));
		__m512i val_1 = _mm512_mask_blend_epi32(blend_5555, tab_1_H, _mm512_swizzle_epi32(tab_1_L, _MM_SWIZ_REG_CDAB));
		__m512i val_2 = _mm512_mask_blend_epi32(blend_5555, tab_2_H, _mm512_swizzle_epi32(tab_2_L, _MM_SWIZ_REG_CDAB));
		key_1 = _mm512_permutevar_epi32(unpack, key_1);
		key_2 = _mm512_permutevar_epi32(unpack, key_2);
		val_1 = _mm512_permutevar_epi32(unpack, val_1);
		val_2 = _mm512_permutevar_epi32(unpack, val_2);
		// matches
		__mmask16 k1 = _mm512_cmpeq_epi32_mask(key_1, key);
		__mmask16 k2 = _mm512_cmpeq_epi32_mask(key_2, key);
		__m512i val = _mm512_mask_blend_epi32(k1, val_2, val_1);
		__mmask16 k = _mm512_kor(k1, k2);
		// store to output
		_mm512_mask_packstorelo_epi32(&buf[b +  0], k, val);
		_mm512_mask_packstorehi_epi32(&buf[b + 16], k, val);
		b += _mm_countbits_64(_mm512_kconcatlo_64(z, k));
		// flush buffer
		if (b > buf_size) {
			b -= buf_size;
			size_t b_i = 0;
			do {
				__m512 x = _mm512_load_ps(&buf[b_i]);
				_mm512_storenrngo_ps(&vals[o], x);
				b_i += 16;
				o += 16;
			} while (b_i != buf_size);
			__m512 x = _mm512_load_ps(&buf[b_i]);
			_mm512_store_ps(&buf[0], x);
		}
	}
	// clean buffer
	size_t b_i = 0;
	while (b_i != b) vals[o++] = buf[b_i++];
	return o;
}

#endif

typedef struct {
	pthread_t id;
	int threads;
	int thread;
	int log_table_bytes;
	uint32_t payload_factor;
	uint32_t *table_factors;
	uint32_t *table_factors_wide;
	bucket_t      *table;
	bucket_wide_t *table_wide;
	size_t inner_size;
	size_t outer_size;
	size_t  join_size;
	uint32_t *inner;
	uint32_t *outer;
	uint64_t *times;
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread, d->threads);
	size_t i, j, b = 0;
	// copy stuff
	int log_table_bytes = d->log_table_bytes;
	uint32_t *inner = d->inner;
	uint32_t *outer = d->outer;
	size_t inner_size = d->inner_size;
	size_t outer_size = d->outer_size;
	uint32_t *table_factors      = d->table_factors;
	uint32_t *table_factors_wide = d->table_factors_wide;
	uint32_t payload_factor = d->payload_factor;
	bucket_t      *table      = d->table;
	bucket_wide_t *table_wide = d->table_wide;
	uint32_t invalid_key = inner[inner_size];
	// build tables
	if (d->thread == 0) {
		size_t f = -1;
		do {
			f++;
			table_factors[0] = (rand() << 1) | 1;
			table_factors[1] = (rand() << 1) | 1;
			i = build(inner, inner_size, table, log_table_bytes,
			          table_factors, invalid_key, payload_factor);
		} while (i != inner_size);
		fprintf(stderr, "Normal table built! (failed %ld times)\n", f);
		f = -1;
		do {
			f++;
			table_factors_wide[0] = (rand() << 1) | 1;
			table_factors_wide[1] = (rand() << 1) | 1;
			i = build_wide(inner, inner_size, table_wide, log_table_bytes,
			               table_factors_wide, invalid_key, payload_factor);
		} while (i != inner_size);
		fprintf(stderr, "Bucketized table built! (failed %ld times)\n", f);
	}
	pthread_barrier_wait(&d->barrier[b++]);
	// begin and end of each table
	size_t beg = ((outer_size / d->threads) & -16) *  d->thread;
	size_t end = ((outer_size / d->threads) & -16) * (d->thread + 1);
	if (d->thread + 1 == d->threads) end = outer_size;
	// compute correct results
	uint64_t checksum = 0;
	size_t join_size = probe_sum(&outer[beg], end - beg, table, log_table_bytes,
	                             table_factors, &checksum);
	// initialize output space
	uint32_t *joined = mamalloc(join_size * sizeof(uint32_t));
	memset(joined, 0xDB, join_size * sizeof(uint32_t));
	// scalar branching
	pthread_barrier_wait(&d->barrier[b++]);
	uint64_t t = thread_time();
	j = probe(&outer[beg], joined, end - beg, table, log_table_bytes, table_factors);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	if (d->thread == 0) fprintf(stderr, "Scalar branching  finished!\n");
	d->times[0] = t;
	assert(j == join_size);
	assert(sum(joined, j) == checksum);
	// scalar branchless
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	j = probe_nob(&outer[beg], joined, end - beg, table, log_table_bytes, table_factors);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	if (d->thread == 0) fprintf(stderr, "Scalar branchless finished!\n");
	d->times[1] = t;
	assert(j == join_size);
	assert(sum(joined, j) == checksum);
	// vector horizontal
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	j = probe_hor(&outer[beg], joined, end - beg, table_wide, log_table_bytes, table_factors_wide);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	if (d->thread == 0) fprintf(stderr, "Vector horizontal finished!\n");
	d->times[2] = t;
	assert(j == join_size);
	assert(sum(joined, j) == checksum);
	// vector vertical (gather both)
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	j = probe_ver(&outer[beg], joined, end - beg, table, log_table_bytes, table_factors);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	if (d->thread == 0) fprintf(stderr, "Vector vertical (blending) finished!\n");
	d->times[3] = t;
	assert(j == join_size);
	assert(sum(joined, j) == checksum);
	// vector vertical (selective gather)
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	j = probe_ver_sel(&outer[beg], joined, end - beg, table, log_table_bytes, table_factors);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	if (d->thread == 0) fprintf(stderr, "Vector vertical (selective) finished!\n");
	d->times[4] = t;
	assert(j == join_size);
	assert(sum(joined, j) == checksum);
	// vector vertical (seperated keys and payloads)
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	j = probe_ver_sep(&outer[beg], joined, end - beg, table, log_table_bytes, table_factors);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	if (d->thread == 0) fprintf(stderr, "Vector vertical (separated) finished!\n");
	d->times[5] = t;
	assert(j == join_size);
	assert(sum(joined, j) == checksum);
	// vector vertical (gather both & emulate)
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	j = probe_ver_emu(&outer[beg], joined, end - beg, table, log_table_bytes, table_factors);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	if (d->thread == 0) fprintf(stderr, "Vector vertical (emulated) finished!\n");
	d->times[6] = t;
	assert(j == join_size);
	assert(sum(joined, j) == checksum);
	// cleanup
	free(joined);
	d->join_size = join_size;
	pthread_exit(NULL);
}

int parse_log_size(const char *s)
{
	char *p = NULL;
	int bits = 0;
	long long base = 1;
	long long num = strtol(s, &p, 10);
	while (base < num) {
		base += base;
		bits++;
	}
	assert(base == num);
	while (isspace(*p)) p++;
	if      (*p == 'K') bits += 10;
	else if (*p == 'M') bits += 20;
	else if (*p == 'G') bits += 30;
	else p--;
	p++;
	assert(*p == 'B');
	p++;
	while (isspace(*p)) p++;
	assert(*p == 0);
	return bits;
}

char *print_log_size(int bits)
{
	static char buf[16];
	if      (bits >= 30) sprintf(buf, "%d GB", 1 << (bits - 30));
	else if (bits >= 20) sprintf(buf, "%d MB", 1 << (bits - 20));
	else if (bits >= 10) sprintf(buf, "%d KB", 1 << (bits - 10));
	else                 sprintf(buf, "%d  B", 1 << bits);
	return buf;
}

int main(int argc, char **argv)
{
	int t, threads = hardware_threads();
	// get arguments
	int log_table_bytes = argc > 1 ? parse_log_size(argv[1]) : 14;
	size_t outer_size = argc > 2 ? atoll(argv[2]) : 1000 * 1000 * 1000;
	double selectivity = argc > 3 ? atof(argv[3]) : 0.9;
	double load_factor = argc > 4 ? atof(argv[4]) : 0.5;
	// extract inner size
	size_t buckets = (((size_t) 1) << log_table_bytes) / sizeof(bucket_t);
	size_t inner_size = load_factor * buckets;
	// print info
	fprintf(stderr, "Threads: %d\n", threads);
	fprintf(stderr, "Table size: %s (2^%d)\n", print_log_size(log_table_bytes), log_table_bytes);
	fprintf(stderr, "Outer tuples: %ld\n", outer_size);
	fprintf(stderr, "Inner tuples: %ld\n", inner_size);
	fprintf(stderr, "Selectivity: %.2f%%\n", selectivity * 100.0);
#ifdef __MIC
	fprintf(stderr, "Running on Phi!\n");
#else
	fprintf(stderr, "Running on Haswell!\n");
#endif
	// method names
	char *names[] = {"Scalar  (branching)",
	                 "Scalar (branchless)",
	                 "Vector (horizontal)",
	                 "Vector   (blending)",
	                 "Vector  (selective)",
	                 "Vector  (separated)",
	                 "Vector   (emulated)"};
	int m, methods = sizeof(names) / sizeof(char*);
	// barriers
	int b, barriers = methods * 2 + 1;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	// initialize data
	uint32_t *inner = NULL;
	uint32_t *outer = NULL;
	size_t join_size = inner_outer(inner_size, outer_size, selectivity, &inner, &outer);
	// init tables
	uint32_t table_factors[2];
	uint32_t table_factors_wide[2];
	uint32_t payload_factor = (rand() << 1) | 1;
	bucket_t      *table      = mamalloc(((size_t) 1) << log_table_bytes);
	bucket_wide_t *table_wide = mamalloc(((size_t) 1) << log_table_bytes);
	// run threads
	info_t info[threads];
	uint64_t times[methods * threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].threads = threads;
		info[t].thread = t;
		info[t].log_table_bytes = log_table_bytes;
		info[t].table     = table;
		info[t].table_wide = table_wide;
		info[t].table_factors      = table_factors;
		info[t].table_factors_wide = table_factors_wide;
		info[t].payload_factor = payload_factor;
		info[t].inner_size = inner_size;
		info[t].outer_size = outer_size;
		info[t].inner = inner;
		info[t].outer = outer;
		info[t].barrier = barrier;
		info[t].times = &times[methods * t];
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	size_t j = 0;
	for (t = 0 ; t != threads ; ++t) {
		pthread_join(info[t].id, NULL);
		j += info[t].join_size;
	}
	assert(j == join_size);
	// cleanup
	free(table);
	free(table_wide);
	free(inner);
	free(outer);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	// print times
	for (m = 0 ; m != methods ; ++m) {
		uint64_t ns = 0;
		for (t = 0 ; t != threads ; ++t)
			ns += info[t].times[m];
		ns /= threads;
		fprintf(stderr, "%s: %.2f btps (%ld ns)\n", names[m], outer_size * 1.0 / ns, ns);
	}
	fflush(stderr);
	printf("%6s", print_log_size(log_table_bytes));
	for (m = 0 ; m != methods ; ++m) {
		uint64_t ns = 0;
		for (t = 0 ; t != threads ; ++t)
			ns += info[t].times[m];
		ns /= threads;
		printf(", %6.3f", outer_size * 1.0 / ns);
	}
	printf("\n");
	return EXIT_SUCCESS;
}
