/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <immintrin.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <sched.h>
#include <time.h>

#include "rand.h"


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *mamalloc(size_t size)
{
	void *ptr = NULL;
	return posix_memalign(&ptr, 64, size) ? NULL : ptr;
}

inline void *align(const void *p)
{
	size_t i = 63 & (size_t) p;
	return (void*) (i ? p + 64 - i : p);
}

int log_2(size_t n)
{
	size_t b = 1;
	int p = 0;
	while (b < n) {
		b += b;
		p++;
	}
	assert(b == n);
	return p;
}

void scalar(const int32_t *keys, void *ranges, size_t size,
            const int32_t *delims, size_t partitions)
{
	size_t i;
	uint8_t  *ranges_8  = (uint8_t*)  ranges;
	uint16_t *ranges_16 = (uint16_t*) ranges;
	const size_t log_partitions = log_2(partitions);
	if (partitions <= 256)
		for (i = 0 ; i != size ; ++i) {
			int32_t key = keys[i];
			size_t lo = 0;
			size_t hi = partitions;
			size_t j = log_partitions;
			do {
				size_t mid = (lo + hi) >> 1;
				if (key > delims[mid - 1])
					lo = mid;
				else
					hi = mid;
			} while (--j);
			ranges_8[i] = lo;
		}
	else
		for (i = 0 ; i != size ; ++i) {
			int32_t key = keys[i];
			size_t lo = 0;
			size_t hi = partitions;
			size_t j = log_partitions;
			do {
				size_t mid = (lo + hi) >> 1;
				if (key > delims[mid - 1])
					lo = mid;
				else
					hi = mid;
			} while (--j);
			ranges_16[i] = lo;
		}
}

#ifndef __MIC

void scalar_nob(const int32_t *keys, void *ranges, size_t size,
                const int32_t *delims, size_t partitions)
{
	size_t i;
	uint8_t  *ranges_8  = (uint8_t*)  ranges;
	uint16_t *ranges_16 = (uint16_t*) ranges;
	const size_t log_partitions = log_2(partitions);
	if (partitions <= 256)
		for (i = 0 ; i != size ; ++i) {
			int32_t key = keys[i];
			size_t lo = 0;
			size_t hi = partitions;
			size_t j = log_partitions;
			do {
				size_t mid = (lo + hi) >> 1;
				asm("cmpl	-4(%4,%2,4), %3\n\t"
				    "cmovgq		%2, %0\n\t"
				    "cmovngq	%2, %1"
				: "=r"(lo), "=r"(hi)
				: "r"(mid), "r"(key), "r"(delims),
				   "0"(lo),  "1"(hi) : "cc");
			} while (--j);
			ranges_8[i] = lo;
		}
	else
		for (i = 0 ; i != size ; ++i) {
			int32_t key = keys[i];
			size_t lo = 0;
			size_t hi = partitions;
			size_t j = log_partitions;
			do {
				size_t mid = (lo + hi) >> 1;
				asm("cmpl	-4(%4,%2,4), %3\n\t"
				    "cmovgq		%2, %0\n\t"
				    "cmovngq	%2, %1"
				: "=r"(lo), "=r"(hi)
				: "r"(mid), "r"(key), "r"(delims),
				   "0"(lo),  "1"(hi) : "cc");
			} while (--j);
			ranges_16[i] = lo;
		}
}

#else

void scalar_nob(const int32_t *keys, void *ranges, size_t size,
                const int32_t *delims, size_t partitions)
{
	size_t i;
	assert(log_2(partitions) > 1);
	uint8_t  *ranges_8  = (uint8_t*)  ranges;
	uint16_t *ranges_16 = (uint16_t*) ranges;
	const size_t partitions_half = partitions >> 1;
	if (partitions <= 256)
		for (i = 0 ; i != size ; ++i) {
			int32_t key = keys[i];
			size_t lo = 0;
			size_t hi = partitions;
			size_t lh = partitions_half;
			do {
				size_t mid = (lo + hi) >> 1;
				int32_t del = delims[mid - 1];
				size_t gt = -(key > del ? 1 : 0);
				lo += lh &  gt;
				hi -= lh & ~gt;
				lh >>= 1;
			} while (lh);
			ranges_8[i] = lo;
		}
	else
		for (i = 0 ; i != size ; ++i) {
			int32_t key = keys[i];
			size_t lo = 0;
			size_t hi = partitions;
			size_t lh = partitions_half;
			do {
				size_t mid = (lo + hi) >> 1;
				int32_t del = delims[mid - 1];
				size_t gt = -(key > del ? 1 : 0);
				lo += lh &  gt;
				hi -= lh & ~gt;
				lh >>= 1;
			} while (lh);
			ranges_16[i] = lo;
		}
}

#endif

#ifndef __MIC

void vector(const int32_t *keys, void *ranges, size_t size,
            const int32_t *delims, size_t partitions)
{
	size_t i;
	uint8_t  *ranges_8  = (uint8_t*)  ranges;
	uint16_t *ranges_16 = (uint16_t*) ranges;
	const size_t log_partitions_1 = log_2(partitions) - 1;
	const __m256i parts_zero = _mm256_set1_epi32(0);
	const __m256i parts_half = _mm256_set1_epi32(partitions >> 1);
	const __m256i parts_size = _mm256_set1_epi32(partitions);
	const __m256i initial_del = _mm256_set1_epi32(delims[(partitions >> 1) - 1]);
	if (partitions <= 256)
		for (i = 0 ; i != size ; i += 8) {
			__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
			__m256i gt = _mm256_cmpgt_epi32(key, initial_del);
			__m256i lo = _mm256_blendv_epi8(parts_zero, parts_half, gt);
			__m256i hi = _mm256_blendv_epi8(parts_half, parts_size, gt);
			size_t j = log_partitions_1;
			do {
				__m256i mid = _mm256_add_epi32(lo, hi);
				mid = _mm256_srli_epi32(mid, 1);
				__m256i del = _mm256_i32gather_epi32(&delims[-1], mid, 4);
				gt = _mm256_cmpgt_epi32(key, del);
				lo = _mm256_blendv_epi8(lo, mid, gt);
				hi = _mm256_blendv_epi8(mid, hi, gt);
			} while (--j);
			hi = _mm256_permute4x64_epi64(lo, _MM_SHUFFLE(1,0,3,2));
			__m128i h = _mm_packs_epi32(_mm256_castsi256_si128(lo),
			                            _mm256_castsi256_si128(hi));
			h = _mm_packus_epi16(h, h);
#if defined __INTEL_COMPILER && __INTEL_COMPILER < 1600
			_mm_stream_si64((long*) (&ranges_8[i]), _mm_cvtsi128_si64(h));
#else
			_mm_stream_si64((long long*) (&ranges_8[i]), _mm_cvtsi128_si64(h));
#endif
		}
	else
		for (i = 0 ; i != size ; i += 8) {
			__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
			__m256i gt = _mm256_cmpgt_epi32(key, initial_del);
			__m256i lo = _mm256_blendv_epi8(parts_zero, parts_half, gt);
			__m256i hi = _mm256_blendv_epi8(parts_half, parts_size, gt);
			size_t j = log_partitions_1;
			do {
				__m256i mid = _mm256_add_epi32(lo, hi);
				mid = _mm256_srli_epi32(mid, 1);
				__m256i del = _mm256_i32gather_epi32(&delims[-1], mid, 4);
				gt = _mm256_cmpgt_epi32(key, del);
				lo = _mm256_blendv_epi8(lo, mid, gt);
				hi = _mm256_blendv_epi8(mid, hi, gt);
			} while (--j);
			hi = _mm256_permute4x64_epi64(lo, _MM_SHUFFLE(1,0,3,2));
			__m128i h = _mm_packs_epi32(_mm256_castsi256_si128(lo),
			                            _mm256_castsi256_si128(hi));
			_mm_stream_si128((__m128i*) &ranges_16[i], h);
		}
}

#else

void vector(const int32_t *keys, void *ranges, size_t size,
            const int32_t *delims, size_t partitions)
{
	size_t i;
	int32_t parts_space[31];
	int32_t *parts = align(parts_space);
	uint8_t  *ranges_8  = (uint8_t*)  ranges;
	uint16_t *ranges_16 = (uint16_t*) ranges;
	const size_t log_partitions_1 = log_2(partitions) - 1;
	const __m512i parts_zero = _mm512_set1_epi32(0);
	const __m512i parts_half = _mm512_set1_epi32(partitions >> 1);
	const __m512i parts_size = _mm512_set1_epi32(partitions);
	const __m512i initial_del = _mm512_set1_epi32(delims[(partitions >> 1) - 1]);
	if (partitions <= 256)
		for (i = 0 ; i != size ; i += 16) {
			__m512i key = _mm512_load_si512(&keys[i]);
			__mmask16 k = _mm512_cmpgt_epi32_mask(key, initial_del);
			__m512i lo = _mm512_mask_blend_epi32(k, parts_zero, parts_half);
			__m512i hi = _mm512_mask_blend_epi32(k, parts_half, parts_size);
			size_t j = log_partitions_1;
			do {
				__m512i mid = _mm512_add_epi32(lo, hi);
				mid = _mm512_srli_epi32(mid, 1);
				__m512i del = _mm512_i32gather_epi32(mid, &delims[-1], 4);
				k = _mm512_cmpgt_epi32_mask(key, del);
				lo = _mm512_mask_blend_epi32(k, lo, mid);
				hi = _mm512_mask_blend_epi32(k, mid, hi);
			} while (--j);
			_mm512_extstore_epi32(&ranges_8[i], lo, _MM_DOWNCONV_EPI32_UINT8, _MM_HINT_NONE);
		}
	else
		for (i = 0 ; i != size ; i += 16) {
			__m512i key = _mm512_load_si512(&keys[i]);
			__mmask16 k = _mm512_cmpgt_epi32_mask(key, initial_del);
			__m512i lo = _mm512_mask_blend_epi32(k, parts_zero, parts_half);
			__m512i hi = _mm512_mask_blend_epi32(k, parts_half, parts_size);
			size_t j = log_partitions_1;
			do {
				__m512i mid = _mm512_add_epi32(lo, hi);
				mid = _mm512_srli_epi32(mid, 1);
				__m512i del = _mm512_i32gather_epi32(mid, &delims[-1], 4);
				k = _mm512_cmpgt_epi32_mask(key, del);
				lo = _mm512_mask_blend_epi32(k, lo, mid);
				hi = _mm512_mask_blend_epi32(k, mid, hi);
			} while (--j);
			_mm512_extstore_epi32(&ranges_16[i], lo, _MM_DOWNCONV_EPI32_UINT16, _MM_HINT_NONE);
		}
}

#endif

#ifndef __MIC

#ifndef __INTEL_COMPILER
inline uint64_t _mm_tzcnt_64(uint64_t i)
{
	uint64_t o;
	asm("tzcntq	%1, %0" : "=r"(o) : "r"(i) : "cc");
	return o;
}
#endif

inline size_t _mm256_movemask_epi32(__m256i x)
{
	size_t r;
	asm("vmovmskps	%1, %0" : "=r"(r) : "x"(x));
	return r;
}

void tree_9(const int32_t *keys, void *ranges, size_t size, int32_t *tree[1])
{
	size_t i;
	int *ranges_32 = (int*) ranges;
	const int32_t *lvl_1 = tree[0];
	const __m256i del = _mm256_load_si256((__m256i*) lvl_1);
	for (i = 0 ; i != size ; i += 4) {
		__m128i key_cmp = _mm_load_si128((__m128i*) &keys[i]);
		__m256i key = _mm256_castsi128_si256(key_cmp);
		key = _mm256_permute4x64_epi64(key, _MM_SHUFFLE(1,0,1,0));
		__m256i key_1 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(0,0,0,0));
		__m256i key_2 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(1,1,1,1));
		__m256i key_3 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(2,2,2,2));
		__m256i key_4 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(3,3,3,3));
		// level 1/1
		__m256i cmp_1 = _mm256_cmpgt_epi32(key_1, del);
		__m256i cmp_2 = _mm256_cmpgt_epi32(key_2, del);
		__m256i cmp_3 = _mm256_cmpgt_epi32(key_3, del);
		__m256i cmp_4 = _mm256_cmpgt_epi32(key_4, del);
		size_t i1 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		size_t i2 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		size_t i3 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		size_t i4 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// store
		size_t i1234 = i1 | (i2 << 8) | (i3 << 16) | (i4 << 24);
		_mm_stream_si32(ranges_32++, i1234);
	}
}

void tree_9x9(const int32_t *keys, void *ranges, size_t size, int32_t *tree[2])
{
	size_t i;
	int *ranges_32 = (int*) ranges;
	const int32_t *lvl_1 = tree[0];
	const int32_t *lvl_2 = tree[1];
	const __m256i del = _mm256_load_si256((__m256i*) lvl_1);
	for (i = 0 ; i != size ; i += 4) {
		__m128i key_cmp = _mm_load_si128((__m128i*) &keys[i]);
		__m256i key = _mm256_castsi128_si256(key_cmp);
		key = _mm256_permute4x64_epi64(key, _MM_SHUFFLE(1,0,1,0));
		__m256i key_1 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(0,0,0,0));
		__m256i key_2 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(1,1,1,1));
		__m256i key_3 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(2,2,2,2));
		__m256i key_4 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(3,3,3,3));
		// level 1/2
		__m256i cmp_1 = _mm256_cmpgt_epi32(key_1, del);
		__m256i cmp_2 = _mm256_cmpgt_epi32(key_2, del);
		__m256i cmp_3 = _mm256_cmpgt_epi32(key_3, del);
		__m256i cmp_4 = _mm256_cmpgt_epi32(key_4, del);
		size_t i1 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		size_t i2 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		size_t i3 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		size_t i4 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// level 2/2
		__m256i del_1 = _mm256_load_si256((__m256i*) &lvl_2[i1 << 3]);
		__m256i del_2 = _mm256_load_si256((__m256i*) &lvl_2[i2 << 3]);
		__m256i del_3 = _mm256_load_si256((__m256i*) &lvl_2[i3 << 3]);
		__m256i del_4 = _mm256_load_si256((__m256i*) &lvl_2[i4 << 3]);
		cmp_1 = _mm256_cmpgt_epi32(key_1, del_1);
		cmp_2 = _mm256_cmpgt_epi32(key_2, del_2);
		cmp_3 = _mm256_cmpgt_epi32(key_3, del_3);
		cmp_4 = _mm256_cmpgt_epi32(key_4, del_4);
		i1 += (i1 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		i2 += (i2 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		i3 += (i3 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		i4 += (i4 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// store
		size_t i1234 = i1 | (i2 << 8) | (i3 << 16) | (i4 << 24);
		_mm_stream_si32(ranges_32++, i1234);
	}
}

void tree_9x9x9(const int32_t *keys, void *ranges, size_t size, int32_t *tree[3])
{
	size_t i;
#if defined __INTEL_COMPILER && __INTEL_COMPILER < 1600
	long *ranges_64 = (long*) ranges;
#else
	long long *ranges_64 = (long long*) ranges;
#endif
	const int32_t *lvl_1 = tree[0];
	const int32_t *lvl_2 = tree[1];
	const int32_t *lvl_3 = tree[2];
	const __m256i del = _mm256_load_si256((__m256i*) lvl_1);
	for (i = 0 ; i != size ; i += 4) {
		__m128i key_cmp = _mm_load_si128((__m128i*) &keys[i]);
		__m256i key = _mm256_castsi128_si256(key_cmp);
		key = _mm256_permute4x64_epi64(key, _MM_SHUFFLE(1,0,1,0));
		__m256i key_1 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(0,0,0,0));
		__m256i key_2 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(1,1,1,1));
		__m256i key_3 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(2,2,2,2));
		__m256i key_4 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(3,3,3,3));
		// level 1/3
		__m256i cmp_1 = _mm256_cmpgt_epi32(key_1, del);
		__m256i cmp_2 = _mm256_cmpgt_epi32(key_2, del);
		__m256i cmp_3 = _mm256_cmpgt_epi32(key_3, del);
		__m256i cmp_4 = _mm256_cmpgt_epi32(key_4, del);
		size_t i1 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		size_t i2 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		size_t i3 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		size_t i4 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// level 2/3
		__m256i del_1 = _mm256_load_si256((__m256i*) &lvl_2[i1 << 3]);
		__m256i del_2 = _mm256_load_si256((__m256i*) &lvl_2[i2 << 3]);
		__m256i del_3 = _mm256_load_si256((__m256i*) &lvl_2[i3 << 3]);
		__m256i del_4 = _mm256_load_si256((__m256i*) &lvl_2[i4 << 3]);
		cmp_1 = _mm256_cmpgt_epi32(key_1, del_1);
		cmp_2 = _mm256_cmpgt_epi32(key_2, del_2);
		cmp_3 = _mm256_cmpgt_epi32(key_3, del_3);
		cmp_4 = _mm256_cmpgt_epi32(key_4, del_4);
		i1 += (i1 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		i2 += (i2 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		i3 += (i3 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		i4 += (i4 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// level 3/3
		del_1 = _mm256_load_si256((__m256i*) &lvl_3[i1 << 3]);
		del_2 = _mm256_load_si256((__m256i*) &lvl_3[i2 << 3]);
		del_3 = _mm256_load_si256((__m256i*) &lvl_3[i3 << 3]);
		del_4 = _mm256_load_si256((__m256i*) &lvl_3[i4 << 3]);
		cmp_1 = _mm256_cmpgt_epi32(key_1, del_1);
		cmp_2 = _mm256_cmpgt_epi32(key_2, del_2);
		cmp_3 = _mm256_cmpgt_epi32(key_3, del_3);
		cmp_4 = _mm256_cmpgt_epi32(key_4, del_4);
		i1 += (i1 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		i2 += (i2 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		i3 += (i3 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		i4 += (i4 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// store
		size_t i1234 = i1 | (i2 << 16) | (i3 << 32) | (i4 << 48);
		_mm_stream_si64(ranges_64++, i1234);
	}
}

void tree_9x9x9x9(const int32_t *keys, void *ranges, size_t size, int32_t *tree[4])
{
	size_t i;
#if defined __INTEL_COMPILER && __INTEL_COMPILER < 1600
	long *ranges_64 = (long*) ranges;
#else
	long long *ranges_64 = (long long*) ranges;
#endif
	const int32_t *lvl_1 = tree[0];
	const int32_t *lvl_2 = tree[1];
	const int32_t *lvl_3 = tree[2];
	const int32_t *lvl_4 = tree[3];
	const __m256i del = _mm256_load_si256((__m256i*) lvl_1);
	for (i = 0 ; i != size ; i += 4) {
		__m128i key_cmp = _mm_load_si128((__m128i*) &keys[i]);
		__m256i key = _mm256_castsi128_si256(key_cmp);
		key = _mm256_permute4x64_epi64(key, _MM_SHUFFLE(1,0,1,0));
		__m256i key_1 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(0,0,0,0));
		__m256i key_2 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(1,1,1,1));
		__m256i key_3 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(2,2,2,2));
		__m256i key_4 = _mm256_shuffle_epi32(key, _MM_SHUFFLE(3,3,3,3));
		// level 1/4
		__m256i cmp_1 = _mm256_cmpgt_epi32(key_1, del);
		__m256i cmp_2 = _mm256_cmpgt_epi32(key_2, del);
		__m256i cmp_3 = _mm256_cmpgt_epi32(key_3, del);
		__m256i cmp_4 = _mm256_cmpgt_epi32(key_4, del);
		size_t i1 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		size_t i2 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		size_t i3 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		size_t i4 = _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// level 2/4
		__m256i del_1 = _mm256_load_si256((__m256i*) &lvl_2[i1 << 3]);
		__m256i del_2 = _mm256_load_si256((__m256i*) &lvl_2[i2 << 3]);
		__m256i del_3 = _mm256_load_si256((__m256i*) &lvl_2[i3 << 3]);
		__m256i del_4 = _mm256_load_si256((__m256i*) &lvl_2[i4 << 3]);
		cmp_1 = _mm256_cmpgt_epi32(key_1, del_1);
		cmp_2 = _mm256_cmpgt_epi32(key_2, del_2);
		cmp_3 = _mm256_cmpgt_epi32(key_3, del_3);
		cmp_4 = _mm256_cmpgt_epi32(key_4, del_4);
		i1 += (i1 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		i2 += (i2 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		i3 += (i3 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		i4 += (i4 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// level 3/4
		del_1 = _mm256_load_si256((__m256i*) &lvl_3[i1 << 3]);
		del_2 = _mm256_load_si256((__m256i*) &lvl_3[i2 << 3]);
		del_3 = _mm256_load_si256((__m256i*) &lvl_3[i3 << 3]);
		del_4 = _mm256_load_si256((__m256i*) &lvl_3[i4 << 3]);
		cmp_1 = _mm256_cmpgt_epi32(key_1, del_1);
		cmp_2 = _mm256_cmpgt_epi32(key_2, del_2);
		cmp_3 = _mm256_cmpgt_epi32(key_3, del_3);
		cmp_4 = _mm256_cmpgt_epi32(key_4, del_4);
		i1 += (i1 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		i2 += (i2 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		i3 += (i3 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		i4 += (i4 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// level 4/4
		del_1 = _mm256_load_si256((__m256i*) &lvl_4[i1 << 3]);
		del_2 = _mm256_load_si256((__m256i*) &lvl_4[i2 << 3]);
		del_3 = _mm256_load_si256((__m256i*) &lvl_4[i3 << 3]);
		del_4 = _mm256_load_si256((__m256i*) &lvl_4[i4 << 3]);
		cmp_1 = _mm256_cmpgt_epi32(key_1, del_1);
		cmp_2 = _mm256_cmpgt_epi32(key_2, del_2);
		cmp_3 = _mm256_cmpgt_epi32(key_3, del_3);
		cmp_4 = _mm256_cmpgt_epi32(key_4, del_4);
		i1 += (i1 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_1) ^ 0x1FF);
		i2 += (i2 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_2) ^ 0x1FF);
		i3 += (i3 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_3) ^ 0x1FF);
		i4 += (i4 << 3) + _mm_tzcnt_64(_mm256_movemask_epi32(cmp_4) ^ 0x1FF);
		// store
		size_t i1234 = i1 | (i2 << 16) | (i3 << 32) | (i4 << 48);
		_mm_stream_si64(ranges_64++, i1234);
	}
}

#else

void tree_17(const int32_t *keys, void *ranges, size_t size, int32_t *tree[1])
{
	size_t i;
	uint32_t *ranges_32 = (uint32_t*) ranges;
	const int32_t *lvl_1 = tree[0];
	const __mmask16 o = _mm512_int2mask(1);
	const __m512i del = _mm512_load_epi32(lvl_1);
	for (i = 0 ; i != size ; i += 4) {
		__m512i key = _mm512_extload_epi32(&keys[i], _MM_UPCONV_EPI32_NONE,
		                                   _MM_BROADCAST_4X16, _MM_HINT_NONE);
		// level 1/1
		__mmask16 k1 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_AAAA));
		__mmask16 k2 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_BBBB));
		__mmask16 k3 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_CCCC));
		__mmask16 k4 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_DDDD));
		size_t i1 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k1));
		size_t i2 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k2));
		size_t i3 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k3));
		size_t i4 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k4));
		// store
		*ranges_32++ = i1 | (i2 << 8) | (i3 << 16) | (i4 << 24);
	}
}

void tree_17x17(const int32_t *keys, void *ranges, size_t size, int32_t *tree[2])
{
	size_t i, p;
	uint64_t *ranges_64 = (uint64_t*) ranges;
	const int32_t *lvl_1 =                  tree[0];
	const int64_t *lvl_2 = (const int64_t*) tree[1];
	const __mmask16 o = _mm512_int2mask(1);
	const __m512i del = _mm512_load_epi32(lvl_1);
	for (i = 0 ; i != size ; i += 4) {
		__m512i key = _mm512_extload_epi32(&keys[i], _MM_UPCONV_EPI32_NONE,
		                                   _MM_BROADCAST_4X16, _MM_HINT_NONE);
		// level 1/2
		__mmask16 k1 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_AAAA));
		__mmask16 k2 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_BBBB));
		__mmask16 k3 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_CCCC));
		__mmask16 k4 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_DDDD));
		size_t i1 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k1));
		size_t i2 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k2));
		size_t i3 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k3));
		size_t i4 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k4));
		// level 2/2
		size_t j1 = i1 << 3;
		size_t j2 = i2 << 3;
		size_t j3 = i3 << 3;
		size_t j4 = i4 << 3;
		__m512i del_1 = _mm512_load_epi32(&lvl_2[j1]);
		__m512i del_2 = _mm512_load_epi32(&lvl_2[j2]);
		__m512i del_3 = _mm512_load_epi32(&lvl_2[j3]);
		__m512i del_4 = _mm512_load_epi32(&lvl_2[j4]);
		k1 = _mm512_cmpge_epi32_mask(del_1, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_AAAA));
		k2 = _mm512_cmpge_epi32_mask(del_2, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_BBBB));
		k3 = _mm512_cmpge_epi32_mask(del_3, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_CCCC));
		k4 = _mm512_cmpge_epi32_mask(del_4, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_DDDD));
		i1 += (j1 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k1));
		i2 += (j2 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k2));
		i3 += (j3 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k3));
		i4 += (j4 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k4));
		// store
		*ranges_64++ = i1 | (i2 << 16) | (i3 << 32) | (i4 << 48);
	}
}

void tree_17x17x17(const int32_t *keys, void *ranges, size_t size, int32_t *tree[3])
{
	size_t i, p;
	uint64_t *ranges_64 = (uint64_t*) ranges;
	const int32_t *lvl_1 =                  tree[0];
	const int64_t *lvl_2 = (const int64_t*) tree[1];
	const int64_t *lvl_3 = (const int64_t*) tree[2];
	const __mmask16 o = _mm512_int2mask(1);
	const __m512i del = _mm512_load_epi32(lvl_1);
	for (i = 0 ; i != size ; i += 4) {
		__m512i key = _mm512_extload_epi32(&keys[i], _MM_UPCONV_EPI32_NONE,
		                                   _MM_BROADCAST_4X16, _MM_HINT_NONE);
		// level 1/3
		__mmask16 k1 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_AAAA));
		__mmask16 k2 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_BBBB));
		__mmask16 k3 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_CCCC));
		__mmask16 k4 = _mm512_cmpge_epi32_mask(del, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_DDDD));
		size_t i1 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k1));
		size_t i2 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k2));
		size_t i3 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k3));
		size_t i4 = _mm_tzcnt_64(_mm512_kconcatlo_64(o, k4));
		// level 2/3
		size_t j1 = i1 << 3;
		size_t j2 = i2 << 3;
		size_t j3 = i3 << 3;
		size_t j4 = i4 << 3;
		__m512i del_1 = _mm512_load_epi32(&lvl_2[j1]);
		__m512i del_2 = _mm512_load_epi32(&lvl_2[j2]);
		__m512i del_3 = _mm512_load_epi32(&lvl_2[j3]);
		__m512i del_4 = _mm512_load_epi32(&lvl_2[j4]);
		k1 = _mm512_cmpge_epi32_mask(del_1, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_AAAA));
		k2 = _mm512_cmpge_epi32_mask(del_2, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_BBBB));
		k3 = _mm512_cmpge_epi32_mask(del_3, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_CCCC));
		k4 = _mm512_cmpge_epi32_mask(del_4, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_DDDD));
		i1 += (j1 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k1));
		i2 += (j2 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k2));
		i3 += (j3 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k3));
		i4 += (j4 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k4));
		// level 3/3
		j1 = i1 << 3;
		j2 = i2 << 3;
		j3 = i3 << 3;
		j4 = i4 << 3;
		del_1 = _mm512_load_epi32(&lvl_3[j1]);
		del_2 = _mm512_load_epi32(&lvl_3[j2]);
		del_3 = _mm512_load_epi32(&lvl_3[j3]);
		del_4 = _mm512_load_epi32(&lvl_3[j4]);
		k1 = _mm512_cmpge_epi32_mask(del_1, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_AAAA));
		k2 = _mm512_cmpge_epi32_mask(del_2, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_BBBB));
		k3 = _mm512_cmpge_epi32_mask(del_3, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_CCCC));
		k4 = _mm512_cmpge_epi32_mask(del_4, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_DDDD));
		i1 += (j1 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k1));
		i2 += (j2 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k2));
		i3 += (j3 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k3));
		i4 += (j4 << 1) + _mm_tzcnt_64(_mm512_kconcatlo_64(o, k4));
		// store
		*ranges_64++ = i1 | (i2 << 16) | (i3 << 32) | (i4 << 48);
	}
}

#endif

size_t sizeof_partitions(const int32_t *delims)
{
	size_t i = 0;
	while (delims[i] != 0x7FFFFFFF) i++;
	return i + 1;
}

size_t check(const int32_t *keys, const void *ranges, size_t size,
             const int32_t *delims, size_t partitions)
{
	size_t i;
	const uint8_t  *ranges_8  = (const uint8_t*)  ranges;
	const uint16_t *ranges_16 = (const uint16_t*) ranges;
	for (i = 1 ; i != partitions ; ++i)
		assert(delims[i - 1] < delims[i]);
	if (partitions <= 256)
		for (i = 0 ; i != size ; ++i) {
			size_t r = ranges_8[i];
			if (r >= partitions) break;
			if (keys[i] > delims[r]) break;
			if (r && keys[i] <= delims[r - 1]) break;
		}
	else
		for (i = 0 ; i != size ; ++i) {
			size_t r = ranges_16[i];
			if (r >= partitions) break;
			if (keys[i] > delims[r]) break;
			if (r && keys[i] <= delims[r - 1]) break;
		}
	return i;
}

typedef struct {
	int32_t **tree_9;
	int32_t **tree_9x9;
	int32_t **tree_9x9x9;
	int32_t **tree_9x9x9x9;
	int32_t **tree_17;
	int32_t **tree_17x17;
	int32_t **tree_17x17x17;
} index_t;

typedef struct {
	pthread_t id;
	int threads;
	int thread;
	int seed;
	size_t size;
	int32_t **delimiters;
	uint64_t *times;
	index_t *indexes;
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread, d->threads);
	size_t i, r, m = 0, b = 0;
	size_t size = d->size;
	// create input
	int32_t *keys = mamalloc(size * sizeof(int32_t));
	rand32_t *gen = rand32_init(d->seed);
	for (i = 0 ; i != size ; ++i)
		keys[i] = (int32_t) rand32_next(gen);
	free(gen);
	// create output
	void *ranges = mamalloc(size * sizeof(uint16_t));
	memset(ranges, 0xDB, size * sizeof(uint16_t));
	// binary search experiments
	for (r = 0 ; d->delimiters[r] != NULL ; ++r) {
		const int32_t *delimiters = d->delimiters[r];
		size_t partitions = sizeof_partitions(delimiters);
		// scalar (branching)
		pthread_barrier_wait(&d->barrier[b++]);
		uint64_t t = thread_time();
		scalar(keys, ranges, size, delimiters, partitions);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (d->thread == 0) fprintf(stderr, "Scalar (branching)  finished! (%ld partitions)\n", partitions);
		assert(check(keys, ranges, size, delimiters, partitions) == size);
		// scalar (branchless)
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		scalar_nob(keys, ranges, size, delimiters, partitions);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (d->thread == 0) fprintf(stderr, "Scalar (branchless) finished! (%ld partitions)\n", partitions);
		assert(check(keys, ranges, size, delimiters, partitions) == size);
		// vector (binary search)
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		vector(keys, ranges, size, delimiters, partitions);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (d->thread == 0) fprintf(stderr, "Vector (vertical)   finished! (%ld partitions)\n", partitions);
		assert(check(keys, ranges, size, delimiters, partitions) == size);
	}
	const int32_t *delimiters = NULL;
	// vector index experiments
#ifndef __MIC
	// 9
	pthread_barrier_wait(&d->barrier[b++]);
	uint64_t t = thread_time();
	tree_9(keys, ranges, size, d->indexes->tree_9);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	d->times[m++] = t;
	if (d->thread == 0) fprintf(stderr, "Index 9 finished!\n");
	delimiters = d->indexes->tree_9[1];
	assert(check(keys, ranges, size, delimiters, 9) == size);
	// 9 x 9
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	tree_9x9(keys, ranges, size, d->indexes->tree_9x9);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	d->times[m++] = t;
	if (d->thread == 0) fprintf(stderr, "Index 9 x 9 finished!\n");
	delimiters = d->indexes->tree_9x9[2];
	assert(check(keys, ranges, size, delimiters, 9 * 9) == size);
	// 9 x 9 x 9
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	tree_9x9x9(keys, ranges, size, d->indexes->tree_9x9x9);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	d->times[m++] = t;
	if (d->thread == 0) fprintf(stderr, "Index 9 x 9 x 9 finished!\n");
	delimiters = d->indexes->tree_9x9x9[3];
	assert(check(keys, ranges, size, delimiters, 9 * 9 * 9) == size);
	// 9 x 9 x 9 x 9
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	tree_9x9x9x9(keys, ranges, size, d->indexes->tree_9x9x9x9);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	d->times[m++] = t;
	if (d->thread == 0) fprintf(stderr, "Index 9 x 9 x 9 x 9 finished!\n");
	delimiters = d->indexes->tree_9x9x9x9[4];
	assert(check(keys, ranges, size, delimiters, 9 * 9 * 9 * 9) == size);
#else
	// 17
	pthread_barrier_wait(&d->barrier[b++]);
	uint64_t t = thread_time();
	tree_17(keys, ranges, size, d->indexes->tree_17);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	d->times[m++] = t;
	if (d->thread == 0) fprintf(stderr, "Index 17 finished!\n");
	delimiters = d->indexes->tree_17[1];
	assert(check(keys, ranges, size, delimiters, 17) == size);
	// 17 x 17
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	tree_17x17(keys, ranges, size, d->indexes->tree_17x17);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	d->times[m++] = t;
	if (d->thread == 0) fprintf(stderr, "Index 17 x 17 finished!\n");
	delimiters = d->indexes->tree_17x17[2];
	assert(check(keys, ranges, size, delimiters, 17 * 17) == size);
	// 17 x 17 x 17
	pthread_barrier_wait(&d->barrier[b++]);
	t = thread_time();
	tree_17x17x17(keys, ranges, size, d->indexes->tree_17x17x17);
	t = thread_time() - t;
	pthread_barrier_wait(&d->barrier[b++]);
	d->times[m++] = t;
	if (d->thread == 0) fprintf(stderr, "Index 17 x 17 x 17 finished!\n");
	delimiters = d->indexes->tree_17x17x17[3];
	assert(check(keys, ranges, size, delimiters, 17 * 17 * 17) == size);
#endif
	// cleanup
	free(keys);
	free(ranges);
	pthread_exit(NULL);
}

int32_t *create_delimiters(size_t size)
{
	size_t i;
	uint32_t *delim = mamalloc(size * sizeof(uint32_t));
	for (i = 0 ; i != size ; ++i)
		delim[i] = (0xFFFFFFFF * (double) (i + 1)) / size;
	assert(delim[i - 1] == 0xFFFFFFFF);
	// check unsigned
	for (i = 1 ; i != size ; ++i)
		assert(delim[i - 1] < delim[i]);
	for (i = 0 ; i != size ; ++i)
		delim[i] -= 1 << 31;
	// check signed
	int32_t *sign_delim = (int32_t*) delim;
	for (i = 1 ; i != size ; ++i)
		assert(sign_delim[i - 1] < sign_delim[i]);
	assert(sign_delim[i - 1] == 0x7FFFFFFF);
	return sign_delim;
}

void build_index_rec(int32_t **tree, size_t fanout, size_t level, size_t levels,
                     const int32_t *ranges, size_t *i_local, size_t *i_global)
{
	size_t f;
	if (level == levels) return;
	build_index_rec(tree, fanout, level + 1, levels, ranges, i_local, i_global);
	for (f = 1 ; f != fanout ; ++f) {
		size_t local = i_local[level]++;
		size_t global = i_global[0]++;
		tree[level][local] = ranges[global];
		build_index_rec(tree, fanout, level + 1, levels, ranges, i_local, i_global);
	}
}

int32_t **build_index(size_t fanout, size_t levels, const int32_t *ranges)
{
	int32_t **tree = malloc((levels + 1) * sizeof(int32_t*));
	size_t l, i, size = 1;
	for (l = 0 ; l != levels ; ++l) {
		tree[l] = mamalloc(size * (fanout - 1) * sizeof(int32_t));
		size *= fanout;
	}
	size_t *i_local = calloc(levels, sizeof(size_t));
	size_t i_global = 0;
	build_index_rec(tree, fanout, 0, levels, ranges, i_local, &i_global);
	assert(i_global + 1 == size);
	size = 1;
	for (l = 0 ; l != levels ; ++l) {
		assert(i_local[l] == size * (fanout - 1));
		for (i = 1 ; i != i_local[l] ; ++i)
			assert(tree[l][i - 1] <= tree[l][i]);
		size *= fanout;
	}
	free(i_local);
	tree[levels] = (int32_t*) ranges;
	return tree;
}

int main(int argc, char **argv)
{
	int t, threads = hardware_threads();
	size_t size = argc > 1 ? atoll(argv[1]) : 1000 * 1000 * 1000;
	fprintf(stderr, "Tuples: %ld\n", size);
	fprintf(stderr, "Threads: %d\n", threads);
	// build trees
	index_t index;
#ifdef __MIC
	fprintf(stderr, "Running on Phi!\n");
	char *extra_names[] = {"17", "17 x 17", "17 x 17 x 17"};
	index.tree_17       = build_index(17, 1, create_delimiters(17));
	index.tree_17x17    = build_index(17, 2, create_delimiters(17 * 17));
	index.tree_17x17x17 = build_index(17, 3, create_delimiters(17 * 17 * 17));
#else
	fprintf(stderr, "Running on Haswell!\n");
	char *extra_names[] = {"9", "9 x 9", "9 x 9 x 9", "9 x 9 x 9 x 9"};
	index.tree_9       = build_index(9, 1, create_delimiters(9));
	index.tree_9x9     = build_index(9, 2, create_delimiters(9 * 9));
	index.tree_9x9x9   = build_index(9, 3, create_delimiters(9 * 9 * 9));
	index.tree_9x9x9x9 = build_index(9, 4, create_delimiters(9 * 9 * 9 * 9));
#endif
	fprintf(stderr, "Vector indexes built!\n");
	// create delimiters for binary search
#ifdef _NO_BS
	size_t partitions[0];
#else
	size_t partitions[] = {8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192};
#endif
	int extra_methods = sizeof(extra_names) / sizeof(char*);
	char *names[] = {"Scalar (branching) ", "Scalar (branchless)", "Vector (vertical)"};
	int m, methods = sizeof(names) / sizeof(char*);
	int r, repeats = sizeof(partitions) / sizeof(size_t);
	int32_t *delimiters[repeats + 1];
	for (r = 0 ; r != repeats ; ++r) {
		delimiters[r] = create_delimiters(partitions[r]);
		assert(sizeof_partitions(delimiters[r]) == partitions[r]);
	}
	delimiters[r] = NULL;
	// barriers
	int b, barriers = (methods * repeats + extra_methods) * 2;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	// run threads
	srand(time(NULL));
	info_t info[threads];
	uint64_t times[(methods * repeats + extra_methods) * threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].threads = threads;
		info[t].thread = t;
		info[t].seed = rand();
		info[t].size = (size / threads) & ~15;
		info[t].times = &times[(methods * repeats + extra_methods) * t];
		info[t].delimiters = delimiters;
		info[t].indexes = &index;
		info[t].barrier = barrier;
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (r = 0 ; r != repeats ; ++r) {
		printf("%4ld", sizeof_partitions(delimiters[r]));
		for (m = 0 ; m != methods ; ++m) {
			uint64_t ns = 0;
			for (t = 0 ; t != threads ; ++t)
				ns += info[t].times[methods * r + m];
			ns /= threads;
			printf(", %6.3f", size * 1.0 / ns);
		}
		printf("\n");
	}
	for (m = 0 ; m != extra_methods ; ++m) {
		uint64_t ns = 0;
		for (t = 0 ; t != threads ; ++t)
			ns += info[t].times[methods * repeats + m];
		ns /= threads;
		printf("%-20s, %6.3f\n", extra_names[m], size * 1.0 / ns);
	}
	return EXIT_SUCCESS;
}
