/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <immintrin.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <sched.h>
#include <time.h>

#include "rand.h"


// scalar 1 -> branching
// scalar 2 -> branchless

// vector 1 -> extract bits then access key & val directly
// vector 2 -> extract bits then gather key & val by rid
// vector 3 -> selective store of rids then gather key & val
// vector 4 -> selective store of key & rid directly


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *thread_set = CPU_ALLOC(threads);
	assert(thread_set != NULL);
	CPU_ZERO_S(size, thread_set);
	CPU_SET_S(thread, size, thread_set);
	assert(pthread_setaffinity_np(pthread_self(), size, thread_set) == 0);
	CPU_FREE(thread_set);
}

void *mamalloc(size_t size)
{
	void *ptr = NULL;
	return posix_memalign(&ptr, 64, size) ? NULL : ptr;
}

void *align(const void *p)
{
	while (63 & (size_t) p) p++;
	return (void*) p;
}

inline void store(int32_t *p, uint32_t v)
{
#ifndef __MIC
	_mm_stream_si32((int*) p, v);
#else
	*p = v;
#endif
}

size_t scalar_1(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o;
	for (i = o = 0 ; i != size ; ++i) {
		int32_t key = keys[i];
		if (key >= key_lo && key <= key_hi) {
			int32_t val = vals[i];
			store(&keys_out[o], key);
			store(&vals_out[o], val);
			o++;
		}
	}
	return o;
}

size_t scalar_2(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o;
	for (i = o = 0 ; i != size ; ++i) {
		int32_t key = keys[i];
		int32_t val = vals[i];
		store(&keys_out[o], key);
		store(&vals_out[o], val);
		o += (key >= key_lo ? 1 : 0) &
		     (key <= key_hi ? 1 : 0);
	}
	return o;
}

#ifndef __MIC

const uint64_t perm[256] = {0x0001020304050607ull,
     0x0102030405060700ull, 0x0002030405060701ull, 0x0203040506070100ull,
     0x0001030405060702ull, 0x0103040506070200ull, 0x0003040506070201ull,
     0x0304050607020100ull, 0x0001020405060703ull, 0x0102040506070300ull,
     0x0002040506070301ull, 0x0204050607030100ull, 0x0001040506070302ull,
     0x0104050607030200ull, 0x0004050607030201ull, 0x0405060703020100ull,
     0x0001020305060704ull, 0x0102030506070400ull, 0x0002030506070401ull,
     0x0203050607040100ull, 0x0001030506070402ull, 0x0103050607040200ull,
     0x0003050607040201ull, 0x0305060704020100ull, 0x0001020506070403ull,
     0x0102050607040300ull, 0x0002050607040301ull, 0x0205060704030100ull,
     0x0001050607040302ull, 0x0105060704030200ull, 0x0005060704030201ull,
     0x0506070403020100ull, 0x0001020304060705ull, 0x0102030406070500ull,
     0x0002030406070501ull, 0x0203040607050100ull, 0x0001030406070502ull,
     0x0103040607050200ull, 0x0003040607050201ull, 0x0304060705020100ull,
     0x0001020406070503ull, 0x0102040607050300ull, 0x0002040607050301ull,
     0x0204060705030100ull, 0x0001040607050302ull, 0x0104060705030200ull,
     0x0004060705030201ull, 0x0406070503020100ull, 0x0001020306070504ull,
     0x0102030607050400ull, 0x0002030607050401ull, 0x0203060705040100ull,
     0x0001030607050402ull, 0x0103060705040200ull, 0x0003060705040201ull,
     0x0306070504020100ull, 0x0001020607050403ull, 0x0102060705040300ull,
     0x0002060705040301ull, 0x0206070504030100ull, 0x0001060705040302ull,
     0x0106070504030200ull, 0x0006070504030201ull, 0x0607050403020100ull,
     0x0001020304050706ull, 0x0102030405070600ull, 0x0002030405070601ull,
     0x0203040507060100ull, 0x0001030405070602ull, 0x0103040507060200ull,
     0x0003040507060201ull, 0x0304050706020100ull, 0x0001020405070603ull,
     0x0102040507060300ull, 0x0002040507060301ull, 0x0204050706030100ull,
     0x0001040507060302ull, 0x0104050706030200ull, 0x0004050706030201ull,
     0x0405070603020100ull, 0x0001020305070604ull, 0x0102030507060400ull,
     0x0002030507060401ull, 0x0203050706040100ull, 0x0001030507060402ull,
     0x0103050706040200ull, 0x0003050706040201ull, 0x0305070604020100ull,
     0x0001020507060403ull, 0x0102050706040300ull, 0x0002050706040301ull,
     0x0205070604030100ull, 0x0001050706040302ull, 0x0105070604030200ull,
     0x0005070604030201ull, 0x0507060403020100ull, 0x0001020304070605ull,
     0x0102030407060500ull, 0x0002030407060501ull, 0x0203040706050100ull,
     0x0001030407060502ull, 0x0103040706050200ull, 0x0003040706050201ull,
     0x0304070605020100ull, 0x0001020407060503ull, 0x0102040706050300ull,
     0x0002040706050301ull, 0x0204070605030100ull, 0x0001040706050302ull,
     0x0104070605030200ull, 0x0004070605030201ull, 0x0407060503020100ull,
     0x0001020307060504ull, 0x0102030706050400ull, 0x0002030706050401ull,
     0x0203070605040100ull, 0x0001030706050402ull, 0x0103070605040200ull,
     0x0003070605040201ull, 0x0307060504020100ull, 0x0001020706050403ull,
     0x0102070605040300ull, 0x0002070605040301ull, 0x0207060504030100ull,
     0x0001070605040302ull, 0x0107060504030200ull, 0x0007060504030201ull,
     0x0706050403020100ull, 0x0001020304050607ull, 0x0102030405060700ull,
     0x0002030405060701ull, 0x0203040506070100ull, 0x0001030405060702ull,
     0x0103040506070200ull, 0x0003040506070201ull, 0x0304050607020100ull,
     0x0001020405060703ull, 0x0102040506070300ull, 0x0002040506070301ull,
     0x0204050607030100ull, 0x0001040506070302ull, 0x0104050607030200ull,
     0x0004050607030201ull, 0x0405060703020100ull, 0x0001020305060704ull,
     0x0102030506070400ull, 0x0002030506070401ull, 0x0203050607040100ull,
     0x0001030506070402ull, 0x0103050607040200ull, 0x0003050607040201ull,
     0x0305060704020100ull, 0x0001020506070403ull, 0x0102050607040300ull,
     0x0002050607040301ull, 0x0205060704030100ull, 0x0001050607040302ull,
     0x0105060704030200ull, 0x0005060704030201ull, 0x0506070403020100ull,
     0x0001020304060705ull, 0x0102030406070500ull, 0x0002030406070501ull,
     0x0203040607050100ull, 0x0001030406070502ull, 0x0103040607050200ull,
     0x0003040607050201ull, 0x0304060705020100ull, 0x0001020406070503ull,
     0x0102040607050300ull, 0x0002040607050301ull, 0x0204060705030100ull,
     0x0001040607050302ull, 0x0104060705030200ull, 0x0004060705030201ull,
     0x0406070503020100ull, 0x0001020306070504ull, 0x0102030607050400ull,
     0x0002030607050401ull, 0x0203060705040100ull, 0x0001030607050402ull,
     0x0103060705040200ull, 0x0003060705040201ull, 0x0306070504020100ull,
     0x0001020607050403ull, 0x0102060705040300ull, 0x0002060705040301ull,
     0x0206070504030100ull, 0x0001060705040302ull, 0x0106070504030200ull,
     0x0006070504030201ull, 0x0607050403020100ull, 0x0001020304050706ull,
     0x0102030405070600ull, 0x0002030405070601ull, 0x0203040507060100ull,
     0x0001030405070602ull, 0x0103040507060200ull, 0x0003040507060201ull,
     0x0304050706020100ull, 0x0001020405070603ull, 0x0102040507060300ull,
     0x0002040507060301ull, 0x0204050706030100ull, 0x0001040507060302ull,
     0x0104050706030200ull, 0x0004050706030201ull, 0x0405070603020100ull,
     0x0001020305070604ull, 0x0102030507060400ull, 0x0002030507060401ull,
     0x0203050706040100ull, 0x0001030507060402ull, 0x0103050706040200ull,
     0x0003050706040201ull, 0x0305070604020100ull, 0x0001020507060403ull,
     0x0102050706040300ull, 0x0002050706040301ull, 0x0205070604030100ull,
     0x0001050706040302ull, 0x0105070604030200ull, 0x0005070604030201ull,
     0x0507060403020100ull, 0x0001020304070605ull, 0x0102030407060500ull,
     0x0002030407060501ull, 0x0203040706050100ull, 0x0001030407060502ull,
     0x0103040706050200ull, 0x0003040706050201ull, 0x0304070605020100ull,
     0x0001020407060503ull, 0x0102040706050300ull, 0x0002040706050301ull,
     0x0204070605030100ull, 0x0001040706050302ull, 0x0104070605030200ull,
     0x0004070605030201ull, 0x0407060503020100ull, 0x0001020307060504ull,
     0x0102030706050400ull, 0x0002030706050401ull, 0x0203070605040100ull,
     0x0001030706050402ull, 0x0103070605040200ull, 0x0003070605040201ull,
     0x0307060504020100ull, 0x0001020706050403ull, 0x0102070605040300ull,
     0x0002070605040301ull, 0x0207060504030100ull, 0x0001070605040302ull,
     0x0107060504030200ull, 0x0007060504030201ull, 0x0706050403020100ull};

#ifndef __INTEL_COMPILER
inline uint64_t _mm_tzcnt_64(uint64_t i)
{
	uint64_t o;
	asm("tzcntq	%1, %0" : "=r"(o) : "r"(i) : "cc");
	return o;
}
#endif

size_t vector_1(int32_t *keys, int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o;
	const __m256i all_1 = _mm256_set1_epi32(-1);
	const __m256i key_L = _mm256_set1_epi32(key_lo);
	const __m256i key_H = _mm256_set1_epi32(key_hi);
	for (i = o = 0 ; i != size ; i += 8) {
		// keys 1 to 4
		__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
		__m256i in_L = _mm256_cmpgt_epi32(key_L, key);
		__m256i in_H = _mm256_cmpgt_epi32(key, key_H);
		__m256i in = _mm256_or_si256(in_L, in_H);
		if (!_mm256_testc_si256(in, all_1)) {
			size_t m = _mm256_movemask_ps(_mm256_castsi256_ps(in));
			m ^= 255;
			do {
				size_t j = _mm_tzcnt_64(m);
				uint32_t k = keys[i + j];
				uint32_t v = vals[i + j];
				_mm_stream_si32(&((int*) keys_out)[o], k);
				_mm_stream_si32(&((int*) vals_out)[o], v);
				o++;
				m &= m - 1;
			} while (m);
		}
	}
	return o;
}

size_t vector_2(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o, b;
	uint32_t buf[8 + 15];
	uint32_t *rids_buf = align(buf);
	const __m256i all_1 = _mm256_set1_epi32(-1);
	const __m256i key_L = _mm256_set1_epi32(key_lo);
	const __m256i key_H = _mm256_set1_epi32(key_hi);
	for (i = o = b = 0 ; i != size ; i += 8) {
		// keys 1 to 4
		__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
		__m256i in_L = _mm256_cmpgt_epi32(key_L, key);
		__m256i in_H = _mm256_cmpgt_epi32(key, key_H);
		__m256i in = _mm256_or_si256(in_L, in_H);
		if (!_mm256_testc_si256(in, all_1)) {
			size_t m = _mm256_movemask_ps(_mm256_castsi256_ps(in));
			m ^= 255;
			do {
				size_t j = _mm_tzcnt_64(m);
				rids_buf[b] = i + j;
				b = (b + 1) & 7;
				if (b == 0) {
					__m256i r = _mm256_load_si256((__m256i*) rids_buf);
					__m256i k = _mm256_i32gather_epi32(keys, r, 4);
					__m256i v = _mm256_i32gather_epi32(vals, r, 4);
					_mm256_stream_si256((__m256i*) &keys_out[o], k);
					_mm256_stream_si256((__m256i*) &vals_out[o], v);
					o += 8;
				}
				m &= m - 1;
			} while (m);
		}
	}
	for (i = 0 ; i != b ; ++i) {
		size_t j = rids_buf[i];
		_mm_stream_si32(&((int*) keys_out)[o], keys[j]);
		_mm_stream_si32(&((int*) vals_out)[o], vals[j]);
		o++;
	}
	return o;
}

size_t vector_3(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o, b;
	const size_t buf_size = 128;
	uint32_t buf[buf_size + 8 + 15];
	uint32_t *rids_buf = align(buf);
	const __m256i all_1 = _mm256_set1_epi32(-1);
	const __m256i key_L = _mm256_set1_epi32(key_lo);
	const __m256i key_H = _mm256_set1_epi32(key_hi);
	__m256i rid = _mm256_set_epi32(7,6,5,4,3,2,1,0);
	__m256i rid_update = _mm256_set1_epi32(8);
	for (i = o = b = 0 ; i != size ; i += 8) {
		__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
		__m256i in_L = _mm256_cmpgt_epi32(key_L, key);
		__m256i in_H = _mm256_cmpgt_epi32(key, key_H);
		__m256i in = _mm256_or_si256(in_L, in_H);
		if (!_mm256_testc_si256(in, all_1)) {
			__m256i out = _mm256_xor_si256(in, all_1);
			size_t m = _mm256_movemask_ps(_mm256_castsi256_ps(out));
			__m128i perm_comp = _mm_loadl_epi64((__m128i*) &perm[m]);
			__m256i perm = _mm256_cvtepi8_epi32(perm_comp);
			__m256i r = _mm256_permutevar8x32_epi32(rid, perm);
			out = _mm256_permutevar8x32_epi32(out, perm);
			_mm256_maskstore_epi32((int*) &rids_buf[b], out, r);
			b += _mm_popcnt_u64(m);
			if (b > buf_size) {
				size_t j = 0;
				do {
					r = _mm256_load_si256((__m256i*) &rids_buf[j]);
					__m256i k = _mm256_i32gather_epi32(keys, r, 4);
					__m256i v = _mm256_i32gather_epi32(vals, r, 4);
					_mm256_stream_si256((__m256i*) &keys_out[o], k);
					_mm256_stream_si256((__m256i*) &vals_out[o], v);
					j += 8;
					o += 8;
				} while (j != buf_size);
				r = _mm256_load_si256((__m256i*) &rids_buf[j]);
				_mm256_store_si256((__m256i*) &rids_buf[0], r);
				b -= buf_size;
			}
		}
		rid = _mm256_add_epi32(rid, rid_update);
	}
	for (i = 0 ; i != b ; ++i) {
		size_t j = rids_buf[i];
		_mm_stream_si32(&((int*) keys_out)[o], keys[j]);
		_mm_stream_si32(&((int*) vals_out)[o], vals[j]);
		o++;
	}
	return o;
}

size_t vector_4(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o, b;
	const size_t buf_size = 128;
	uint32_t buf[(buf_size + 8) * 2 + 15];
	uint32_t *keys_buf = align(buf);
	uint32_t *vals_buf = &keys_buf[buf_size + 8];
	const __m256i all_1 = _mm256_set1_epi32(-1);
	const __m256i key_L = _mm256_set1_epi32(key_lo);
	const __m256i key_H = _mm256_set1_epi32(key_hi);
	for (i = o = b = 0 ; i != size ; i += 8) {
		__m256i key = _mm256_load_si256((__m256i*) &keys[i]);
		__m256i in_L = _mm256_cmpgt_epi32(key_L, key);
		__m256i in_H = _mm256_cmpgt_epi32(key, key_H);
		__m256i in = _mm256_or_si256(in_L, in_H);
		if (!_mm256_testc_si256(in, all_1)) {
			__m256i val = _mm256_load_si256((__m256i*) &vals[i]);
			__m256i out = _mm256_xor_si256(in, all_1);
			size_t m = _mm256_movemask_ps(_mm256_castsi256_ps(out));
			__m128i perm_comp = _mm_loadl_epi64((__m128i*) &perm[m]);
			__m256i perm = _mm256_cvtepi8_epi32(perm_comp);
			key = _mm256_permutevar8x32_epi32(key, perm);
			val = _mm256_permutevar8x32_epi32(val, perm);
			out = _mm256_permutevar8x32_epi32(out, perm);
			_mm256_maskstore_epi32((int*) &keys_buf[b], out, key);
			_mm256_maskstore_epi32((int*) &vals_buf[b], out, val);
			b += _mm_popcnt_u64(m);
			if (b > buf_size) {
				size_t j = 0;
				do {
					key = _mm256_load_si256((__m256i*) &keys_buf[j]);
					val = _mm256_load_si256((__m256i*) &vals_buf[j]);
					_mm256_stream_si256((__m256i*) &keys_out[o], key);
					_mm256_stream_si256((__m256i*) &vals_out[o], val);
					j += 8;
					o += 8;
				} while (j != buf_size);
				key = _mm256_load_si256((__m256i*) &keys_buf[j]);
				val = _mm256_load_si256((__m256i*) &vals_buf[j]);
				_mm256_store_si256((__m256i*) &keys_buf[0], key);
				_mm256_store_si256((__m256i*) &vals_buf[0], val);
				b -= buf_size;
			}
		}
	}
	for (i = 0 ; i != b ; ++i) {
		_mm_stream_si32(&((int*) keys_out)[o], keys_buf[i]);
		_mm_stream_si32(&((int*) vals_out)[o], vals_buf[i]);
		o++;
	}
	return o;
}

#else

size_t vector_1(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o, b;
	uint32_t buf[47];
	uint32_t *keys_buf = align(buf);
	uint32_t *vals_buf = &keys_buf[16];
	const __mmask16 z = _mm512_int2mask(0);
	const __m512i key_L = _mm512_set1_epi32(key_lo);
	const __m512i key_H = _mm512_set1_epi32(key_hi);
	for (i = o = b = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32(&keys[i]);
		__mmask16 k = _mm512_cmpge_epi32_mask(key, key_L);
		k = _mm512_mask_cmple_epi32_mask(k, key, key_H);
		if (!_mm512_kortestz(k, k)) {
			uint64_t m = _mm512_kconcatlo_64(z, k);
			size_t j = _mm_tzcnt_64(m);
			do {
				uint32_t x = keys[i + j];
				uint32_t y = vals[i + j];
				keys_buf[b] = x;
				vals_buf[b] = y;
				b = (b + 1) & 15;
				if (b == 0) {
					__m512 x = _mm512_load_ps(keys_buf);
					__m512 y = _mm512_load_ps(vals_buf);
					_mm512_storenrngo_ps(&keys_out[o], x);
					_mm512_storenrngo_ps(&vals_out[o], y);
					o += 16;
				}
				j = _mm_tzcnti_64(j, m);
			} while (j != 64);
		}
	}
	for (i = 0 ; i != b ; ++i) {
		keys_out[o] = keys_buf[i];
		vals_out[o] = vals_buf[i];
		o++;
	}
	return o;
}

size_t vector_2(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o, b;
	uint32_t buf[31];
	uint32_t *rids_buf = align(buf);
	const __mmask16 z = _mm512_int2mask(0);
	const __m512i key_L = _mm512_set1_epi32(key_lo);
	const __m512i key_H = _mm512_set1_epi32(key_hi);
	for (i = o = b = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32(&keys[i]);
		__mmask16 k = _mm512_cmpge_epi32_mask(key, key_L);
		k = _mm512_mask_cmple_epi32_mask(k, key, key_H);
		if (!_mm512_kortestz(k, k)) {
			uint64_t m = _mm512_kconcatlo_64(z, k);
			size_t j = _mm_tzcnt_64(m);
			do {
				rids_buf[b] = i + j;
				b = (b + 1) & 15;
				if (b == 0) {
					__m512i r = _mm512_load_epi32(rids_buf);
					__m512 x = _mm512_i32gather_ps(r, keys, 4);
					__m512 y = _mm512_i32gather_ps(r, vals, 4);
					_mm512_storenrngo_ps(&keys_out[o], x);
					_mm512_storenrngo_ps(&vals_out[o], y);
					o += 16;
				}
				j = _mm_tzcnti_64(j, m);
			} while (j != 64);
		}
	}
	for (i = 0 ; i != b ; ++i) {
		size_t j = rids_buf[i];
		keys_out[o] = keys[j];
		vals_out[o] = vals[j];
		o++;
	}
	return o;
}

size_t vector_3(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o, b;
	const size_t buf_size = 128;
	uint32_t buf[buf_size + 16 + 15];
	uint32_t *rids_buf = align(buf);
	const __mmask16 z = _mm512_int2mask(0);
	const __m512i key_L = _mm512_set1_epi32(key_lo);
	const __m512i key_H = _mm512_set1_epi32(key_hi);
	__m512i rid = _mm512_set_epi32(15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0);
	__m512i rid_update = _mm512_set1_epi32(16);
	for (i = o = b = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32(&keys[i]);
		__mmask16 k = _mm512_cmpge_epi32_mask(key, key_L);
		k = _mm512_mask_cmple_epi32_mask(k, key, key_H);
		if (!_mm512_kortestz(k, k)) {
			_mm512_mask_packstorelo_epi32(&rids_buf[b +  0], k, rid);
			_mm512_mask_packstorehi_epi32(&rids_buf[b + 16], k, rid);
			b += _mm_countbits_64(_mm512_kconcatlo_64(z, k));
			if (b > buf_size) {
				size_t j = 0;
				do {
					__m512i r = _mm512_load_epi32(&rids_buf[j]);
					__m512 x = _mm512_i32gather_ps(r, keys, 4);
					__m512 y = _mm512_i32gather_ps(r, vals, 4);
					_mm512_storenrngo_ps(&keys_out[o], x);
					_mm512_storenrngo_ps(&vals_out[o], y);
					j += 16;
					o += 16;
				} while (j != buf_size);
				__m512i r = _mm512_load_epi32(&rids_buf[j]);
				_mm512_store_epi32(&rids_buf[0], r);
				b -= buf_size;
			}
		}
		rid = _mm512_add_epi32(rid, rid_update);
	}
	for (i = 0 ; i != b ; ++i) {
		size_t j = rids_buf[i];
		keys_out[o] = keys[j];
		vals_out[o] = vals[j];
		o++;
	}
	return o;
}

size_t vector_4(const int32_t *keys, const int32_t *vals, size_t size,
                int32_t *keys_out, int32_t *vals_out,
                int32_t key_lo, int32_t key_hi)
{
	size_t i, o, b;
	const size_t buf_size = 128;
	uint32_t buf[(buf_size + 16) * 2 + 15];
	uint32_t *keys_buf = align(buf);
	uint32_t *vals_buf = &keys_buf[buf_size + 16];
	const __mmask16 z = _mm512_int2mask(0);
	const __m512i key_L = _mm512_set1_epi32(key_lo);
	const __m512i key_H = _mm512_set1_epi32(key_hi);
	for (i = o = b = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32(&keys[i]);
		__mmask16 k = _mm512_cmpge_epi32_mask(key, key_L);
		k = _mm512_mask_cmple_epi32_mask(k, key, key_H);
		if (!_mm512_kortestz(k, k)) {
			__m512i val = _mm512_load_epi32(&vals[i]);
			_mm512_mask_packstorelo_epi32(&keys_buf[b +  0], k, key);
			_mm512_mask_packstorehi_epi32(&keys_buf[b + 16], k, key);
			_mm512_mask_packstorelo_epi32(&vals_buf[b +  0], k, val);
			_mm512_mask_packstorehi_epi32(&vals_buf[b + 16], k, val);
			b += _mm_countbits_64(_mm512_kconcatlo_64(z, k));
			if (b > buf_size) {
				size_t j = 0;
				do {
					__m512 x = _mm512_load_ps(&keys_buf[j]);
					__m512 y = _mm512_load_ps(&vals_buf[j]);
					_mm512_storenrngo_ps(&keys_out[o], x);
					_mm512_storenrngo_ps(&vals_out[o], y);
					j += 16;
					o += 16;
				} while (j != buf_size);
				__m512i x = _mm512_load_epi32(&keys_buf[j]);
				__m512i y = _mm512_load_epi32(&vals_buf[j]);
				_mm512_store_epi32(&keys_buf[0], x);
				_mm512_store_epi32(&vals_buf[0], y);
				b -= buf_size;
			}
		}
	}
	for (i = 0 ; i != b ; ++i) {
		keys_out[o] = keys_buf[i];
		vals_out[o] = vals_buf[i];
		o++;
	}
	return o;
}

#endif

size_t check(const int32_t *keys, const int32_t *vals, size_t size,
             const int32_t *keys_out, const int32_t *vals_out,
             int32_t key_lo, int32_t key_hi)
{
	size_t i, o;
	for (i = o = 0 ; i != size ; ++i)
		if (keys[i] >= key_lo && keys[i] <= key_hi) {
			if (keys[i] != keys_out[o] || vals[i] != vals_out[o]) break;
			o++;
		}
	return o;
}

typedef struct {
	pthread_t id;
	int threads;
	int thread;
	int seed;
	uint64_t size;
	size_t selectivities;
	double *selectivity;
	size_t *count;
	uint64_t *times;
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread, d->threads);
	size_t s, i, b = 0, m = 0;
	size_t size = d->size;
	assert(size % 16 == 0);
	int32_t *keys     = mamalloc(size * sizeof(int32_t));
	int32_t *vals     = mamalloc(size * sizeof(int32_t));
	int32_t *keys_out = mamalloc(size * sizeof(int32_t));
	int32_t *vals_out = mamalloc(size * sizeof(int32_t));
	rand32_t *gen = rand32_init(d->seed);
	int32_t factor = rand32_next(gen) | 1;
	for (i = 0 ; i != size ; ++i) {
		keys[i] = rand32_next(gen);
		vals[i] = keys[i] * factor;
		keys_out[i] = 0xCAFEBABE;
		vals_out[i] = 0xDEADBEEF;
	}
	free(gen);
	if (!d->thread) fprintf(stderr, "Generation finished\n");
	for (s = 0 ; s != d->selectivities ; ++s) {
		double perc = d->selectivity[s];
		uint32_t range = ~0;
		range *= perc;
		uint32_t half_range = range >> 1;
		int32_t key_lo = -half_range;
		int32_t key_hi =  half_range;
		size_t c = 0;
		for (i = 0 ; i != size ; ++i)
			if (keys[i] >= key_lo && keys[i] <= key_hi) c++;
		d->count[s] = c;
		// scalar 1 -> branching
		pthread_barrier_wait(&d->barrier[b++]);
		uint64_t t = thread_time();
		i = scalar_1(keys, vals, size, keys_out, vals_out, key_lo, key_hi);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (!d->thread) fprintf(stderr, "Scalar 1 finished (%.2f%%)\n", i * 100.0 / size);
		assert(c == i);
		assert(c == check(keys, vals, size, keys_out, vals_out, key_lo, key_hi));
		// scalar 2 -> branchless
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		i = scalar_2(keys, vals, size, keys_out, vals_out, key_lo, key_hi);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (!d->thread) fprintf(stderr, "Scalar 2 finished (%.2f%%)\n", i * 100.0 / size);
		assert(c == i);
		assert(c == check(keys, vals, size, keys_out, vals_out, key_lo, key_hi));
		// vector 1 -> extract bits then access key & val directly
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		i = vector_1(keys, vals, size, keys_out, vals_out, key_lo, key_hi);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (!d->thread) fprintf(stderr, "Vector 1 finished (%.2f%%)\n", i * 100.0 / size);
		assert(c == i);
		assert(c == check(keys, vals, size, keys_out, vals_out, key_lo, key_hi));
		// vector 4 -> selective store of key & rid directly
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		i = vector_4(keys, vals, size, keys_out, vals_out, key_lo, key_hi);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (!d->thread) fprintf(stderr, "Vector 4 finished (%.2f%%)\n", i * 100.0 / size);
		assert(c == i);
		assert(c == check(keys, vals, size, keys_out, vals_out, key_lo, key_hi));
		// vector 2 -> extract bits then gather key & val by rid
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		i = vector_2(keys, vals, size, keys_out, vals_out, key_lo, key_hi);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (!d->thread) fprintf(stderr, "Vector 2 finished (%.2f%%)\n", i * 100.0 / size);
		assert(c == i);
		assert(c == check(keys, vals, size, keys_out, vals_out, key_lo, key_hi));
		// vector 3 -> selective store of rids then gather key & val
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		i = vector_3(keys, vals, size, keys_out, vals_out, key_lo, key_hi);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		if (!d->thread) fprintf(stderr, "Vector 3 finished (%.2f%%)\n", i * 100.0 / size);
		assert(c == i);
		assert(c == check(keys, vals, size, keys_out, vals_out, key_lo, key_hi));
	}
	free(keys);
	free(vals);
	free(keys_out);
	free(vals_out);
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	int t, threads = hardware_threads();
	size_t size = argc > 1 ? atoll(argv[1]) : 500 * 1000 * 1000;
	fprintf(stderr, "Threads: %d\n", threads);
	fprintf(stderr, "Tuples: %ld\n", size);
#ifdef __MIC
	fprintf(stderr, "Running on Phi!\n");
#else
	fprintf(stderr, "Running on Haswell!\n");
#endif
	double selectivity[] = {0, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1.0};
	size_t s, selectivities = sizeof(selectivity) / sizeof(double);
	char *names[] = {"Scalar 1",
	                 "Scalar 2",
	                 "Vector 1",
	                 "Vector 2",
	                 "Vector 3",
	                 "Vector 4"};
	int m, methods = sizeof(names) / sizeof(char*);
	int b, barriers = selectivities * methods * 2;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	info_t info[threads];
	srand(time(NULL));
	size_t   count[threads * selectivities];
	uint64_t times[threads * selectivities * methods];
	for (t = 0 ; t != threads ; ++t) {
		info[t].size = (size / threads) & ~15;
		info[t].seed = rand();
		info[t].thread = t;
		info[t].threads = threads;
		info[t].count = &count[t * selectivities];
		info[t].times = &times[t * selectivities * methods];
		info[t].selectivity   = selectivity;
		info[t].selectivities = selectivities;
		info[t].barrier = barrier;
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	for (s = 0 ; s != selectivities ; ++s) {
		size_t c = 0;
		for (t = 0 ; t != threads ; ++t)
			c += info[t].count[s];
		fprintf(stderr, "Expected: %.2f%%\n", selectivity[s] * 100.0);
		fprintf(stderr, "Measured: %.2f%% (%ld)\n", c * 100.0 / size, c);
		for (m = 0 ; m != methods ; ++m) {
			uint64_t ns = 0;
			for (t = 0 ; t != threads ; ++t)
				ns += info[t].times[s * methods + m];
			ns /= threads;
			fprintf(stderr, "%s: %.3f btps\n", names[m], size * 1.0 / ns);
		}
	}
	fflush(stderr);
	for (s = 0 ; s != selectivities ; ++s) {
		size_t c = 0;
		for (t = 0 ; t != threads ; ++t)
			c += info[t].count[s];
		printf("%5.1f%%", selectivity[s] * 100);
		for (m = 0 ; m != methods ; ++m) {
			uint64_t ns = 0;
			for (t = 0 ; t != threads ; ++t)
				ns += info[t].times[s * methods + m];
			ns /= threads;
			printf(", %6.3f", size * 1.0 / ns);
		}
		printf("\n");
	}
	return EXIT_SUCCESS;
}
