/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#ifdef _NO_VECTOR
#ifndef _NO_VECTOR_HASHING
#define _NO_VECTOR_HASHING
#endif
#ifndef _NO_VECTOR_PARTITIONING
#define _NO_VECTOR_PARTITIONING
#endif
#endif

#ifdef _NO_VECTOR_HASHING
#ifdef _NO_VECTOR_PARTITIONING
#ifndef _NO_VECTOR
#define _NO_VECTOR
#endif
#endif
#endif

#ifndef _NO_VECTOR
#include <immintrin.h>
#endif

#include <immintrin.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <sched.h>
#include <time.h>

#include "rand.h"


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

uint64_t real_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_REALTIME, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *mamalloc(size_t size)
{
	void *ptr = NULL;
	return posix_memalign(&ptr, 64, size) ? NULL : ptr;
}

void *align(const void *p)
{
	size_t i = 63 & (size_t) p;
	return (void*) (i ? p + 64 - i : p);
}

int power_of_2(uint64_t x)
{
	return x > 0 && (x & (x - 1)) == 0;
}

int prime(uint64_t x)
{
	if ((x & 1) == 0) return x == 2;
	uint64_t d, sx = sqrt(x);
	for (d = 3 ; d <= sx ; d += 2)
		if (x % d == 0) return 0;
	return x > 2;
}

#ifndef _NO_VECTOR_HASHING

void set(uint64_t *dst, size_t size, uint32_t value)
{
	uint64_t *dst_end = &dst[size];
	uint64_t *dst_aligned = align(dst);
	__m512i x = _mm512_set1_epi64(value);
	while (dst != dst_end && dst != dst_aligned)
		*dst++ = value;
	dst_aligned = &dst[(dst_end - dst) & ~7];
	while (dst != dst_aligned) {
		_mm512_store_epi64(dst, x);
		dst += 8;
	}
	while (dst != dst_end)
		*dst++ = value;
}

void build(const uint32_t *keys, const uint32_t *vals, size_t size,
           uint64_t *table, size_t buckets, uint32_t factor, uint32_t empty)
{
	set(table, buckets, empty);
	// set constants
	__m512i mask_1 = _mm512_set1_epi32(1);
	__m512i mask_empty = _mm512_set1_epi32(empty);
	__m512i mask_factor = _mm512_set1_epi32(factor);
	__m512i mask_buckets = _mm512_set1_epi32(buckets);
	__m512i mask_pack = _mm512_set_epi32(15, 7, 14, 6, 13, 5, 12, 4, 11, 3, 10, 2, 9, 1, 8, 0);
	__mmask16 blend_0000 = _mm512_int2mask(0x0000);
	__mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	__mmask16 blend_5555 = _mm512_int2mask(0x5555);
	// main loop
	size_t i = 0;
	size_t size_minus_16 = size - 16;
	__mmask16 k = _mm512_kxnor(k, k);
	__m512i key, val, off;
	if (size >= 16) do {
		// replace invalid keys & payloads
		key = _mm512_mask_loadunpacklo_epi32(key, k, &keys[i]);
		key = _mm512_mask_loadunpackhi_epi32(key, k, &keys[i + 16]);
		val = _mm512_mask_loadunpacklo_epi32(val, k, &vals[i]);
		val = _mm512_mask_loadunpackhi_epi32(val, k, &vals[i + 16]);
		off = _mm512_mask_xor_epi32(off, k, off, off);
		i += _mm_countbits_64(_mm512_kconcatlo_64(blend_0000, k));
		// hash keys
		__m512i hash = _mm512_mullo_epi32(key, mask_factor);
		hash = _mm512_mulhi_epu32(hash, mask_buckets);
		hash = _mm512_add_epi32(hash, off);
		k = _mm512_cmpge_epu32_mask(hash, mask_buckets);
		hash = _mm512_mask_sub_epi32(hash, k, hash, mask_buckets);
		// load keys from table and detect conflicts
		__m512i tab = _mm512_i32gather_epi32(hash, table, 8);
		k = _mm512_cmpeq_epi32_mask(tab, mask_empty);
		_mm512_mask_i32scatter_epi32(table, k, hash, mask_pack, 8);
		tab = _mm512_mask_i32gather_epi32(tab, k, hash, table, 8);
		k = _mm512_mask_cmpeq_epi32_mask(k, tab, mask_pack);
		// mix keys and payloads in pairs
		__m512i key_tmp = _mm512_permutevar_epi32(mask_pack, key);
		__m512i val_tmp = _mm512_permutevar_epi32(mask_pack, val);
		__m512i lo = _mm512_mask_blend_epi32(blend_AAAA, key_tmp, _mm512_swizzle_epi32(val_tmp, _MM_SWIZ_REG_CDAB));
		__m512i hi = _mm512_mask_blend_epi32(blend_5555, val_tmp, _mm512_swizzle_epi32(key_tmp, _MM_SWIZ_REG_CDAB));
		// store valid pairs
		_mm512_mask_i32loscatter_epi64(table, k, hash, lo, 8);
		__mmask16 kr = _mm512_kmerge2l1h(k, k);
		hash = _mm512_permute4f128_epi32(hash, _MM_PERM_BADC);
		_mm512_mask_i32loscatter_epi64(table, kr, hash, hi, 8);
		off = _mm512_add_epi32(off, mask_1);
	} while (i <= size_minus_16);
	// save last items
	uint32_t keys_last[32];
	uint32_t vals_last[32];
	k = _mm512_knot(k);
	_mm512_mask_packstorelo_epi32(&keys_last[0],  k, key);
	_mm512_mask_packstorehi_epi32(&keys_last[16], k, key);
	_mm512_mask_packstorelo_epi32(&vals_last[0],  k, val);
	_mm512_mask_packstorehi_epi32(&vals_last[16], k, val);
	size_t j = _mm_countbits_64(_mm512_kconcatlo_64(blend_0000, k));
	for (; i != size ; ++i, ++j) {
		keys_last[j] = keys[i];
		vals_last[j] = vals[i];
	}
	// process last items in scalar code
	for (i = 0 ; i != j ; ++i) {
		uint32_t key = keys_last[i];
		uint64_t pair = vals_last[i];
		pair = (pair << 32) | key;
		size_t h = (uint32_t) (key * factor);
		h = (h * buckets) >> 32;
		while (empty != (uint32_t) table[h])
			if (++h == buckets) h = 0;
		table[h] = pair;
	}
}

size_t probe(const uint32_t *keys, const uint32_t *vals, size_t size,
             const uint64_t *table, size_t buckets, size_t threads,
             uint32_t thread_factor, uint32_t hash_factor, uint32_t empty,
             uint32_t *keys_out, uint32_t *vals_out, uint32_t *tabs_out,
             size_t block_size, size_t block_limit, volatile size_t *counter)
{
	assert(keys_out == align(keys_out));
	assert(vals_out == align(vals_out));
	// generate masks
	__m512i mask_1 = _mm512_set1_epi32(1);
	__m512i mask_empty = _mm512_set1_epi32(empty);
	__m512i mask_hash_factor = _mm512_set1_epi32(hash_factor);
	__m512i mask_thread_factor = _mm512_set1_epi32(thread_factor);
	__m512i mask_threads = _mm512_set1_epi32(threads);
	__m512i mask_buckets = _mm512_set1_epi32(buckets);
	__m512i mask_unpack = _mm512_set_epi32(15, 13, 11, 9, 7, 5, 3, 1, 14, 12, 10, 8, 6, 4, 2, 0);
	__mmask16 blend_0000 = _mm512_int2mask(0x0000);
	__mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	__mmask16 blend_5555 = _mm512_int2mask(0x5555);
	// space for buffers
	const size_t buffer_size = 256;
	uint32_t buffer_space[(buffer_size + 16) * 3 + 15];
	uint32_t *keys_buf = align(buffer_space);
	uint32_t *vals_buf = &keys_buf[buffer_size + 16];
	uint32_t *tabs_buf = &vals_buf[buffer_size + 16];
	// main loop
	const size_t size_vec = size - 16;
	size_t b, i = 0, j = 0;
	size_t o = __sync_fetch_and_add(counter, 1);
	assert(o <= block_limit);
	o *= block_size;
	__mmask16 k = _mm512_kxnor(k, k);
	__m512i key, val, off;
	if (size >= 16) do {
		// replace invalid keys & payloads
		key = _mm512_mask_loadunpacklo_epi32(key, k, &keys[i]);
		key = _mm512_mask_loadunpackhi_epi32(key, k, &keys[i + 16]);
		val = _mm512_mask_loadunpacklo_epi32(val, k, &vals[i]);
		val = _mm512_mask_loadunpackhi_epi32(val, k, &vals[i + 16]);
		off = _mm512_mask_xor_epi32(off, k, off, off);
		i += _mm_countbits_64(_mm512_kconcatlo_64(blend_0000, k));
		// hash keys using either 1st or 2nd function
		__m512i hash = _mm512_mullo_epi32(key, mask_hash_factor);
		__m512i thread = _mm512_mullo_epi32(key, mask_thread_factor);
		hash = _mm512_mulhi_epu32(hash, mask_buckets);
		thread = _mm512_mulhi_epu32(thread, mask_threads);
		hash = _mm512_add_epi32(hash, off);
		k = _mm512_cmpge_epu32_mask(hash, mask_buckets);
		hash = _mm512_mask_sub_epi32(hash, k, hash, mask_buckets);
		hash = _mm512_fmadd_epi32(thread, mask_buckets, hash);
		// load keys from table and update offsets
		__m512i lo = _mm512_i32logather_epi64(hash, table, 8);
		hash = _mm512_permute4f128_epi32(hash, _MM_PERM_BADC);
		__m512i hi = _mm512_i32logather_epi64(hash, table, 8);
		off = _mm512_add_epi32(off, mask_1);
		// split keys and payloads
		__m512i tab_key = _mm512_mask_blend_epi32(blend_AAAA, lo, _mm512_swizzle_epi32(hi, _MM_SWIZ_REG_CDAB));
		__m512i tab_val = _mm512_mask_blend_epi32(blend_5555, hi, _mm512_swizzle_epi32(lo, _MM_SWIZ_REG_CDAB));
		tab_key = _mm512_permutevar_epi32(mask_unpack, tab_key);
		tab_val = _mm512_permutevar_epi32(mask_unpack, tab_val);
		// compare
		__mmask16 m = _mm512_cmpeq_epi32_mask(tab_key, key);
		k = _mm512_cmpeq_epi32_mask(tab_key, mask_empty);
#ifdef _UNIQUE
		k = _mm512_kor(k, m);
#endif
		// pack store matches
		_mm512_mask_packstorelo_epi32(&keys_buf[j], m, key);
		_mm512_mask_packstorehi_epi32(&keys_buf[j + 16], m, key);
		_mm512_mask_packstorelo_epi32(&vals_buf[j], m, val);
		_mm512_mask_packstorehi_epi32(&vals_buf[j + 16], m, val);
		_mm512_mask_packstorelo_epi32(&tabs_buf[j], m, tab_val);
		_mm512_mask_packstorehi_epi32(&tabs_buf[j + 16], m, tab_val);
		j += _mm_countbits_64(_mm512_kconcatlo_64(blend_0000, m));
		if (j >= buffer_size) {
			j -= buffer_size;
			for (b = 0 ; b != buffer_size ; b += 16, o += 16) {
				__m512 x = _mm512_load_ps(&keys_buf[b]);
				__m512 y = _mm512_load_ps(&vals_buf[b]);
				__m512 z = _mm512_load_ps(&tabs_buf[b]);
				_mm512_storenrngo_ps(&keys_out[o], x);
				_mm512_storenrngo_ps(&vals_out[o], y);
				_mm512_storenrngo_ps(&tabs_out[o], z);
			}
			__m512 x = _mm512_load_ps(&keys_buf[b]);
			__m512 y = _mm512_load_ps(&vals_buf[b]);
			__m512 z = _mm512_load_ps(&tabs_buf[b]);
			_mm512_store_ps(keys_buf, x);
			_mm512_store_ps(vals_buf, y);
			_mm512_store_ps(tabs_buf, z);
			if ((o & (block_size - 1)) == 0) {
				o = __sync_fetch_and_add(counter, 1);
				assert(o <= block_limit);
				o *= block_size;
			}
		}
	} while (i <= size_vec);
	// flush last items
	for (b = 0 ; b != j ; ++b, ++o) {
		keys_out[o] = keys_buf[b];
		vals_out[o] = vals_buf[b];
		tabs_out[o] = tabs_buf[b];
	}
	// save last items
	uint32_t keys_last[32];
	uint32_t vals_last[32];
	uint32_t offs_last[32];
	k = _mm512_knot(k);
	_mm512_mask_packstorelo_epi32(&keys_last[0],  k, key);
	_mm512_mask_packstorehi_epi32(&keys_last[16], k, key);
	_mm512_mask_packstorelo_epi32(&vals_last[0],  k, val);
	_mm512_mask_packstorehi_epi32(&vals_last[16], k, val);
	_mm512_mask_packstorelo_epi32(&offs_last[0],  k, off);
	_mm512_mask_packstorehi_epi32(&offs_last[16], k, off);
	j = _mm_countbits_64(_mm512_kconcatlo_64(blend_0000, k));
	for (; i != size ; ++i, ++j) {
		keys_last[j] = keys[i];
		vals_last[j] = vals[i];
		offs_last[j] = 0;
	}
	// process last items in scalar code
	for (i = 0 ; i != j ; ++i) {
		uint32_t key = keys_last[i];
		uint32_t val = vals_last[i];
		size_t t = (uint32_t) (key * thread_factor);
		size_t h = (uint32_t) (key * hash_factor);
		t = (t * threads) >> 32;
		h = (h * buckets) >> 32;
		h += offs_last[i];
		if (h >= buckets)
			h -= buckets;
		const uint64_t *sub_table = &table[t * buckets];
		uint64_t tab = sub_table[h];
		while (empty != (uint32_t) tab) {
			if (key == (uint32_t) tab) {
				tabs_out[o] = tab >> 32;
				vals_out[o] = val;
				keys_out[o++] = key;
				if ((o & (block_size - 1)) == 0) {
					o = __sync_fetch_and_add(counter, 1);
					assert(o <= block_limit);
					o *= block_size;
				}
#ifdef _UNIQUE
				break;
#endif
			}
			if (++h == buckets) h = 0;
			tab = sub_table[h];
		}
	}
	return o;
}

#else

void build(const uint32_t *keys, const uint32_t *vals, size_t size,
           uint64_t *table, size_t buckets, uint32_t factor, uint32_t empty)
{
	size_t i;
	for (i = 0 ; i != buckets ; ++i)
		table[i] = empty;
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		uint64_t pair = vals[i];
		pair = (pair << 32) | key;
		size_t h = (uint32_t) (key * factor);
		h = (h * buckets) >> 32;
		while (empty != (uint32_t) table[h])
			if (++h == buckets) h = 0;
		table[h] = pair;
	}
}

size_t probe(const uint32_t *keys, const uint32_t *vals, size_t size,
             const uint64_t *table, size_t buckets, size_t threads,
             uint32_t thread_factor, uint32_t hash_factor, uint32_t empty,
             uint32_t *keys_out, uint32_t *vals_out, uint32_t *tabs_out,
             size_t block_size, size_t block_limit, volatile size_t *counter)
{
	size_t i, o = __sync_fetch_and_add(counter, 1);
	assert(o <= block_limit);
	o *= block_size;
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		uint32_t val = vals[i];
		size_t t = (uint32_t) (key * thread_factor);
		size_t h = (uint32_t) (key * hash_factor);
		t = (t * threads) >> 32;
		h = (h * buckets) >> 32;
		const uint64_t *sub_table = &table[t * buckets];
		uint64_t tab = sub_table[h];
		while (empty != (uint32_t) tab) {
			if (key == (uint32_t) tab) {
				tabs_out[o] = tab >> 32;
				vals_out[o] = val;
				keys_out[o++] = key;
				if ((o & (block_size - 1)) == 0) {
					o = __sync_fetch_and_add(counter, 1);
					assert(o <= block_limit);
					o *= block_size;
				}
#ifdef _UNIQUE
				break;
#endif
			}
			if (++h == buckets) h = 0;
			tab = sub_table[h];
		}
	}
	return o;
}

#endif

#ifndef _NO_VECTOR_PARTITIONING

void histogram(const uint32_t *keys, size_t size, uint32_t *counts,
               uint32_t factor, size_t partitions)
{
	// partition vector space
	uint32_t parts_space[31];
	uint32_t *parts = align(parts_space);
	// create masks
	__mmask16 blend_0 = _mm512_int2mask(0);
	__m512i mask_0 = _mm512_set1_epi32(0);
	__m512i mask_1 = _mm512_set1_epi32(1);
	__m512i mask_16 = _mm512_set1_epi32(16);
	__m512i mask_255 = _mm512_set1_epi32(255);
	__m512i mask_factor = _mm512_set1_epi32(factor);
	__m512i mask_partitions = _mm512_set1_epi32(partitions);
	__m512i mask_lanes = _mm512_set_epi32(15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0);
	// reset counts
	size_t p, partitions_x16 = partitions << 4;
	uint8_t all_counts_space[partitions_x16 + 127];
	uint8_t *all_counts = align(all_counts_space);
	for (p = 0 ; p < partitions_x16 ; p += 64)
		_mm512_store_epi32(&all_counts[p], mask_0);
	for (p = 0 ; p != partitions ; ++p)
		counts[p] = 0;
	// before alignment
	const uint32_t *keys_end = &keys[size];
	const uint32_t *keys_aligned = align(keys);
	while (keys != keys_end && keys != keys_aligned) {
		uint32_t key = *keys++;
		p = (uint32_t) (key * factor);
		p = (p * partitions) >> 32;
		counts[p]++;
	}
	// aligned
	keys_aligned = &keys[(keys_end - keys) & ~15];
	while (keys != keys_aligned) {
		__m512i key = _mm512_load_epi32(keys);
		keys += 16;
		__m512i part = _mm512_mullo_epi32(key, mask_factor);
		part = _mm512_mulhi_epu32(part, mask_partitions);
		__m512i part_lanes = _mm512_fmadd_epi32(part, mask_16, mask_lanes);
		__m512i count = _mm512_i32extgather_epi32(part_lanes, all_counts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__mmask16 k = _mm512_cmpeq_epi32_mask(count, mask_255);
		count = _mm512_add_epi32(count, mask_1);
		count = _mm512_and_epi32(count, mask_255);
		_mm512_i32extscatter_epi32(all_counts, part_lanes, count, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		if (!_mm512_kortestz(k, k)) {
			_mm512_store_epi32(parts, part);
			size_t mask = _mm512_kconcatlo_64(blend_0, k);
			size_t b = _mm_tzcnt_64(mask);
			do {
				p = parts[b];
				counts[p] += 256;
				b = _mm_tzcnti_64(b, mask);
			} while (b != 64);
		}
	}
	// after alignment
	while (keys != keys_end) {
		uint32_t key = *keys++;
		p = (uint32_t) (key * factor);
		p = (p * partitions) >> 32;
		counts[p]++;
	}
	// merge counts
	for (p = 0 ; p != partitions ; ++p) {
		__m512i sum = _mm512_extload_epi32(&all_counts[p << 4], _MM_UPCONV_EPI32_UINT8, _MM_BROADCAST32_NONE, _MM_HINT_NONE);
		counts[p] += _mm512_reduce_add_epi32(sum);
	}
#ifdef BG
	size_t i;
	for (i = p = 0 ; p != partitions ; ++p)
		i += counts[p];
	assert(i == size);
#endif
}

void partition_global(const uint32_t *keys, const uint32_t *vals, size_t size,
                      uint32_t *offsets, uint64_t *buffers, uint32_t *keys_out,
                      uint32_t *vals_out, uint32_t factor, size_t partitions)
{
	size_t i, p;
	assert(partitions > 1);
	assert(keys_out == align(keys_out));
	assert(vals_out == align(vals_out));
	// conflict space
	uint8_t conflicts_space[partitions + 63];
	uint8_t *conflicts = align(conflicts_space);
	// partition vector space
	uint32_t parts_space[31];
	uint32_t *parts = align(parts_space);
	// generate masks
	__m512i mask_factor = _mm512_set1_epi32(factor);
	__m512i mask_partitions = _mm512_set1_epi32(partitions);
	__m512i mask_pack = _mm512_set_epi32(15, 7, 14, 6, 13, 5, 12, 4, 11, 3, 10, 2, 9, 1, 8, 0);
	__m512i mask_unpack = _mm512_set_epi32(15, 13, 11, 9, 7, 5, 3, 1, 14, 12, 10, 8, 6, 4, 2, 0);
	__m512i mask_16 = _mm512_set1_epi32(16);
	__m512i mask_15 = _mm512_set1_epi32(15);
	__m512i mask_1 = _mm512_set1_epi32(1);
	__mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	__mmask16 blend_5555 = _mm512_int2mask(0x5555);
	__mmask16 blend_0000 = _mm512_int2mask(0x0000);
	// unaligned loop
	size_t size_minus_16 = size - 16;
	__mmask16 k = _mm512_kxnor(k, k);
	__m512i key, val;
	i = 0;
	if (size >= 16) do {
		// load new keys and payloads
		key = _mm512_mask_loadunpacklo_epi32(key, k, &keys[i]);
		key = _mm512_mask_loadunpackhi_epi32(key, k, &keys[i + 16]);
		val = _mm512_mask_loadunpacklo_epi32(val, k, &vals[i]);
		val = _mm512_mask_loadunpackhi_epi32(val, k, &vals[i + 16]);
		// hash keys
		__m512i part = _mm512_mullo_epi32(key, mask_factor);
		part = _mm512_mulhi_epu32(part, mask_partitions);
		// detect conflicts
		_mm512_i32extscatter_epi32(conflicts, part, mask_pack, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i back = _mm512_i32extgather_epi32(part, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		size_t c = _mm512_kconcatlo_64(blend_0000, k);
		k = _mm512_cmpeq_epi32_mask(back, mask_pack);
		// split the data into low and high part
		__m512i key_tmp = _mm512_permutevar_epi32(mask_pack, key);
		__m512i val_tmp = _mm512_permutevar_epi32(mask_pack, val);
		__m512i lo = _mm512_mask_blend_epi32(blend_AAAA, key_tmp, _mm512_swizzle_epi32(val_tmp, _MM_SWIZ_REG_CDAB));
		__m512i hi = _mm512_mask_blend_epi32(blend_5555, val_tmp, _mm512_swizzle_epi32(key_tmp, _MM_SWIZ_REG_CDAB));
		// update offsets and detect conflits
		__m512i offset = _mm512_mask_i32gather_epi32(offset, k, part, offsets, 4);
		__m512i offset_plus_1 = _mm512_add_epi32(offset, mask_1);
		_mm512_mask_i32scatter_epi32(offsets, k, part, offset_plus_1, 4);
		// compute block offsets
		offset = _mm512_and_epi32(offset, mask_15);
		__mmask16 eq = _mm512_mask_cmpeq_epi32_mask(k, offset, mask_15);
		offset = _mm512_fmadd_epi32(part, mask_16, offset);
		// write interleaved keys and payloads to buffers
		_mm512_mask_i32loscatter_epi64(buffers, k, offset, lo, 8);
		offset = _mm512_permute4f128_epi32(offset, _MM_PERM_BCDC);
		__mmask16 r_k = _mm512_kmerge2l1h(k, k);
		_mm512_mask_i32loscatter_epi64(buffers, r_k, offset, hi, 8);
		c = _mm_countbits_64(c);
		// flush full blocks (taken ~ 65%)
		if (!_mm512_kortestz(eq, eq)) {
			_mm512_store_epi32(parts, part);
			size_t mask = _mm512_kconcatlo_64(blend_0000, eq);
			size_t b = _mm_tzcnt_64(mask);
			do {
				size_t l = parts[b];
				size_t o = offsets[l];
				l <<= 4;
				lo = _mm512_load_epi64(&buffers[l + 0]);
				hi = _mm512_load_epi64(&buffers[l + 8]);
				key_tmp = _mm512_mask_blend_epi32(blend_AAAA, lo, _mm512_swizzle_epi32(hi, _MM_SWIZ_REG_CDAB));
				val_tmp = _mm512_mask_blend_epi32(blend_5555, hi, _mm512_swizzle_epi32(lo, _MM_SWIZ_REG_CDAB));
				key_tmp = _mm512_permutevar_epi32(mask_unpack, key_tmp);
				val_tmp = _mm512_permutevar_epi32(mask_unpack, val_tmp);
				_mm512_storenrngo_ps(&keys_out[o - 16], _mm512_castsi512_ps(key_tmp));
				_mm512_storenrngo_ps(&vals_out[o - 16], _mm512_castsi512_ps(val_tmp));
				b = _mm_tzcnti_64(b, mask);
			} while (b != 64);
		}
		i += c;
	} while (i <= size_minus_16);
	// store last items in stack
	uint32_t tmp_key[32];
	uint32_t tmp_rid[32];
	k = _mm512_knot(k);
	_mm512_mask_packstorelo_epi32(&tmp_key[0],  k, key);
	_mm512_mask_packstorehi_epi32(&tmp_key[16], k, key);
	_mm512_mask_packstorelo_epi32(&tmp_rid[0],  k, val);
	_mm512_mask_packstorehi_epi32(&tmp_rid[16], k, val);
	size_t c = _mm_countbits_64(_mm512_kconcatlo_64(blend_0000, k));
	assert(size - i + c <= 32);
	for (; i != size ; ++i, ++c) {
		tmp_key[c] = keys[i];
		tmp_rid[c] = vals[i];
	}
	// partition last items using scalar code
	for (i = 0 ; i != c ; ++i) {
		uint64_t kv = tmp_rid[i];
		kv = (kv << 32) | tmp_key[i];
		p = (uint32_t) (factor * (uint32_t) kv);
		p = (p * partitions) >> 32;
		uint64_t *buf = &buffers[p << 4];
		size_t o = offsets[p]++;
		size_t b = o & 15;
		buf[b] = kv;
		if (b == 15) {
			__m512i lo_out = _mm512_load_epi64(&buf[0]);
			__m512i hi_out = _mm512_load_epi64(&buf[8]);
			key = _mm512_mask_blend_epi32(blend_AAAA, lo_out, _mm512_swizzle_epi32(hi_out, _MM_SWIZ_REG_CDAB));
			val = _mm512_mask_blend_epi32(blend_5555, hi_out, _mm512_swizzle_epi32(lo_out, _MM_SWIZ_REG_CDAB));
			key = _mm512_permutevar_epi32(mask_unpack, key);
			val = _mm512_permutevar_epi32(mask_unpack, val);
			_mm512_storenrngo_ps(&keys_out[o - 15], _mm512_castsi512_ps(key));
			_mm512_storenrngo_ps(&vals_out[o - 15], _mm512_castsi512_ps(val));
		}
	}
}

void copy(uint32_t *dst, const uint32_t *src, size_t size)
{
	uint32_t *dst_end = &dst[size];
	uint32_t *dst_aligned = align(dst);
	while (dst != dst_end && dst != dst_aligned)
		*dst++ = *src++;
	dst_aligned = &dst[(dst_end - dst) & ~15];
	if (src == align(src))
		while (dst != dst_aligned) {
			__m512 x = _mm512_load_ps(src);
			_mm512_storenrngo_ps(dst, x);
			src += 16;
			dst += 16;
		}
	else
		while (dst != dst_aligned) {
			__m512 x;
			x = _mm512_loadunpacklo_ps(x, src);
			src += 16;
			x = _mm512_loadunpackhi_ps(x, src);
			_mm512_storenrngo_ps(dst, x);
			dst += 16;
		}
	while (dst != dst_end)
		*dst++ = *src++;
}

size_t interleave(uint32_t **counts, uint32_t *offsets, uint32_t *counts_sum,
                  size_t partitions, size_t thread, size_t threads)
{
	size_t i = 0, p = 0;
	assert(offsets == align(offsets));
	assert(counts_sum == align(counts_sum));
	while (p != partitions) {
		size_t s = partitions - p;
		if (s > 16) s = 16;
		__mmask16 k = _mm512_int2mask((1 << s) - 1);
		__m512i sum = _mm512_xor_epi32(sum, sum);
		size_t t = 0;
		if (thread) do {
			__m512i cur = _mm512_mask_load_epi32(cur, k, &counts[t][p]);
			sum = _mm512_add_epi32(sum, cur);
		} while (++t != thread);
		_mm512_mask_store_epi32(&offsets[p], k, sum);
		do {
			__m512i cur = _mm512_mask_load_epi32(cur, k, &counts[t][p]);
			sum = _mm512_add_epi32(sum, cur);
		} while (++t != threads);
		_mm512_mask_store_epi32(&counts_sum[p], k, sum);
		do {
			offsets[p] += i;
			i += counts_sum[p++];
		} while (--s);
	}
	return i;
}

#else

void histogram(const uint32_t *keys, size_t size, uint32_t *counts,
               uint32_t factor, size_t partitions)
{
	size_t i, p;
	for (p = 0 ; p != partitions ; ++p)
		counts[p] = 0;
	for (i = 0 ; i != size ; ++i) {
		p = (uint32_t) (keys[i] * factor);
		p = (p * partitions) >> 32;
		counts[p]++;
	}
}

void partition_global(const uint32_t *keys, const uint32_t *vals, size_t size,
                      uint32_t *offsets, uint64_t *buffers, uint32_t *keys_out,
                      uint32_t *vals_out, uint32_t factor, size_t partitions)
{
	size_t i, p;
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		uint64_t key_val = vals[i];
		key_val = (key_val << 32) | key;
		p = (uint32_t) (key * factor);
		p = (p * partitions) >> 32;
		size_t o = offsets[p]++;
		size_t b = o & 15;
		uint64_t *buf = &buffers[p << 4];
		buf[b] = key_val;
		if (b == 15) {
			uint32_t *k_out = &keys_out[o - 15];
			uint32_t *v_out = &vals_out[o - 15];
			for (b = 0 ; b != 16 ; ++b) {
				key_val = buf[b];
				k_out[b] = key_val;
				v_out[b] = key_val >> 32;
			}
		}
	}
}

void copy(uint32_t *dst, const uint32_t *src, size_t size)
{
	size_t i;
	for (i = 0 ; i != size ; ++i)
		dst[i] = src[i];
}

size_t interleave(uint32_t **counts, uint32_t *offsets, uint32_t *counts_sum,
                  size_t partitions, size_t thread, size_t threads)
{
	size_t i, p, t;
	for (i = p = 0 ; p != partitions ; ++p) {
		size_t s = i;
		for (t = 0 ; t != thread ; ++t)
			i += counts[t][p];
		offsets[p] = i;
		for (; t != threads ; ++t)
			i += counts[t][p];
		counts_sum[p] = i - s;
	}
	return i;
}

#endif

void flush(const uint32_t *counts, const uint32_t *offsets, const uint64_t *buffers,
           uint32_t *keys_out, uint32_t *vals_out, size_t partitions)
{
	size_t p;
	for (p = 0 ; p != partitions ; ++p) {
		const uint64_t *buf = &buffers[p << 4];
		size_t off = offsets[p];
		size_t cnt = counts[p];
		size_t end = off & 15;
		size_t beg = end > cnt ? end - cnt : 0;
		off -= end - beg;
		while (beg != end) {
			keys_out[off] = buf[beg];
			vals_out[off] = buf[beg] >> 32;
			off++;
			beg++;
		}
	}
}

typedef struct {
	size_t beg;
	size_t end;
} size_pair_t;

int pair_cmp(const void *x, const void *y)
{
	size_t a = ((size_pair_t*) x)->beg;
	size_t b = ((size_pair_t*) y)->beg;
	return a < b ? -1 : a > b ? 1 : 0;
}

size_t close_gaps(uint32_t *keys, uint32_t *vals, uint32_t *tabs,
                  const size_t *offsets, size_t count,
                  size_t block_size, volatile size_t *counter)
{
	assert(power_of_2(block_size));
	size_pair_t *holes = malloc(count * sizeof(size_pair_t));
	size_t i, c = 0, l = 0, h = count - 1;
	for (i = 0 ; i != count ; ++i) {
		holes[i].beg = offsets[i];
		holes[i].end = (offsets[i] & ~(block_size - 1)) + block_size;
	}
	qsort(holes, count, sizeof(size_pair_t), pair_cmp);
	size_t src = holes[h].end;
	i = __sync_fetch_and_add(counter, 1);
	while (l <= h) {
		size_t fill = src - holes[h].end;
		if (fill == 0) {
			src = holes[h].beg;
			if (!h--) break;
			continue;
		}
		size_t hole = holes[l].end - holes[l].beg;
		if (hole == 0) {
			l++;
			continue;
		}
		size_t cnt = fill < hole ? fill : hole;
		size_t dst = holes[l].beg;
		holes[l].beg += cnt;
		src -= cnt;
		if (c++ == i) {
			copy(&keys[dst], &keys[src], cnt);
			copy(&vals[dst], &vals[src], cnt);
			copy(&tabs[dst], &tabs[src], cnt);
			i = __sync_fetch_and_add(counter, 1);
		}
	}
	free(holes);
	return src;
}

size_t thread_beg(size_t size, size_t alignment, size_t thread, size_t threads)
{
	assert(power_of_2(alignment));
	size_t part = (size / threads) & ~(alignment - 1);
	return part * thread;
}

size_t thread_end(size_t size, size_t alignment, size_t thread, size_t threads)
{
	assert(power_of_2(alignment));
	size_t part = (size / threads) & ~(alignment - 1);
	if (thread + 1 == threads) return size;
	return part * (thread + 1);
}

void swap(uint32_t **x, uint32_t **y)
{
	uint32_t *t = *x; *x = *y; *y = t;
}

size_t max(size_t x, size_t y)
{
	return x > y ? x : y;
}

size_t min(size_t x, size_t y)
{
	return x < y ? x : y;
}

size_t binary_search(const size_t *array, size_t size, size_t value)
{
	size_t lo = 0;
	size_t hi = size;
	do {
		size_t mid = (lo + hi) >> 1;
		if (value > array[mid])
			lo = mid + 1;
		else
			hi = mid;
	} while (lo < hi);
	return array[lo];
}

void shuffle(uint32_t *data, size_t size, rand32_t *gen)
{
	size_t i, j;
	for (i = 0 ; i != size ; ++i) {
		j = rand32_next(gen);
		j *= size - i;
		j >>= 32;
		j += i;
		uint32_t t = data[i];
		data[i] = data[j];
		data[j] = t;
	}
}

void unique(uint32_t *keys, size_t size, volatile uint32_t *table, size_t buckets,
            uint32_t factor, uint32_t empty, rand32_t *gen)
{
	assert(prime(buckets));
	size_t i = 0;
	while (i != size) {
		uint32_t key;
		do {
			key = rand32_next(gen);
		} while (key == empty);
		size_t h = (uint32_t) (key * factor);
		h = (h * buckets) >> 32;
		uint32_t tab = table[h];
		while (tab != key) {
			if (tab == empty) {
				tab = __sync_val_compare_and_swap(&table[h], empty, key);
				if (tab == empty) {
					keys[i++] = key;
					break;
				}
				if (tab == key) break;
			}
			if (++h == buckets) h = 0;
			tab = table[h];
		}
	}
}

typedef struct {
	pthread_t id;
	int seed;
	int thread;
	int threads;
	uint32_t thread_factor;
	uint32_t inner_factor;
	uint32_t outer_factor;
	size_t inner_tuples;
	size_t outer_tuples;
	size_t join_tuples;
	size_t inner_distinct;
	size_t outer_distinct;
	size_t join_distinct;
	uint32_t *inner_keys[2];
	uint32_t *inner_vals[2];
	uint32_t *outer_keys;
	uint32_t *outer_vals;
	uint32_t *join_keys;
	uint32_t *join_outer_vals;
	uint32_t *join_inner_vals;
	uint32_t **inner_counts;
	uint64_t *inner_checksum;
	size_t *inner_buckets;
	size_t *final_offsets;
	uint64_t real_time;
	uint64_t thread_time;
	uint64_t *hash_table;
	size_t hash_buckets_total;
	double hash_table_load;
	uint32_t hash_factor;
	uint32_t *unique;
	uint32_t *unique_table;
	size_t unique_buckets;
	uint32_t unique_factor;
	size_t block_size;
	size_t block_limit;
	size_t buffer_size;
	uint64_t times[4];
	volatile size_t *counters;
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	pthread_barrier_t *barrier = d->barrier;
	bind_thread(d->thread, d->threads);
	uint32_t *inner_keys_in  = d->inner_keys[0];
	uint32_t *inner_keys_out = d->inner_keys[1];
	uint32_t *inner_vals_in  = d->inner_vals[0];
	uint32_t *inner_vals_out = d->inner_vals[1];
	uint32_t *outer_keys = d->outer_keys;
	uint32_t *outer_vals  = d->outer_vals;
	size_t i, o, j, u, t, f, h;
	size_t thread  = d->thread;
	size_t threads = d->threads;
	rand32_t *gen = rand32_init(d->seed);
	// generate unique items
	size_t join_distinct = d->join_distinct;
	size_t inner_distinct = d->inner_distinct;
	size_t outer_distinct = d->outer_distinct;
	size_t distinct = inner_distinct + outer_distinct - join_distinct;
	size_t distinct_beg = thread_beg(distinct, 1, thread, threads);
	size_t distinct_end = thread_end(distinct, 1, thread, threads);
	unique(&d->unique[distinct_beg], distinct_end - distinct_beg,
	       d->unique_table, d->unique_buckets, d->unique_factor, 0, gen);
	// generate keys from unique items
	pthread_barrier_wait(barrier++);
	if (thread == 0) free((void*) d->unique_table);
	uint32_t *inner_unique = d->unique;
	size_t inner_beg = thread_beg(d->inner_tuples, 16, thread, threads);
	size_t inner_end = thread_end(d->inner_tuples, 16, thread, threads);
	size_t inner_distinct_beg = thread_beg(inner_distinct, 16, thread, threads);
	size_t inner_distinct_end = thread_end(inner_distinct, 16, thread, threads);
	uint64_t inner_checksum = 0;
	u = inner_distinct_beg;
	for (i = inner_beg ; i != inner_end ; ++i) {
		if (u != inner_distinct_end)
			inner_keys_in[i] = inner_unique[u++];
		else {
			uint64_t r = rand32_next(gen);
			r = (r * inner_distinct) >> 32;
			inner_keys_in[i] = inner_unique[r];
		}
		assert(inner_keys_in[i] != 0);
		inner_checksum += inner_keys_in[i];
	}
	assert(u == inner_distinct_end);
	uint32_t *outer_unique = &d->unique[inner_distinct - join_distinct];
	size_t outer_beg = thread_beg(d->outer_tuples, 16, thread, threads);
	size_t outer_end = thread_end(d->outer_tuples, 16, thread, threads);
	size_t outer_distinct_beg = thread_beg(outer_distinct, 16, thread, threads);
	size_t outer_distinct_end = thread_end(outer_distinct, 16, thread, threads);
	u = outer_distinct_beg;
	for (o = outer_beg ; o != outer_end ; ++o) {
		if (u != outer_distinct_end)
			outer_keys[o] = outer_unique[u++];
		else {
			uint64_t r = rand32_next(gen);
			r = (r * outer_distinct) >> 32;
			outer_keys[o] = outer_unique[r];
		}
		assert(outer_keys[o] != 0);
	}
	assert(u == outer_distinct_end);
	pthread_barrier_wait(barrier++);
	if (thread == 0) {
		free(d->unique);
		fprintf(stderr, "Shuffling ... ");
		shuffle(inner_keys_in, d->inner_tuples, gen);
		shuffle(outer_keys, d->outer_tuples, gen);
		fprintf(stderr, "done!\n");
	}
	// generate payloads and outputs
	pthread_barrier_wait(barrier++);
	uint32_t inner_factor = d->inner_factor;
	for (i = inner_beg ; i != inner_end ; ++i) {
		inner_vals_in[i] = inner_keys_in[i] * inner_factor;
		inner_keys_out[i] = 0xCAFEBABE;
		inner_vals_out[i] = 0xDEADBEEF;
	}
	uint32_t outer_factor = d->outer_factor;
	for (o = outer_beg ; o != outer_end ; ++o)
		outer_vals[o] = outer_keys[o] * outer_factor;
	size_t join_beg = thread_beg(d->block_limit * d->block_size, 1, thread, threads);
	size_t join_end = thread_end(d->block_limit * d->block_size, 1, thread, threads);
	uint32_t *keys_out = d->join_keys;
	uint32_t *vals_out = d->join_outer_vals;
	uint32_t *tabs_out = d->join_inner_vals;
	for (j = join_beg ; j != join_end ; ++j) {
		keys_out[j] = 0xCAFEBABE;
		vals_out[j] = 0xDEADBEEF;
		tabs_out[j] = 0x12345678;
	}
	// initialize table
	size_t table_beg = thread_beg(d->hash_buckets_total, 1, thread, threads);
	size_t table_end = thread_end(d->hash_buckets_total, 1, thread, threads);
	uint64_t *table = (uint64_t*) d->hash_table;
	for (t = table_beg ; t != table_end ; ++t)
		table[t] = 0;
	d->inner_checksum[thread] = inner_checksum;
	pthread_barrier_wait(barrier++);
	uint64_t inner_sum = 0;
	for (t = 0 ; t != threads ; ++t)
		inner_sum += d->inner_checksum[t];
	if (thread == 0) fprintf(stderr, "Data generated!\n");
	// start timinig
	pthread_barrier_wait(barrier++);
	uint64_t tt = thread_time();
	uint64_t rt = real_time();
	// shared partitioning
	d->inner_buckets[thread] = (inner_end - inner_beg) / d->hash_table_load;
	if (threads > 1) {
		uint64_t unaligned_buffers_space[threads * 32 + 7];
		uint64_t *buffers_space = align(unaligned_buffers_space);
		uint64_t *inner_buffers = &buffers_space[threads *  0];
		size_t threads_aligned = (threads + 15) & ~15;
		uint32_t unaligned_counts_space[threads_aligned * 6 + 15];
		uint32_t *counts_space = align(unaligned_counts_space);
		uint32_t *inner_offsets = &counts_space[threads_aligned * 0];
		uint32_t *inner_counts = &counts_space[threads_aligned * 2];
		d->inner_counts[thread] = inner_counts;
		uint32_t factor = d->thread_factor;
		histogram(&inner_keys_in[inner_beg], inner_end - inner_beg, inner_counts, factor, threads);
		pthread_barrier_wait(barrier++);
		inner_counts = &counts_space[threads_aligned * 4];
		i = interleave(d->inner_counts, inner_offsets, inner_counts, threads, thread, threads);
		assert(i == d->inner_tuples);
		partition_global(&inner_keys_in[inner_beg], &inner_vals_in[inner_beg], inner_end - inner_beg,
		                 inner_offsets, inner_buffers, inner_keys_out, inner_vals_out, factor, threads);
		inner_beg = 0;
		for (t = 0 ; t != thread ; ++t)
			inner_beg += inner_counts[t];
		inner_end = inner_beg + inner_counts[t];
		d->inner_buckets[thread] = (inner_end - inner_beg) / (d->hash_table_load + 0.05);
		pthread_barrier_wait(barrier++);
		flush(d->inner_counts[thread], inner_offsets, inner_buffers, inner_keys_out, inner_vals_out, threads);
		pthread_barrier_wait(barrier++);
		swap(&inner_keys_in, &inner_keys_out);
		swap(&inner_vals_in, &inner_vals_out);
	}
#ifdef BG
	pthread_barrier_wait(barrier++);
	inner_checksum = 0;
	for (i = inner_beg ; i != inner_end ; ++i) {
		inner_checksum += inner_keys_in[i];
		assert(inner_vals_in[i] == inner_keys_in[i] * inner_factor);
	}
	d->inner_checksum[thread] = inner_checksum;
	pthread_barrier_wait(barrier++);
	inner_checksum = 0;
	for (t = 0 ; t != threads ; ++t)
		inner_checksum += d->inner_checksum[t];
	assert(inner_checksum == inner_sum);
	fprintf(stderr, "Checksum verified [%ld / %ld]\n", thread + 1, threads);
#endif
	uint64_t ptt = tt;
	uint64_t ntt = thread_time();
	d->times[0] = ntt - ptt;
	// compute hash buckets
	size_t hash_buckets = 0;
	for (t = 0 ; t != threads ; ++t)
		if (d->inner_buckets[t] > hash_buckets)
			hash_buckets = d->inner_buckets[t];
	if (thread == 0 && hash_buckets * threads > d->hash_buckets_total)
		fprintf(stderr, "Buckets overflow: %ld > %ld\n", hash_buckets * threads, d->hash_buckets_total);
	assert(hash_buckets * threads <= d->hash_buckets_total);
	// build hash table
	build(&inner_keys_in[inner_beg],
	      &inner_vals_in[inner_beg],
	      inner_end - inner_beg,
	      &d->hash_table[hash_buckets * thread],
	      hash_buckets,
	      d->hash_factor, 0);
	ptt = ntt;
	ntt = thread_time();
	d->times[1] = ntt - ptt;
	// sync
	pthread_barrier_wait(barrier++);
	// probe hash table
	o = probe(&outer_keys[outer_beg],
	          &outer_vals[outer_beg],
	          outer_end - outer_beg,
	          d->hash_table,
	          hash_buckets,
	          threads,
	          d->thread_factor,
	          d->hash_factor, 0,
	          d->join_keys,
	          d->join_outer_vals,
	          d->join_inner_vals,
		      d->block_size,
		      d->block_limit,
		      &d->counters[0]);
	d->final_offsets[thread] = o;
	ptt = ntt;
	ntt = thread_time();
	d->times[2] = ntt - ptt;
	// sync
	pthread_barrier_wait(barrier++);
	// close gaps
	d->join_tuples = close_gaps(d->join_keys,
	                            d->join_outer_vals,
	                            d->join_inner_vals,
	                            d->final_offsets,
	                            threads, 
	                            d->block_size,
	                            &d->counters[1]);
	assert(d->join_tuples <= d->block_size * d->block_limit);
	ptt = ntt;
	ntt = thread_time();
	d->times[3] = ntt - ptt;
	// finish timinig
	rt = real_time() - rt;
	tt = thread_time() - tt;
	pthread_barrier_wait(barrier++);
	d->real_time = rt;
	d->thread_time = tt;
//	fprintf(stderr, "Checksum verified [%ld / %ld]\n", thread + 1, threads);
	join_beg = thread_beg(d->join_tuples, 1, thread, threads);
	join_end = thread_end(d->join_tuples, 1, thread, threads);
	for (j = join_beg ; j != join_end ; ++j) {
		assert(vals_out[j] == keys_out[j] * outer_factor);
		assert(tabs_out[j] == keys_out[j] * inner_factor);
	}
//	fprintf(stderr, "Join verified [%ld / %ld]\n", thread + 1, threads);
	free(gen);
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	// arguments
	int t, threads = argc > 1 ? atoi(argv[1]) : hardware_threads();
	size_t outer_tuples = argc > 2 ? atoll(argv[2]) : 200 * 1000 * 1000;
	size_t inner_tuples = argc > 3 ? atoll(argv[3]) : 200 * 1000 * 1000;
	size_t outer_distinct = argc > 4 ? atoll(argv[4]) : min(inner_tuples, outer_tuples);
	size_t inner_distinct = argc > 5 ? atoll(argv[5]) : min(inner_tuples, outer_tuples);
	size_t join_distinct  = argc > 6 ? atoll(argv[6]) : min(inner_distinct, outer_distinct);
	double outer_repeats = outer_tuples * 1.0 / outer_distinct;
	double inner_repeats = inner_tuples * 1.0 / inner_distinct;
	size_t join_tuples = outer_repeats * inner_repeats * join_distinct;
	// other parameters
	double hash_table_load = 0.25;
	size_t block_size = 256 * 256;
	// compute hash buckets
	size_t hash_buckets = inner_tuples / hash_table_load;
	// print info
	assert(inner_distinct <= inner_tuples);
	assert(outer_distinct <= outer_tuples);
	assert(join_distinct <= min(inner_distinct, outer_distinct));
	assert(threads > 0 && threads <= hardware_threads());
	assert(power_of_2(block_size));
#ifdef _UNIQUE
	fprintf(stderr, "Enforcing unique keys\n");
	assert(inner_tuples == inner_distinct);
#endif
#ifdef _NO_VECTOR_PARTITIONING
	fprintf(stderr, "Vectorized partitioning disabled!\n");
#endif
#ifdef _NO_VECTOR_HASHING
	fprintf(stderr, "Vectorized hashing disabled!\n");
#endif
	fprintf(stderr, "Threads: %d\n", threads);
	fprintf(stderr, "Inner tuples: %9ld\n", inner_tuples);
	fprintf(stderr, "Outer tuples: %9ld\n", outer_tuples);
	fprintf(stderr, "Inner distinct values: %9ld\n", inner_distinct);
	fprintf(stderr, "Outer distinct values: %9ld\n", outer_distinct);
	fprintf(stderr, "Join  distinct values: %9ld\n", join_distinct);
	fprintf(stderr, "Join tuples (expected): %9ld\n", join_tuples);
	// compute space
	double space = inner_tuples * 16 + outer_tuples * 8 + hash_buckets * 8 + join_tuples * 12;
	space /= 1024 * 1024 * 1024;
	fprintf(stderr, "Space: %.2f GB\n", space);
	assert(space <= 14.5);
	// parameters
	srand(time(NULL));
	uint32_t factors[7];
	for (t = 0 ; t != 7 ; ++t)
		factors[t] = (rand() << 1) | 1;
	// inner side & buffers
	uint32_t *inner_keys_1 = mamalloc(inner_tuples * sizeof(uint32_t));
	uint32_t *inner_keys_2 = mamalloc(inner_tuples * sizeof(uint32_t));
	uint32_t *inner_vals_1 = mamalloc(inner_tuples * sizeof(uint32_t));
	uint32_t *inner_vals_2 = mamalloc(inner_tuples * sizeof(uint32_t));
	// outer side & buffers
	uint32_t *outer_keys = mamalloc(outer_tuples * sizeof(uint32_t));
	uint32_t *outer_vals = mamalloc(outer_tuples * sizeof(uint32_t));
	// hash table
	uint64_t *hash_table = mamalloc(hash_buckets * sizeof(uint64_t));
	// unique table and unique items
	size_t distinct = outer_distinct + inner_distinct - join_distinct;
	size_t unique_buckets = distinct * 2 + 1;
	while (!prime(unique_buckets)) unique_buckets += 2;
	uint32_t *unique = malloc(distinct * sizeof(uint32_t));
	uint32_t *unique_table = calloc(unique_buckets, sizeof(uint32_t));
	// join result
	size_t block_limit = join_tuples * 1.05 / block_size + threads * 2;
	uint32_t *join_keys       = mamalloc(block_limit * block_size * sizeof(uint32_t));
	uint32_t *join_inner_vals = mamalloc(block_limit * block_size * sizeof(uint32_t));
	uint32_t *join_outer_vals = mamalloc(block_limit * block_size * sizeof(uint32_t));
	// run threads
	int b, barriers = 64;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	info_t info[threads];
	size_t final_offsets[threads];
	size_t inner_buckets[threads];
	uint32_t *inner_counts[threads];
	uint64_t inner_checksum[threads];
	volatile size_t counters[2] = {0, 0};
	for (t = 0 ; t != threads ; ++t) {
		info[t].thread = t;
		info[t].threads = threads;
		info[t].seed = rand();
		info[t].join_tuples = join_tuples;
		info[t].block_limit = block_limit;
		info[t].outer_tuples = outer_tuples;
		info[t].inner_tuples = inner_tuples;
		info[t].join_distinct = join_distinct;
		info[t].inner_distinct = inner_distinct;
		info[t].outer_distinct = outer_distinct;
		info[t].inner_keys[0] = inner_keys_1;
		info[t].inner_keys[1] = inner_keys_2;
		info[t].inner_vals[0] = inner_vals_1;
		info[t].inner_vals[1] = inner_vals_2;
		info[t].outer_keys = outer_keys;
		info[t].outer_vals = outer_vals;
		info[t].inner_buckets = inner_buckets;
		info[t].inner_counts = inner_counts;
		info[t].inner_checksum = inner_checksum;
		info[t].final_offsets = final_offsets;
		info[t].join_keys = join_keys;
		info[t].join_inner_vals = join_inner_vals;
		info[t].join_outer_vals = join_outer_vals;
		info[t].unique = unique;
		info[t].unique_table = unique_table;
		info[t].unique_buckets = unique_buckets;
		info[t].unique_factor = factors[0];
		info[t].hash_factor = factors[1];
		info[t].thread_factor = factors[2];
		info[t].inner_factor = factors[3];
		info[t].outer_factor = factors[4];
		info[t].hash_table = hash_table;
		info[t].hash_buckets_total = hash_buckets;
		info[t].hash_table_load = hash_table_load;
		info[t].block_size = block_size;
		info[t].counters = counters;
		info[t].barrier = barrier;
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	join_tuples = info[0].join_tuples;
	fprintf(stderr, "Join tuples (measured): %9ld\n", join_tuples);
	fprintf(stderr, "Join selectivity: %.3f\n", join_tuples * 1.0 / (inner_tuples + outer_tuples));
	// assemble results
	uint64_t min_tt = ~0, max_tt = 0, avg_tt = 0;
	uint64_t min_rt = ~0, max_rt = 0, avg_rt = 0;
	for (t = 0 ; t != threads ; ++t) {
		uint64_t rt = info[t].real_time;
		uint64_t tt = info[t].thread_time;
		max_rt = max(rt, max_rt);
		max_tt = max(tt, max_tt);
		min_rt = min(rt, min_rt);
		min_tt = min(tt, min_tt);
		avg_rt += rt;
		avg_tt += tt;
		assert(info[t].join_tuples == join_tuples);
	}
	avg_tt /= threads;
	avg_rt /= threads;
	double billion_th = 1.0 / 1000000000;
	fprintf(stderr, "Min: %6.4f sec (real)\n", min_rt * billion_th);
	fprintf(stderr, "Avg: %6.4f sec (real)\n", avg_rt * billion_th);
	fprintf(stderr, "Max: %6.4f sec (real)\n", max_rt * billion_th);
	fprintf(stderr, "Min: %6.4f sec (thread)\n", min_tt * billion_th);
	fprintf(stderr, "Avg: %6.4f sec (thread)\n", avg_tt * billion_th);
	fprintf(stderr, "Max: %6.4f sec (thread)\n", max_tt * billion_th);
	size_t p;
	for (p = 0 ; p != 4 ; ++p) {
		uint64_t partial_tt = 0;
		for (t = 0 ; t != threads ; ++t)
			partial_tt += info[t].times[p];
		if (partial_tt == 0) continue;
		partial_tt /= threads;
		double r = partial_tt * 1.0 / avg_tt;
		fprintf(stderr, "Phase %ld: %5.2f%% (%.4f)\n", p + 1, r * 100, r * max_rt * billion_th);
	}
	printf("%d, %.4f\n", threads, max_rt * billion_th);
	free(join_keys);
	free(join_inner_vals);
	free(join_outer_vals);
	free(inner_keys_1);
	free(inner_keys_2);
	free(inner_vals_1);
	free(inner_vals_2);
	free(outer_keys);
	free(outer_vals);
	free(hash_table);
	return EXIT_SUCCESS;
}
