/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <immintrin.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <sched.h>
#include <time.h>

#include "rand.h"


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread_id)
{
	int threads = hardware_threads();
	assert(thread_id >= 0 && thread_id < threads);
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread_id, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *mamalloc(size_t size)
{
	void *ptr = NULL;
	return posix_memalign(&ptr, 64, size) ? NULL : ptr;
}

inline void *align(const void *p)
{
	size_t i = 63 & (size_t) p;
	return (void*) (i ? p + 64 - i : p);
}

void print_vector(const char *m, __m512i x)
{
	uint32_t space[31];
	uint32_t *arr = align(space);
	_mm512_store_epi32(arr, x);
	fprintf(stderr, "%s: [%3u", m, arr[0]);
	size_t i = 1;
	do {
		fprintf(stderr, ", %3u", arr[i]);
	} while (++i != 16);
	fprintf(stderr, "]\n");
}

void histogram_1(const uint32_t *keys, size_t size, uint32_t *counts,
                 uint8_t shift_bits, uint8_t radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	for (p = 0 ; p != partitions ; ++p)
		counts[p] = 0;
	for (i = 0 ; i != size ; ++i) {
		p = keys[i];
		p >>= shift_bits;
		p &= partitions - 1;
		counts[p]++;
	}
	for (i = p = 0 ; p != partitions ; ++p)
		i += counts[p];
	assert(i == size);
}

void histogram_2(const uint32_t *keys, size_t size, uint32_t *counts,
                 uint8_t shift_bits, uint8_t radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	assert(keys == align(keys));
	assert((size & 15) == 0);
	// conflict space
	uint8_t conflicts_space[partitions + 63];
	uint8_t *conflicts = align(conflicts_space);
	// generate masks
	__m512i mask_1 = _mm512_set1_epi32(1);
	__m512i mask_shift = _mm512_set1_epi32(shift_bits);
	__m512i mask_partitions_minus_1 = _mm512_set1_epi32(partitions - 1);
	__m512i mask_reverse = _mm512_set_epi32(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
	// reset counts
	for (p = 0 ; p != partitions ; ++p)
		counts[p] = 0;
	// aligned loop
	for (i = 0 ; i != size ; i += 16) {
		// compute partition
		__m512i key = _mm512_load_epi32(&keys[i]);
		__m512i part = _mm512_srlv_epi32(key, mask_shift);
		part = _mm512_and_epi32(part, mask_partitions_minus_1);
		// detect conflicts
		__m512i part_reverse = _mm512_permutevar_epi32(mask_reverse, part);
		_mm512_i32extscatter_epi32(conflicts, part_reverse, mask_reverse, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i back = _mm512_i32extgather_epi32(part_reverse, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i update = mask_1;
		__mmask16 k = _mm512_cmpneq_epi32_mask(back, mask_reverse);
		if (!_mm512_kortestz(k, k)) do {
			update = _mm512_mask_add_epi32(update, k, update, mask_1);
			_mm512_mask_i32extscatter_epi32(conflicts, k, part_reverse, mask_reverse, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
			back = _mm512_mask_i32extgather_epi32(back, k, part_reverse, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
			k = _mm512_cmpneq_epi32_mask(back, mask_reverse);
		} while (!_mm512_kortestz(k, k));
		// update counts
		__m512i offset = _mm512_i32gather_epi32(part, counts, 4);
		update = _mm512_permutevar_epi32(mask_reverse, update);
		offset = _mm512_add_epi32(offset, update);
		_mm512_i32scatter_epi32(counts, part, offset, 4);
	}
}

void histogram_3(const uint32_t *keys, size_t size, uint32_t *counts,
                 uint8_t shift_bits, uint8_t radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	assert(keys == align(keys));
	assert((size & 15) == 0);
	// create masks
	__m512i mask_0 = _mm512_set1_epi32(0);
	__m512i mask_1 = _mm512_set1_epi32(1);
	__m512i mask_16 = _mm512_set1_epi32(16);
	__m512i mask_shift = _mm512_set1_epi32(shift_bits);
	__m512i mask_partitions_minus_1 = _mm512_set1_epi32(partitions - 1);
	__m512i mask_lanes = _mm512_set_epi32(15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0);
	// reset counts
	size_t partitions_x16 = partitions << 4;
	uint32_t all_counts_space[partitions_x16 + 15];
	uint32_t *all_counts = align(all_counts_space);
	for (p = 0 ; p != partitions_x16 ; p += 16)
		_mm512_store_epi32(&all_counts[p], mask_0);
	// main loop
	for (i = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32(&keys[i]);
		__m512i part = _mm512_srlv_epi32(key, mask_shift);
		part = _mm512_and_epi32(part, mask_partitions_minus_1);
		part = _mm512_fmadd_epi32(part, mask_16, mask_lanes);
		__m512i count = _mm512_i32gather_epi32(part, all_counts, 4);
		count = _mm512_add_epi32(count, mask_1);
		_mm512_i32scatter_epi32(all_counts, part, count, 4);
	}
	// merge counts
	for (p = 0 ; p != partitions ; ++p) {
		__m512i sum = _mm512_load_epi32(&all_counts[p << 4]);
		counts[p] = _mm512_reduce_add_epi32(sum);
	}
}

void histogram_4(const uint32_t *keys, size_t size, uint32_t *counts,
                 uint8_t shift_bits, uint8_t radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	assert(keys == align(keys));
	assert((size & 15) == 0);
	// partition vector space
	uint32_t parts_space[31];
	uint32_t *parts = align(parts_space);
	// create masks
	__mmask16 blend_0 = _mm512_int2mask(0);
	__m512i mask_0 = _mm512_set1_epi32(0);
	__m512i mask_1 = _mm512_set1_epi32(1);
	__m512i mask_16 = _mm512_set1_epi32(16);
	__m512i mask_255 = _mm512_set1_epi32(255);
	__m512i mask_shift = _mm512_set1_epi32(shift_bits);
	__m512i mask_partitions_minus_1 = _mm512_set1_epi32(partitions - 1);
	__m512i mask_lanes = _mm512_set_epi32(15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0);
	// reset counts
	size_t partitions_x16 = partitions << 4;
	uint8_t all_counts_space[partitions_x16 + 127];
	uint8_t *all_counts = align(all_counts_space);
	for (p = 0 ; p < partitions_x16 ; p += 64)
		_mm512_store_epi32(&all_counts[p], mask_0);
	for (p = 0 ; p != partitions ; ++p)
		counts[p] = 0;
	// main loop
	for (i = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32(&keys[i]);
		__m512i part = _mm512_srlv_epi32(key, mask_shift);
		part = _mm512_and_epi32(part, mask_partitions_minus_1);
		__m512i part_lanes = _mm512_fmadd_epi32(part, mask_16, mask_lanes);
		__m512i count = _mm512_i32extgather_epi32(part_lanes, all_counts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__mmask16 k = _mm512_cmpeq_epi32_mask(count, mask_255);
		count = _mm512_add_epi32(count, mask_1);
		count = _mm512_and_epi32(count, mask_255);
		_mm512_i32extscatter_epi32(all_counts, part_lanes, count, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		if (!_mm512_kortestz(k, k)) {
			_mm512_store_epi32(parts, part);
			uint64_t mask = _mm512_kconcatlo_64(blend_0, k);
			size_t b = _mm_tzcnt_64(mask);
			do {
				p = parts[b];
				counts[p] += 256;
				b = _mm_tzcnti_64(b, mask);
			} while (b != 64);
		}
	}
	// merge counts
	for (p = 0 ; p != partitions ; ++p) {
		__m512i sum = _mm512_extload_epi32(&all_counts[p << 4], _MM_UPCONV_EPI32_UINT8, _MM_BROADCAST32_NONE, _MM_HINT_NONE);
		counts[p] += _mm512_reduce_add_epi32(sum);
	}
}

void partition_1(const uint32_t *keys, const uint32_t *vals, size_t size,
                 const uint32_t *counts, uint32_t *keys_out, uint32_t *vals_out,
                 uint8_t shift_bits, uint8_t radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	uint32_t offsets[partitions];
	for (i = p = 0 ; p != partitions ; ++p) {
		offsets[p] = i;
		i += counts[p];
	}
	assert(i == size);
	for (i = 0 ; i != size ; ++i) {
		uint32_t k = keys[i];
		uint32_t v = vals[i];
		p = k;
		p >>= shift_bits;
		p &= partitions - 1;
		size_t o = offsets[p]++;
		keys_out[o] = k;
		vals_out[o] = v;
	}
	for (i = p = 0 ; p != partitions ; ++p) {
		i += counts[p];
		assert(i == offsets[p]);
	}
	assert(i == size);
}

void partition_2(const uint32_t *keys, const uint32_t *vals, size_t size,
                 const uint32_t *counts, uint32_t *keys_out, uint32_t *vals_out,
                 uint8_t shift_bits, uint8_t radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	uint32_t offsets[partitions];
	uint64_t buffers_space[partitions * 16 + 7];
	uint64_t *buffers = align(buffers_space);
	for (i = p = 0 ; p != partitions ; ++p) {
		offsets[p] = i;
		i += counts[p];
	}
	assert(i == size);
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		uint64_t key_val = vals[i];
		key_val = (key_val << 32) | key;
		p = key;
		p >>= shift_bits;
		p &= partitions - 1;
		size_t o = offsets[p]++;
		size_t b = o & 15;
		uint64_t *buf = &buffers[p << 4];
		buf[b] = key_val;
		if (b == 15) {
			uint32_t *k_out = &keys_out[o - 15];
			uint32_t *v_out = &vals_out[o - 15];
			for (b = 0 ; b != 16 ; ++b) {
				key_val = buf[b];
				k_out[b] = key_val;
				v_out[b] = key_val >> 32;
			}
		}
	}
	for (i = p = 0 ; p != partitions ; ++p) {
		uint64_t *buf = &buffers[p << 4];
		size_t off = offsets[p];
		size_t cnt = counts[p];
		size_t end = off & 15;
		size_t beg = end > cnt ? end - cnt : 0;
		off -= end - beg;
		while (beg != end) {
			uint64_t key_val = buf[beg];
			keys_out[off] = key_val;
			vals_out[off] = key_val >> 32;
			off++;
			beg++;
		}
		i += cnt;
		assert(i == off);
	}
	assert(i == size);
}

void partition_3(const uint32_t *keys, const uint32_t *vals, size_t size,
                 const uint32_t *counts, uint32_t *keys_out, uint32_t *vals_out,
                 uint8_t shift_bits, uint8_t radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	assert(keys == align(keys));
	assert(vals == align(vals));
	assert((size & 15) == 0);
	// conflict space
	uint8_t conflicts_space[partitions + 63];
	uint8_t *conflicts = align(conflicts_space);
	// offset space
	uint32_t offsets_space[partitions + 15];
	uint32_t *offsets = align(offsets_space);
	// initialize offsets
	for (i = p = 0 ; p != partitions ; ++p) {
		offsets[p] = i;
		i += counts[p];
	}
	assert(i == size);
	// generate masks
	__m512i mask_1 = _mm512_set1_epi32(1);
	__m512i mask_shift = _mm512_set1_epi32(shift_bits);
	__m512i mask_partitions_minus_1 = _mm512_set1_epi32(partitions - 1);
	__m512i mask_reverse = _mm512_set_epi32(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
	// aligned loop
	for (i = 0 ; i != size ; i += 16) {
		// load keys and payloads
		__m512i key = _mm512_load_epi32(&keys[i]);
		__m512i val = _mm512_load_epi32(&vals[i]);
		// compute partitions
		__m512i part = _mm512_srlv_epi32(key, mask_shift);
		part = _mm512_and_epi32(part, mask_partitions_minus_1);
		// detect conflicts
		__m512i part_reverse = _mm512_permutevar_epi32(mask_reverse, part);
		_mm512_i32extscatter_epi32(conflicts, part_reverse, mask_reverse, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i back = _mm512_i32extgather_epi32(part_reverse, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i update = mask_1;
		__mmask16 k = _mm512_cmpneq_epi32_mask(back, mask_reverse);
		if (!_mm512_kortestz(k, k)) do {
			update = _mm512_mask_add_epi32(update, k, update, mask_1);
			_mm512_mask_i32extscatter_epi32(conflicts, k, part_reverse, mask_reverse, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
			back = _mm512_mask_i32extgather_epi32(back, k, part_reverse, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
			k = _mm512_cmpneq_epi32_mask(back, mask_reverse);
		} while (!_mm512_kortestz(k, k));
		// update offsets and detect conflits
		__m512i offset = _mm512_i32gather_epi32(part, offsets, 4);
		update = _mm512_permutevar_epi32(mask_reverse, update);
		offset = _mm512_add_epi32(offset, update);
		_mm512_i32scatter_epi32(offsets, part, offset, 4);
		// write data to output
		_mm512_i32scatter_epi32(&keys_out[-1], offset, key, 4);
		_mm512_i32scatter_epi32(&vals_out[-1], offset, val, 4);
	}
	// check partition sizes
	for (i = p = 0 ; p != partitions ; ++p) {
		i += counts[p];
		assert(i == offsets[p]);
	}
	assert(i == size);
}

void partition_4(const uint32_t *keys, const uint32_t *vals, size_t size,
                 const uint32_t *counts, uint32_t *keys_out, uint32_t *vals_out,
                 uint8_t shift_bits, uint8_t radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	assert(keys == align(keys));
	assert(vals == align(vals));
	assert(keys_out == align(keys_out));
	assert(vals_out == align(vals_out));
	assert((size & 15) == 0);
	// conflict space
	uint8_t conflicts_space[partitions + 63];
	uint8_t *conflicts = align(conflicts_space);
	// partition vector space
	uint32_t parts_space[31];
	uint32_t *parts = align(parts_space);
	// buffer space
	uint64_t buffers_space[(partitions << 4) + 7];
	uint64_t *buffers = align(buffers_space);
	// offset space
	uint32_t offsets_space[partitions + 15];
	uint32_t *offsets = align(offsets_space);
	// initialize offsets
	for (i = p = 0 ; p != partitions ; ++p) {
		offsets[p] = i;
		i += counts[p];
	}
	assert(i == size);
	// generate masks
	__m512i mask_shift = _mm512_set1_epi32(shift_bits);
	__m512i mask_partitions_minus_1 = _mm512_set1_epi32(partitions - 1);
	__m512i mask_pack = _mm512_set_epi32(15, 7, 14, 6, 13, 5, 12, 4, 11, 3, 10, 2, 9, 1, 8, 0);
	__m512i mask_unpack = _mm512_set_epi32(15, 13, 11, 9, 7, 5, 3, 1, 14, 12, 10, 8, 6, 4, 2, 0);
	__m512i mask_reverse = _mm512_set_epi32(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
	__m512i mask_16 = _mm512_set1_epi32(16);
	__m512i mask_15 = _mm512_set1_epi32(15);
	__m512i mask_1 = _mm512_set1_epi32(1);
	__mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	__mmask16 blend_5555 = _mm512_int2mask(0x5555);
	__mmask16 blend_0000 = _mm512_int2mask(0x0000);
	// aligned loop
	for (i = 0 ; i != size ; i += 16) {
		// load keys and payloads
		__m512i key = _mm512_load_epi32(&keys[i]);
		__m512i val = _mm512_load_epi32(&vals[i]);
		// compute partitions
		__m512i part = _mm512_srlv_epi32(key, mask_shift);
		part = _mm512_and_epi32(part, mask_partitions_minus_1);
		// detect conflicts
		__m512i part_reverse = _mm512_permutevar_epi32(mask_reverse, part);
		_mm512_i32extscatter_epi32(conflicts, part_reverse, mask_reverse, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i back = _mm512_i32extgather_epi32(part_reverse, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i update = mask_1;
		__mmask16 k = _mm512_cmpneq_epi32_mask(back, mask_reverse);
		if (!_mm512_kortestz(k, k)) do {
			update = _mm512_mask_add_epi32(update, k, update, mask_1);
			_mm512_mask_i32extscatter_epi32(conflicts, k, part_reverse, mask_reverse, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
			back = _mm512_mask_i32extgather_epi32(back, k, part_reverse, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
			k = _mm512_cmpneq_epi32_mask(back, mask_reverse);
		} while (!_mm512_kortestz(k, k));
		update = _mm512_permutevar_epi32(mask_reverse, update);
		// split the data into low and high part
		key = _mm512_permutevar_epi32(mask_pack, key);
		val = _mm512_permutevar_epi32(mask_pack, val);
		__m512i lo = _mm512_mask_blend_epi32(blend_AAAA, key, _mm512_swizzle_epi32(val, _MM_SWIZ_REG_CDAB));
		__m512i hi = _mm512_mask_blend_epi32(blend_5555, val, _mm512_swizzle_epi32(key, _MM_SWIZ_REG_CDAB));
		// update offsets and detect conflits
		__m512i offset = _mm512_i32gather_epi32(part, offsets, 4);
		__m512i offset_back = _mm512_add_epi32(offset, update);
		_mm512_i32scatter_epi32(offsets, part, offset_back, 4);
		// compute block offsets and block bounds
		offset = _mm512_and_epi32(offset, mask_15);
		offset = _mm512_add_epi32(offset, update);
		__mmask16 le = _mm512_cmple_epi32_mask(offset, mask_16);
		__mmask16 eq = _mm512_cmpeq_epi32_mask(offset, mask_16);
		__mmask16 gt = _mm512_knot(le);
		__m512i offset_lo = _mm512_fmadd_epi32(part, mask_16, offset);
		__m512i offset_hi = _mm512_permute4f128_epi32(offset_lo, _MM_PERM_BCDC);
		// write interleaved keys and payloads to buffers
		_mm512_mask_i32loscatter_epi64(&buffers[-1], le, offset_lo, lo, 8);
		le = _mm512_kmerge2l1h(le, le);
		_mm512_mask_i32loscatter_epi64(&buffers[-1], le, offset_hi, hi, 8);
		// flush full blocks (taken ~ 65%)
		if (!_mm512_kortestz(eq, eq)) {
			_mm512_store_epi32(parts, part);
			size_t mask = _mm512_kconcatlo_64(blend_0000, eq);
			size_t b = _mm_tzcnt_64(mask);
			do {
				size_t l = parts[b];
				size_t o = offsets[l] & ~15;
				l <<= 4;
				__m512i lo_out = _mm512_load_epi64(&buffers[l + 0]);
				__m512i hi_out = _mm512_load_epi64(&buffers[l + 8]);
				key = _mm512_mask_blend_epi32(blend_AAAA, lo_out, _mm512_swizzle_epi32(hi_out, _MM_SWIZ_REG_CDAB));
				val = _mm512_mask_blend_epi32(blend_5555, hi_out, _mm512_swizzle_epi32(lo_out, _MM_SWIZ_REG_CDAB));
				key = _mm512_permutevar_epi32(mask_unpack, key);
				val = _mm512_permutevar_epi32(mask_unpack, val);
				_mm512_storenrngo_ps(&keys_out[o - 16], _mm512_castsi512_ps(key));
				_mm512_storenrngo_ps(&vals_out[o - 16], _mm512_castsi512_ps(val));
				b = _mm_tzcnti_64(b, mask);
			} while (b != 64);
			// write overflows (taken < 3%)
			if (!_mm512_kortestz(gt, gt)) {
				_mm512_mask_i32loscatter_epi64(&buffers[-17], gt, offset_lo, lo, 8);
				gt = _mm512_kmerge2l1h(gt, gt);
				_mm512_mask_i32loscatter_epi64(&buffers[-17], gt, offset_hi, hi, 8);
			}
		}
	}
	// clean up buffers and check partition sizes
	for (i = p = 0 ; p != partitions ; ++p) {
		uint64_t *buf = &buffers[p << 4];
		size_t off = offsets[p];
		size_t cnt = counts[p];
		size_t end = off & 15;
		size_t beg = end > cnt ? end - cnt : 0;
		off -= end - beg;
		while (beg != end) {
			vals_out[off] = buf[beg] >> 32;
			keys_out[off] = buf[beg];
			off++;
			beg++;
		}
		i += cnt;
		assert(i == off);
	}
	assert(i == size);
}

uint64_t checksum(uint32_t *data, size_t size)
{
	uint64_t sum = 0;
	size_t i;
	for (i = 0 ; i != size ; ++i)
		sum += data[i];
	return sum;
}

size_t partitioned(uint32_t *keys, size_t size, uint8_t shift_bits, uint8_t radix_bits)
{
	size_t partitions = ((size_t) 1) << radix_bits;
	size_t i, old_p = 0;
	for (i = 0 ; i != size ; ++i) {
		size_t p = keys[i];
		p >>= shift_bits;
		p &= partitions - 1;
		if (old_p > p) break;
		old_p = p;
	}
	return i;
}

typedef struct {
	pthread_t id;
	int cpu;
	int seed;
	size_t size;
	uint64_t *times;
	uint8_t *partition_bits;
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->cpu);
	size_t i, p, b = 0, m = 0, size = d->size;
	uint32_t *keys     = mamalloc(size * sizeof(uint32_t));
	uint32_t *vals     = mamalloc(size * sizeof(uint32_t));
	uint32_t *keys_out = mamalloc(size * sizeof(uint32_t));
	uint32_t *vals_out = mamalloc(size * sizeof(uint32_t));
	unsigned int seed = (unsigned int) d->seed;
	rand32_t *gen = rand32_init(d->seed);
	for (i = 0 ; i != size ; ++i) {
		keys[i] = rand32_next(gen);
		vals[i] = ~keys[i];
		keys_out[i] = 0xCAFEBABE;
		vals_out[i] = 0xDEADBEEF;
	}
	uint64_t key_checksum = checksum(keys, size);
	uint64_t val_checksum = checksum(vals, size);
	for (p = 0 ; d->partition_bits[p] ; ++p) {
		uint8_t radix_bits = d->partition_bits[p];
		uint8_t shift_bits = rand32_next(gen) % (33 - radix_bits);
		size_t partitions = ((size_t) 1) << radix_bits;
		uint32_t counts[partitions];
		pthread_barrier_wait(&d->barrier[b++]);
		uint64_t t = thread_time();
		histogram_1(keys, size, counts, shift_bits, radix_bits);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		assert(checksum(counts, partitions) == size);
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		histogram_2(keys, size, counts, shift_bits, radix_bits);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		assert(checksum(counts, partitions) == size);
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		histogram_3(keys, size, counts, shift_bits, radix_bits);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		assert(checksum(counts, partitions) == size);
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		histogram_4(keys, size, counts, shift_bits, radix_bits);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		assert(checksum(counts, partitions) == size);
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		partition_1(keys, vals, size, counts, keys_out, vals_out, shift_bits, radix_bits);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		assert(size == partitioned(keys_out, size, shift_bits, radix_bits));
		assert(key_checksum == checksum(keys_out, size));
		assert(val_checksum == checksum(vals_out, size));
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		partition_2(keys, vals, size, counts, keys_out, vals_out, shift_bits, radix_bits);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		assert(size == partitioned(keys_out, size, shift_bits, radix_bits));
		assert(key_checksum == checksum(keys_out, size));
		assert(val_checksum == checksum(vals_out, size));
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		partition_3(keys, vals, size, counts, keys_out, vals_out, shift_bits, radix_bits);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		assert(size == partitioned(keys_out, size, shift_bits, radix_bits));
		assert(key_checksum == checksum(keys_out, size));
		assert(val_checksum == checksum(vals_out, size));
		pthread_barrier_wait(&d->barrier[b++]);
		t = thread_time();
		partition_4(keys, vals, size, counts, keys_out, vals_out, shift_bits, radix_bits);
		t = thread_time() - t;
		pthread_barrier_wait(&d->barrier[b++]);
		d->times[m++] = t;
		assert(size == partitioned(keys_out, size, shift_bits, radix_bits));
		assert(key_checksum == checksum(keys_out, size));
		assert(val_checksum == checksum(vals_out, size));
		if (!d->cpu) fprintf(stderr, "%ld partitions finished!\n", partitions);
	}
	pthread_barrier_wait(&d->barrier[b++]);
	free(gen);
	free(keys);
	free(vals);
	free(keys_out);
	free(vals_out);
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	srand(time(NULL));
	int t, threads = argc > 1 ? atoi(argv[1]) : hardware_threads();
	double gigs = argc > 2 ? atof(argv[2]) : 6.0;
	size_t p, size = gigs * 128 * 1024 * 1024;
	fprintf(stderr, "Tuples: %ld\n", size);
	fprintf(stderr, "Threads: %d\n", threads);
	char *names[] = {"Histogram scalar",
	                 "Histogram SIMD detecting",
	                 "Histogram SIMD uncompressed",
	                 "Histogram SIMD compressed",
	                 "Partition scalar unbuffered",
	                 "Partition scalar buffered",
	                 "Partition SIMD unbuffered",
	                 "Partition SIMD buffered"};
	uint8_t partition_bits[] = {3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 0};
	for (p = 0 ; partition_bits[p] ; ++p);
	int m, methods = sizeof(names) / sizeof(char*);
	int b, barriers = p * methods * 2 + 1;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	info_t info[threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].cpu = t;
		info[t].seed = rand();
		info[t].partition_bits = partition_bits;
		info[t].times = malloc(p * methods * sizeof(uint64_t));
		info[t].size = (size / threads) & ~15;
		info[t].barrier = barrier;
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	for (p = 0 ; partition_bits[p] ; ++p) {
		size_t partitions = ((size_t) 1) << partition_bits[p];
		fprintf(stderr, "Partitions: %ld\n", partitions);
		printf("%4ld", partitions);
		for (m = 0 ; m != methods ; ++m) {
			uint64_t nanos = 0;
			for (t = 0 ; t != threads ; ++t)
				nanos += info[t].times[p * methods + m];
			nanos /= threads;
			double billion = 1000 * 1000 * 1000;
			fprintf(stderr, "%-30s %6.4f secs %6.2f btps\n",
			        names[m], nanos / billion, size * 1.0 / nanos);
			printf(", %5.2f", size * 1.0 / nanos);
		}
		printf("\n");
	}
	for (t = 0 ; t != threads ; ++t)
		free(info[t].times);
	return EXIT_SUCCESS;
}
