/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#ifndef _NO_VECTOR
#include <immintrin.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <sched.h>
#include <time.h>

#include "rand.h"


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

uint64_t real_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_REALTIME, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread_id)
{
	int threads = hardware_threads();
	assert(thread_id >= 0 && thread_id < threads);
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread_id, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *mamalloc(size_t size)
{
	void *ptr = NULL;
	return posix_memalign(&ptr, 64, size) ? NULL : ptr;
}

void *align(const void *p)
{
	size_t i = 63 & (size_t) p;
	return (void*) (i ? p + 64 - i : p);
}

#ifndef _NO_VECTOR

void histogram(const uint32_t *keys, size_t size, uint32_t *counts,
               int shift_bits, int radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	assert(keys == align(keys));
	assert((size & 15) == 0);
	// create masks
	__m512i mask_0 = _mm512_set1_epi32(0);
	__m512i mask_1 = _mm512_set1_epi32(1);
	__m512i mask_16 = _mm512_set1_epi32(16);
	__m512i mask_shift_lo = _mm512_set1_epi32(32 - radix_bits);
	__m512i mask_shift_hi = _mm512_set1_epi32(32 - radix_bits - shift_bits);
	__m512i mask_lanes = _mm512_set_epi32(15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0);
	// reset counts
	size_t partitions_x16 = partitions << 4;
	uint32_t all_counts_space[partitions_x16 + 15];
	uint32_t *all_counts = align(all_counts_space);
	for (p = 0 ; p != partitions_x16 ; p += 16)
		_mm512_store_epi32(&all_counts[p], mask_0);
	// main loop
	for (i = 0 ; i != size ; i += 16) {
		__m512i key = _mm512_load_epi32(&keys[i]);
		__m512i part = _mm512_sllv_epi32(key, mask_shift_hi);
		part = _mm512_srlv_epi32(part, mask_shift_lo);
		part = _mm512_fmadd_epi32(part, mask_16, mask_lanes);
		__m512i count = _mm512_i32gather_epi32(part, all_counts, 4);
		count = _mm512_add_epi32(count, mask_1);
		_mm512_i32scatter_epi32(all_counts, part, count, 4);
	}
	// merge counts
	for (p = 0 ; p != partitions ; ++p) {
		__m512i sum = _mm512_load_epi32(&all_counts[p << 4]);
		counts[p] = _mm512_reduce_add_epi32(sum);
	}
}

int partition(const uint32_t *keys, size_t size, uint32_t *offsets,
              uint32_t *buffers, uint32_t *keys_out, int shift_bits, int radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	assert(keys == align(keys));
	assert(keys_out == align(keys_out));
	assert((size & 15) == 0);
	// conflict space
	uint8_t conflicts_space[partitions + 63];
	uint8_t *conflicts = align(conflicts_space);
	// partition vector space
	uint32_t parts_space[31];
	uint32_t *parts = align(parts_space);
	// generate masks
	__m512i mask_shift_lo = _mm512_set1_epi32(32 - radix_bits);
	__m512i mask_shift_hi = _mm512_set1_epi32(32 - radix_bits - shift_bits);
	__m512i mask_pack = _mm512_set_epi32(15, 7, 14, 6, 13, 5, 12, 4, 11, 3, 10, 2, 9, 1, 8, 0);
	__m512i mask_unpack = _mm512_set_epi32(15, 13, 11, 9, 7, 5, 3, 1, 14, 12, 10, 8, 6, 4, 2, 0);
	__m512i mask_reverse = _mm512_set_epi32(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
	__m512i mask_16 = _mm512_set1_epi32(16);
	__m512i mask_15 = _mm512_set1_epi32(15);
	__m512i mask_1 = _mm512_set1_epi32(1);
	__mmask16 blend_AAAA = _mm512_int2mask(0xAAAA);
	__mmask16 blend_5555 = _mm512_int2mask(0x5555);
	__mmask16 blend_0000 = _mm512_int2mask(0x0000);
	// aligned loop
	for (i = 0 ; i != size ; i += 16) {
		// load keys
		__m512i key = _mm512_load_epi32(&keys[i]);
		// compute partitions
		__m512i part = _mm512_sllv_epi32(key, mask_shift_hi);
		part = _mm512_srlv_epi32(part, mask_shift_lo);
		// detect conflicts
		__m512i part_reverse = _mm512_permutevar_epi32(mask_reverse, part);
		_mm512_i32extscatter_epi32(conflicts, part_reverse, mask_reverse, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i back = _mm512_i32extgather_epi32(part_reverse, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
		__m512i update = mask_1;
		__mmask16 k = _mm512_cmpneq_epi32_mask(back, mask_reverse);
		if (!_mm512_kortestz(k, k)) do {
			update = _mm512_mask_add_epi32(update, k, update, mask_1);
			_mm512_mask_i32extscatter_epi32(conflicts, k, part_reverse, mask_reverse, _MM_DOWNCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
			back = _mm512_mask_i32extgather_epi32(back, k, part_reverse, conflicts, _MM_UPCONV_EPI32_UINT8, 1, _MM_HINT_NONE);
			k = _mm512_cmpneq_epi32_mask(back, mask_reverse);
		} while (!_mm512_kortestz(k, k));
		update = _mm512_permutevar_epi32(mask_reverse, update);
		// update offsets and detect conflits
		__m512i offset = _mm512_i32gather_epi32(part, offsets, 4);
		__m512i offset_back = _mm512_add_epi32(offset, update);
		_mm512_i32scatter_epi32(offsets, part, offset_back, 4);
		// compute block offsets and block bounds
		offset = _mm512_and_epi32(offset, mask_15);
		offset = _mm512_add_epi32(offset, update);
		__mmask16 le = _mm512_cmple_epi32_mask(offset, mask_16);
		__mmask16 eq = _mm512_cmpeq_epi32_mask(offset, mask_16);
		__mmask16 gt = _mm512_knot(le);
		offset = _mm512_fmadd_epi32(part, mask_16, offset);
		// write keys to buffers
		_mm512_mask_i32scatter_epi32(&buffers[-1], le, offset, key, 4);
		// flush full blocks (taken ~ 65%)
		if (!_mm512_kortestz(eq, eq)) {
			_mm512_store_epi32(parts, part);
			size_t mask = _mm512_kconcatlo_64(blend_0000, eq);
			size_t b = _mm_tzcnt_64(mask);
			do {
				size_t l = parts[b];
				size_t o = offsets[l] & ~15;
				__m512 out = _mm512_load_ps(&buffers[l << 4]);
				_mm512_storenrngo_ps(&keys_out[o - 16], out);
				b = _mm_tzcnti_64(b, mask);
			} while (b != 64);
			// write overflows (taken < 3%)
			if (!_mm512_kortestz(gt, gt))
				_mm512_mask_i32scatter_epi32(&buffers[-17], gt, offset, key, 4);
		}
	}
	return 1;
}

size_t interleave(uint32_t **counts, uint32_t *offsets,
                  int radix_bits, size_t thread, size_t threads)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	assert(radix_bits >= 4);
	assert(offsets == align(offsets));
	uint32_t counts_sum_space[31];
	uint32_t *counts_sum = align(counts_sum_space);
	for (i = p = 0 ; p != partitions ; p += 16) {
		__m512i sum = _mm512_xor_epi32(sum, sum);
		size_t t = 0;
		if (thread) do {
			__m512i cur = _mm512_load_epi32(&counts[t][p]);
			sum = _mm512_add_epi32(sum, cur);
		} while (++t != thread);
		_mm512_store_epi32(&offsets[p], sum);
		do {
			__m512i cur = _mm512_load_epi32(&counts[t][p]);
			sum = _mm512_add_epi32(sum, cur);
		} while (++t != threads);
		_mm512_store_epi32(counts_sum, sum);
		for (t = 0 ; t != 16 ; ++t) {
			offsets[p + t] += i;
			i += counts_sum[t];
		}
	}
	return i;
}

#else

void histogram(const uint32_t *keys, size_t size, uint32_t *counts,
               int shift_bits, int radix_bits)
{
	size_t i, p, partitions = ((size_t) 1) << radix_bits;
	for (p = 0 ; p != partitions ; ++p)
		counts[p] = 0;
	for (i = 0 ; i != size ; ++i) {
		p = keys[i] >> shift_bits;
		p &= partitions - 1;
		counts[p]++;
	}
}

#ifndef _NO_BUFFER

int partition(const uint32_t *keys, size_t size, uint32_t *offsets,
              uint32_t *buffers, uint32_t *keys_out, int shift_bits, int radix_bits)
{
	size_t i, partitions = ((size_t) 1) << radix_bits;
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		size_t p = key;
		p >>= shift_bits;
		p &= partitions - 1;
		size_t o = offsets[p]++;
		size_t b = o & 15;
		uint32_t *buf = &buffers[p << 4];
		buf[b] = key;
		if (b == 15) {
			uint32_t *k_out = &keys_out[o - 15];
			for (b = 0 ; b != 16 ; ++b)
				k_out[b] = buf[b];
		}
	}
	return 1;
}

#else

int partition(const uint32_t *keys, size_t size, uint32_t *offsets,
              uint32_t *buffers, uint32_t *keys_out, int shift_bits, int radix_bits)
{
	size_t i, partitions = ((size_t) 1) << radix_bits;
	for (i = 0 ; i != size ; ++i) {
		uint32_t key = keys[i];
		size_t p = key;
		p >>= shift_bits;
		p &= partitions - 1;
		size_t o = offsets[p]++;
		keys_out[o] = key;
	}
	return 0;
}

#endif

size_t interleave(uint32_t **counts, uint32_t *offsets,
                  int radix_bits, size_t thread, size_t threads)
{
	size_t i, t, p, partitions = ((size_t) 1) << radix_bits;
	for (i = p = 0 ; p != partitions ; ++p) {
		for (t = 0 ; t != thread ; ++t)
			i += counts[t][p];
		offsets[p] = i;
		for (; t != threads ; ++t)
			i += counts[t][p];
	}
	return i;
}

#endif

void flush(const uint32_t *counts, const uint32_t *offsets, const uint32_t *buffers,
           uint32_t *keys_out, int radix_bits)
{
	size_t p, partitions = ((size_t) 1) << radix_bits;
	for (p = 0 ; p != partitions ; ++p) {
		const uint32_t *buf = &buffers[p << 4];
		size_t off = offsets[p];
		size_t cnt = counts[p];
		size_t end = off & 15;
		size_t beg = end > cnt ? end - cnt : 0;
		off -= end - beg;
		while (beg != end)
			keys_out[off++] = buf[beg++];
	}
}

void swap(uint32_t **x, uint32_t **y)
{
	uint32_t *t = *x; *x = *y; *y = t;
}

typedef struct {
	pthread_t id;
	int seed;
	int thread;
	int threads;
	int *bits;
	size_t size;
	uint32_t factor;
	uint32_t *keys[2];
	uint32_t **counts;
	uint64_t times[8];
	uint64_t *checksum[2];
	uint64_t real_time;
	uint64_t thread_time;
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread);
	uint32_t *keys_in  = d->keys[0];
	uint32_t *keys_out = d->keys[1];
	uint32_t **counts = d->counts;
	size_t size = d->size;
	size_t thread  = d->thread;
	size_t threads = d->threads;
	// read partitioning steps
	size_t max_partitions = 0;
	int *bits = d->bits;
	int sum_bits = 0;
	while (sum_bits < 32) {
		size_t partitions = ((size_t) 1) << *bits;
		if (partitions > max_partitions)
			max_partitions = partitions;
		sum_bits += *bits++;
	}
	assert(sum_bits == 32);
	bits = d->bits;
	// initialize parts
	size_t i, p;
	size_t part = (size / threads) & ~15;
	size_t beg = part *  thread;
	size_t end = part * (thread + 1);
	if (thread + 1 == threads) end = size;
	rand32_t *gen = rand32_init(d->seed);
	uint64_t sum = 0;
	uint32_t factor = d->factor;
	for (i = beg ; i != end ; ++i) {
		keys_in[i] = rand32_next(gen);
		keys_out[i] = 0xCAFEBABE;
		sum += keys_in[i];
	}
	d->checksum[0][thread] = sum;
	free(gen);
	for (p = 0 ; p != 8 ; ++p)
		d->times[p] = 0;
	pthread_barrier_t *barrier = d->barrier;
	pthread_barrier_wait(barrier++);
	// sort
	uint64_t tt = thread_time();
	uint64_t rt = real_time();
	uint32_t buffers_space[(max_partitions << 4) + 15];
	uint32_t offsets_space[max_partitions + 15];
	uint32_t counts_space[max_partitions + 15];
	uint32_t *buffers = align(buffers_space);
	uint32_t *offsets = align(offsets_space);
	counts[thread] = align(counts_space);
	int shift_bits = 0;
	uint64_t ptt = tt;
	for (p = 0 ;; ++p) {
		int radix_bits = *bits++;
		histogram(&keys_in[beg], end - beg,
		          counts[thread], shift_bits, radix_bits);
		pthread_barrier_wait(barrier++);
		i = interleave(counts, offsets, radix_bits, thread, threads);
		assert(i == size);
		if (partition(&keys_in[beg], end - beg, offsets,
		              buffers, keys_out, shift_bits, radix_bits)) {
			pthread_barrier_wait(barrier++);
			flush(counts[thread], offsets, buffers, keys_out, radix_bits);
		}
		shift_bits += radix_bits;
		uint64_t ntt = thread_time();
		d->times[p] = ntt - ptt;
		ptt = ntt;
		if (shift_bits == 32) break;
		pthread_barrier_wait(barrier++);
		swap(&keys_in, &keys_out);
	}
	rt = real_time() - rt;
	tt = thread_time() - tt;
	// validate
	pthread_barrier_wait(barrier++);
	uint32_t prev_key = beg ? keys_out[beg - 1] : 0;
	sum = 0;
	for (i = beg ; i != end ; ++i) {
		uint32_t curr_key = keys_out[i];
		assert(curr_key >= prev_key);
		sum += curr_key;
		prev_key = curr_key;
	}
	d->checksum[1][thread] = sum;
	d->thread_time = tt;
	d->real_time = rt;
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	int million_tuples = argc > 1 ? atoi(argv[1]) : 800;
	int threads = argc > 2 ? atoi(argv[2]) : hardware_threads();
	assert(million_tuples <= 1600);
	assert(threads > 0 && threads <= hardware_threads());
	size_t size = million_tuples * (size_t) 1000000;
	fprintf(stderr, "Tuples: %ld\n", size);
	fprintf(stderr, "Threads: %d\n", threads);
#ifdef _NO_VECTOR
	fprintf(stderr, "Vectorization disabled!\n");
#ifdef _NO_BUFFER
	fprintf(stderr, "Buffering disabled!\n");
#endif
#endif
	int t, b, barriers = 32;
	int bits[] = {6, 6, 6, 7, 7};
	fprintf(stderr, "Bits: %d", bits[0]);
	for (b = 1 ; b * sizeof(int) != sizeof(bits) ; ++b)
		fprintf(stderr, " -> %d", bits[b]);
	fprintf(stderr, "\n");
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	uint32_t *keys_1 = mamalloc(size * sizeof(uint32_t));
	uint32_t *keys_2 = mamalloc(size * sizeof(uint32_t));
	uint32_t *counts[threads];
	uint64_t checksum[2][threads];
	info_t info[threads];
	srand(time(NULL));
	uint32_t factor = rand();
	factor += factor + 1;
	for (t = 0 ; t != threads ; ++t) {
		info[t].bits = bits;
		info[t].seed = rand();
		info[t].thread = t;
		info[t].threads = threads;
		info[t].size = size;
		info[t].factor = factor;
		info[t].keys[0] = keys_1;
		info[t].keys[1] = keys_2;
		info[t].checksum[0] = checksum[0];
		info[t].checksum[1] = checksum[1];
		info[t].counts = counts;
		info[t].barrier = barrier;
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	uint64_t checksum_in = 0, checksum_out = 0;
	uint64_t min_tt = ~0, max_tt = 0, avg_tt = 0;
	uint64_t min_rt = ~0, max_rt = 0, avg_rt = 0;
	for (t = 0 ; t != threads ; ++t) {
		uint64_t rt = info[t].real_time;
		uint64_t tt = info[t].thread_time;
		if (rt > max_rt) max_rt = rt;
		if (tt > max_tt) max_tt = tt;
		if (rt < min_rt) min_rt = rt;
		if (tt < min_tt) min_tt = tt;
		avg_rt += rt;
		avg_tt += tt;
		checksum_in  += checksum[0][t];
		checksum_out += checksum[1][t];
	}
	avg_tt /= threads;
	avg_rt /= threads;
	assert(checksum_in == checksum_out);
	double max_rt_secs = max_rt / 1000000000.0;
	double min_rt_secs = min_rt / 1000000000.0;
	double avg_rt_secs = avg_rt / 1000000000.0;
	double max_tt_secs = max_tt / 1000000000.0;
	double min_tt_secs = min_tt / 1000000000.0;
	double avg_tt_secs = avg_tt / 1000000000.0;
	double gigs = size * 4.0 / (1024 * 1024 * 1024);
	fprintf(stderr, "Min: %6.4f sec | %5.3f GB/sec (real)\n", min_rt_secs, gigs / min_rt_secs);
	fprintf(stderr, "Avg: %6.4f sec | %5.3f GB/sec (real)\n", avg_rt_secs, gigs / avg_rt_secs);
	fprintf(stderr, "Max: %6.4f sec | %5.3f GB/sec (real)\n", max_rt_secs, gigs / max_rt_secs);
	fprintf(stderr, "Min: %6.4f sec | %5.3f GB/sec (thread)\n", min_tt_secs, gigs / min_tt_secs);
	fprintf(stderr, "Avg: %6.4f sec | %5.3f GB/sec (thread)\n", avg_tt_secs, gigs / avg_tt_secs);
	fprintf(stderr, "Max: %6.4f sec | %5.3f GB/sec (thread)\n", max_tt_secs, gigs / max_tt_secs);
	size_t p;
	for (p = 0 ; p != 8 ; ++p) {
		uint64_t partial_tt = 0;
		for (t = 0 ; t != threads ; ++t)
			partial_tt += info[t].times[p];
		if (partial_tt == 0) continue;
		partial_tt /= threads;
		double perc = partial_tt * 1.0 / avg_tt;
		fprintf(stderr, "Phase %ld: %5.2f%%, %6.4f secs\n", p + 1, perc * 100, perc * max_tt_secs);
	}
	printf("%d, %d, %.4f\n", threads, million_tuples, max_rt_secs);
	free(keys_1);
	free(keys_2);
	return EXIT_SUCCESS;
}
